/*
 * This Script is meant as a sample for making your own custom tools that enhances NeoDex.
 */

--name of macroscript, internalCategory, compatibility.
declareNeoDexPlugin "NeoDex_Scene_Manager" "Tools" NeoDexFullCompatibility

-- All the extra tools must be a macroscript.
-- GMAX Compatibility Note: category must be declared before internal category in the macroscript.
macroscript NeoDex_Scene_Manager
    buttonText:"Scene Manager"
    category:"NeoDex Toolkit"
	internalCategory:"Tools" 
(
	local version = "1.0"
	--("Scene Manager " + version)
	rollout nodeRollout "Node Manager" width:188 height:352
	(
		fn isGeometry o = classof o == Editable_Mesh
		fn isAttachmentPoint o =
		(	Try (Classof o == Wc3AttachPoint)
			Catch (false)
		)
		fn isCollisionShapeGmax o = (classOf o == Box or classOf o == Sphere)
		fn isCollisionShapeMax o =((classof o) as string == "Wc3CollisionBox") or ((classof o) as string == "Wc3CollisionSphere")
		local isCollisionShape = if gmax then isCollisionShapeGmax else isCollisionShapeMax
		fn isEventObject o = ((substring o.name 1 4) == "Obj:")
		fn isEffect o =
		(
			local t = classof o
			t == SuperSpray or t == BlizzParticle2 or t == BlizRibbon
		)
		mapped fn grayOut o b = o.xray = b
		GroupBox grp1 "Types" pos:[16,14] width:153 height:149
		checkbox geoCheck "Geometry" pos:[28,33] width:105 height:16
		checkbox bonesCheck "MaxBones" pos:[28,51] width:72 height:16
		checkbox attachmentsCheck "Attachments and  EventObjects" pos:[28,91] width:123 height:27
		checkbox lightsCheck "Lights" pos:[28,74] width:82 height:14
		checkbox effectsCheck "Effects" pos:[28,118] width:70 height:17
		checkbox colShapesCheck "Collision Shapes" pos:[28,140] width:105 height:15
		GroupBox grp5 "Actions" pos:[16,167] width:153 height:176
		GroupBox grp6 "Show" pos:[27,183] width:126 height:49
		GroupBox grp7 "Freeze" pos:[27,234] width:127 height:47
		GroupBox grp8 "See Through" pos:[27,282] width:126 height:53
		button grayOutSet "Set" pos:[37,305] width:39 height:17
		button grayOutUnset "Unset" pos:[101,304] width:39 height:17
		button freezeSet "Set" pos:[38,252] width:39 height:17
		button freezeUnset "Unset" pos:[102,251] width:39 height:17
		button showSet "Set" pos:[38,207] width:39 height:17
		button showUnset "Unset" pos:[102,206] width:39 height:17
		
		fn getExpectedGroup =
		(
			local result = #()
			if geoCheck.checked then
			(
				a = (for o in objects where isGeometry o collect o)
				join result a
			)
			if bonesCheck.checked then
			(
				a = (for o in objects where Filters.Is_Bone o collect o)
				join result a
			)
			if attachmentsCheck.checked then
			(
				a = (for o in objects where isAttachmentPoint o or isEventObject o collect o)
				join result a
			)
			if lightsCheck.checked then
			(
				a = (for o in objects where Filters.Is_Light o collect o)
				join result a
			)
			if effectsCheck.checked then
			(
				a = (for o in objects where isEffect o collect o)
				join result a
			)
			if colShapesCheck.checked then
			(
				a = (for o in objects where isCollisionShape o collect o)
				join result a
			)
			result
		)
		
		on grayOutSet pressed do
		(
			grayOut (getExpectedGroup ()) true
		)
		on grayOutUnset pressed do
		(
			grayOut (getExpectedGroup ()) false
		)
		on freezeSet pressed do
		(
			freeze (getExpectedGroup ())
		)
		on freezeUnset pressed do
		(
			unfreeze (getExpectedGroup ())
		)
		on showSet pressed do
		(
			unhide (getExpectedGroup ())
		)
		on showUnset pressed do
		(
			hide (getExpectedGroup ())
		)
	)
	
	
	rollout colorRollout "Color Manager" width:188 height:137
	(
		colorPicker cp1 "" pos:[131,38] width:29 height:24
		GroupBox grp1 "Team Color Manager" pos:[14,10] width:159 height:113
		spinner spn1 "Team Color" pos:[42,42] width:82 height:16 range:[1,16,1] type:#integer scale:1
		button btn1 "Change" pos:[27,88] width:86 height:20
		on spn1 changed newState do
		(
			cp1.color = NeoDexWc3MaterialTeamColors[newState] -- This global is declared in Wc3Material.ms
		)
		on btn1 pressed do
		(
			NeoDexUpdateTeamColor (spn1.value as integer) -- This function is declared in Wc3Material.ms
		)
	)
	local Main = newrolloutfloater "Scene Manager" 200 520
	addrollout nodeRollout Main
	addrollout colorRollout Main
	--mainRollout.title = "Scene Manager" + " " + (version as string) --fucking GMAX Compatibility
	colorRollout.cp1.color = NeoDexWc3MaterialTeamColors[NeoDexTeamColor]
	colorRollout.spn1.value = NeoDexTeamColor
)