
declareNeoDexPlugin "NeoDex_Sequence_Manager" "Tools" NeoDexFullCompatibility

macroscript NeoDex_Sequence_Manager
    buttonText:"Sequence Manager"
    category:"NeoDex Toolkit"
    internalCategory:"Tools" 
(
	include "NeoDexModules\\Wc3Model.ms"
	include "NeoDexModules\\UtilityFunctions.ms"
	local Sequences
	fn GetSequenceData =
	(
		local result = #()
		if HasNoteTracks Rootnode then
		(
			local Nt = GetNoteTrack Rootnode 1
			local keys = Nt.keys
			for k in keys do
			(
				local notFound = true
				for i = 1 to result.count where notFound do
				(
					if result[i][1].value == k.value then
					(
						append result[i] k
						notFound = false
					)
				)
				if notFound then
					append result #(k)
			)
		)
		for i = result.count to 1 by -1 where result[i].count != 2 do deleteItem result i
		result
	)

	fn readSequenceData key key2 =
	(
		local data = filterString key.value "\"\n\t\x0D"
		local animName = data[1]
		local noLoop = false
		local rarity, movespeed
		for i = 2 to data.count do 
		(
			local c = data[i]
			if stricmp c "Nonlooping" == 0 then NoLoop = true
			if stricmp (substring c 1 6) "Rarity" == 0 then rarity = ((substring c 7 -1) as integer)
			if stricmp (substring c 1 9) "Movespeed" == 0 then movespeed = ((substring c 10 -1) as integer)
		)
		Wc3Sequence animName key.time key2.time NoLoop Rarity Movespeed undefined
	)
	
	fn LoadSequences =
	(
		ArrayMap (GetSequenceData ()) (fn tmp a = readSequenceData a[1] a[2])
	)
	
	fn playSequence i =
	(
		if Sequences != undefined then
			if i > 0 and i <= Sequences.count then
			(
				animationRange = interval Sequences[i].startTime Sequences[i].endTime
				max time start
			)
	)
	
	rollout mainRollout "Sequence Manager" width:286 height:217
	(
		GroupBox grp6 "Data" pos:[102,6] width:175 height:167
		listbox lbx1 "Sequences" pos:[6,14] width:88 height:11
		edittext edt1 "" pos:[154,26] width:112 height:15 enabled:false
		label lbl1 "Name:" pos:[110,28] width:40 height:17
		checkbox chk1 "NonLooping" pos:[114,59] width:150 height:17 enabled:false
		spinner spn1 "Rarity" pos:[126,82] width:56 height:16 enabled:false range:[0,9,0] type:#integer scale:1
		spinner spn2 "MoveSpeed" pos:[138,106] width:117 height:16 enabled:false range:[0,1000,0] type:#integer scale:10
		button btn66 "Play" pos:[185,182] width:60 height:26
		button btn67 "Save Changes" pos:[146,134] width:80 height:26 enabled:false
		button btn68 "Add New" pos:[17,186] width:64 height:21 enabled:false
		button btn75 "Set Anim" pos:[114,182] width:57 height:26
		on lbx1 doubleClicked sel do
		(
			playSequence lbx1.selection
		)
		on btn66 pressed do
		(
			max time play
		)
		on btn75 pressed do
		(
			playSequence lbx1.selection
		)
	)
	
	Sequences = LoadSequences ()
	createDialog mainRollout
	mainRollout.lbx1.items = for s in Sequences collect s.name
)