
fn getTVFaces m map =
(
	local faces = #()
	local num = getNumFaces m
	if num != 0 then faces[num] = undefined
	for i = 1 to num do faces[i] = meshop.getMapFace m map i
	faces
)

fn getTVFacesOld m =
(
	local faces = #()
	local num = getNumFaces m
	if num != 0 then faces[num] = undefined
	for i = 1 to num do faces[i] = getTVFace m i
	faces
)

fn getFaces m =
(
	local faces = #()
	local num = getNumFaces m
	if num != 0 then faces[num] = undefined
	for i = 1 to num do faces[i] = getFace m i
	faces
)

fn splitBySmoothGroup obj faces =
(
	local result = #()
	local tmp = #()
	for f in faces do
	(
		local sg = getFaceSmoothGroup obj f
		local index = #()
		for i = 1 to tmp.count where (bit.and tmp[i] sg) != 0 do append index i
		if index.count != 0 then
		(
			local first = index[1]
			for i = index.count to 2 by -1 do
			(
				local j = index[i]
				join result[first] result[j]
				tmp[first] = bit.or tmp[first] tmp[j]
				deleteItem tmp j
				deleteItem result j
			)
			tmp[first] = bit.or tmp[first] sg
			append result[first] f
		)
		else
		(
			append tmp sg
			append result #(f)
		)
	)
	result
)

fn gatherSmoothGroupData obj =
(
	local numVerts = getNumVerts obj
	local baseFaces = getFaces obj
	local vertexMapper = #()
	local normals = #()
	vertexMapper[numVerts] = undefined
	normals[numVerts] = undefined
	local faces = getFaces obj
	local currentVert = 0
	local faceNormals = #()
	faceNormals[faces.count] = undefined
	local func = meshop.getFaceRNormals
	for f = 1 to faces.count do
		faceNormals[f] = func obj f
	func = splitBySmoothGroup
	local func2 = meshop.getFacesUsingVert
	for i = 1 to numVerts do
	(
		local g = func obj (func2 obj #{i})
		for a in g do
		(
			local k = 1
			for j = 1 to 3 where baseFaces[a[1]][j] == i do k = j
			currentVert += 1
			vertexMapper[currentVert] = i
			normals[currentVert] = faceNormals[a[1]][k]
			for f in a do
			(
				k = 1
				for j = 1 to 3 where baseFaces[f][j] == i do k = j
				faces[f][k] = currentVert
			)
		)
	)
	for i = vertexMapper.count to 1 by -1 where vertexMapper[i] == undefined do
	(
		deleteItem vertexMapper i
		deleteItem normals i
	)
	#(faces,normals,vertexMapper)
)

-- This function is simplified for only 1 map and no facegroups
fn breakUnwrapFast obj =
(
	local vertices = #()
	local texNum = getNumTVerts obj
	vertices[texNum] = undefined
	local faces = getFaces obj
	local tvfaces = getTVFacesOld obj
	local tmp = #()
	local count = faces.count
	for i = 1 to count do
	(
		-- there will be some over writting, but it's okay
		local tvf = tvfaces[i]
		local f = faces[i]
		vertices[tvf[1]] = f[1]
		vertices[tvf[2]] = f[2]
		vertices[tvf[3]] = f[3]
	)
	for f in tvfaces do
		for i = 1 to 3 do
		(
			local index = findItem tmp f[i]
			if index != 0 then f[i] = index else
			(
				append tmp f[i]
				f[i] = tmp.count
			)
		)
	#(vertices,tvfaces,tmp) --returns a mapper, fixed faces and count of vertices
)


--This function breaks faces on a mesh for a single map
fn breakUnwrapMap obj faces map =
(
	local vertices = #()
	local texNum = meshop.getNumMapVerts obj map
	vertices[texNum] = undefined
	local tvfaces = getTVFaces obj map
	local tmp = #()
	local count = faces.count
	local extraVertices = #()
	local indexer = #()
	for i = 1 to count do
	(
		-- there will be some over writting, but it's okay
		local tvf = tvfaces[i]
		local f = faces[i]
		for j = 1 to 3 do
		(
			-- We must check if the faces were partitioned before
			if vertices[tvf[j]] == undefined then
				vertices[tvf[j]] = f[j]
			else
				if vertices[tvf[j]] != f[j] then 
				( -- we found a partition
					local index = findItem indexer tvf[j]
					if index != 0 then
						tvf[j] = texNum + index
					else 
					(
						append indexer tvf[j]
						tvf[j] = indexer.count + texNum
						append extraVertices f[j]
					)
				)
		)
	)
	join vertices extraVertices
	for f in tvfaces do
		for i = 1 to 3 do
		(
			local index = findItem tmp f[i]
			if index != 0 then f[i] = index else
			(
				append tmp f[i]
				f[i] = tmp.count
			)
		)
	tvertices = #()
	tvertices[tmp.count] = undefined
	local func = meshop.getMapVert
	for i = 1 to tmp.count do
	(
		local index = tmp[i]
		if index > texNum then index = indexer[index-texNum]
		tvertices[i] = func obj map index
	)
	#(vertices,tvfaces,tmp,tvertices) --returns a mapper, fixed faces and count of vertices
)

--This function will break the mesh for each map continously
fn breakUnwrap obj faces:undefined numVerts:undefined =
(
	fn compareMaps a b = if (a[2] < b[2]) then -1 else 1
	if faces == undefined then
		faces = getFaces obj
	if numVerts == undefined then numVerts = getNumVerts obj
	local oldMapper = #()
	oldMapper[numVerts] = undefined
	for i = 1 to numVerts do oldMapper[i] = i
	local Maps = #()
	local tmp = #()
	-- We must gather the valid maps and number of vertices
	for i =1 to (meshop.getNumMaps obj) do
		if (meshop.getMapSupport obj i) then
			append tmp #(i,(meshop.getNumMapVerts obj i))
	-- Now we sort them according to highest number of tvertices
	qsort tmp compareMaps
	-- we do that to avoid problems and reduce partition calculations.
	Maps[tmp.count] = undefined
	for i = 1 to tmp.count do Maps[i] = tmp[i][1]
	local oldTVertices = #()
	for map in Maps do
	(
		local tmp = breakUnwrapMap obj faces map
		faces = tmp[2]
		local transformer = tmp[1]
		local vertexIndices = tmp[3]
		local count = vertexIndices.count
		local newMapper = #()
		local currentMaps = oldTVertices.count
		local newTVertices = #()
		local maptvertices = #()
		local tverts = tmp[4]
		if (currentMaps != 0) then
			newTVertices = ArrayRepeat currentMaps  #()
		for i = 1 to oldTvertices.count do
			newTVertices[i][count] = undefined
		newMapper[count] = undefined
		for i = 1 to count do
		(
			local v = transformer[vertexIndices[i]]
			newMapper[i] = oldMapper[v]
			for j = 1 to currentMaps do
				newTVertices[j][i] = oldTVertices[j][v]
		)
		oldMapper = newMapper
		append newTVertices tverts
		oldTVertices = newTVertices
	)
	for i = 1 to Maps.count do
		Maps[i] = #(oldTVertices[i],Maps[i])
	qsort Maps compareMaps
	for i = 1 to Maps.count do
		oldTVertices[i] = Maps[i][1]
	-- we will return final faces, final vertex groups and the mapper for vertices, normals and matrixindices
	#(faces,oldTVertices,oldMapper)
)

fn ColorVertices obj newColor =
(
	defaultVCFaces obj
	local vertNum = getNumCPVVerts obj
	for i = 1 to vertNum do setVertColor obj i newColor
)

--This function will split faces by elements
fn getFaceElements faces =
(
	fn faceToBitArray f = #{f[1],f[2],f[3]}
	local mapper = #{}
	mapper.count = faces.count
	mapper = -mapper
	mapper[mapper.count] = false
	local bitFaces = #()
	bitFaces[faces.count] = undefined
	for i = 1 to faces.count do
		bitFaces[i] = faceToBitArray faces[i]
	local result = #(bitFaces[mapper.count])
	local result2 = #(#{mapper.count})
	while (not mapper.isEmpty) do
	(
		local grouped = false
		for i = 1 to result.count do
		(
			local found = false
			local elem = result[i]
			for b in mapper do
			(
				local tmp = elem*bitFaces[b]
				if (not tmp.isEmpty) then
				(
					elem = elem + bitFaces[b]
					mapper[b] = false
					result2[i][b] = true
					found = true
				)
			)
			result[i] = elem
			grouped = grouped or found
		)
		if ((not grouped) and (not mapper.isEmpty)) then
		(
			local newFace
			local i = mapper.count
			while (i != 0) do
				if mapper[i] then
				(
					newFace = bitFaces[i]
					mapper[i] = false
					append result newFace
					append result2 #{i}
					i = 0
				) else i = i - 1
		)
	)
	result2
)

