
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *  Class: Graph
 *  Allows you to build graphs for finding face groups
 *  Author: Fernando A. Sahmkow
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */
(
    fn getFirstBitArray bitarr =
    (
        result = undefined
        for b in bitarr where result == undefined do result = b
        result
    )
    fn depthFirstSearchHelper startNode connections &visited &nodeGroup &func:unsupplied =
    (
        if (not nodeGroup[startNode]) then
        (
            visited = visited + 1
            nodeGroup[startNode] = true
            if func != unsupplied then
                func startNode
            for c in connections[startNode] where c != undefined do
                depthFirstSearchHelper c connections &visited nodeGroup func:func
        )
        OK
    )
    fn relateNodes connections &startNode &endNode =
    (
        local index = findItem connections[startNode] endNode
        if index== 0 then
            append connections[startNode] endNode
    )
    struct _Graph
    (
        connections = #(),
        self,
        fn create numNodes =
        (
            g = _Graph ()
            g.connections[numNodes] = undefined
            for i = 1 to numNodes do
            (
                g.connections[i] = #()
            )
            g.self = g
            g
        ),
        fn addConnection startNode endNode bothSided:false =
        (
            relateNodes connections startNode endNode
            if bothSided then
               relateNodes connections endNode startNode
			OK
        ),
        fn depthFirstSearch &func:unsupplied returnQuery:true =
        (
            local notSeen = #{}
			local numNodes = connections.count
			notSeen.count = numNodes
			notSeen = -notSeen
            local result = #()
            local visited = 0
            while (visited != numNodes) do
            (
                local nodeGroup = #{}
                nodeGroup.count = numNodes
                local startnode = getFirstBitArray notSeen
                depthFirstSearchHelper startNode connections &visited nodeGroup func:func
				append result nodeGroup
                notSeen = notSeen - nodeGroup
            )
            if returnQuery then
                result
            else OK
        )
    )
    global Graph
    fn Graph numNodes =
    (
        _Graph.create numNodes
    )
)

/*
g = Graph 5
g.addConnection 1 3
g.addConnection 2 1
g.addConnection 3 2
g.addConnection 4 5
g.addConnection 5 4
g.depthfirstSearch ()
*/