
global NeoDexIO = #Loaded
global outstream
global instream
global IOLevel = 1
global Tabs = #("","\t","\t\t","\t\t\t","\t\t\t\t","\t\t\t\t\t","\t\t\t\t\t\t", "\t\t\t\t\t\t\t")

---------------------------------------------------------------------------------------------------------------------------------
-- 												Write Functions											  				  ---
---------------------------------------------------------------------------------------------------------------------------------
fn SetOutStream os = outstream = os
fn SetInStream os = instream = os

fn isSmoothAnim type = (type == #Bezier or type == #Hermite)

fn checkExtents ex = ex.minExtent != ex.maxExtent
	
fn MaxMaterialAlphaToWc3Alpha &alphaKeys =
(
	if (alphaKeys.type == #Static) then \
		alphaKeys.data = alphaKeys.data/100.0
	else
	(
		local data = alphaKeys.data
		local IsSmooth =  isSmoothAnim data.type
		for k in data.keys do
		(
			k.value = k.value/100.0
			if IsSmooth then
			(
				k.intan = k.intan /100.0
				k.outtan = k.outtan / 100.0
			)
		)
	)
	alphaKeys
)

fn writterTransformKeys &Keys &func =
(
	if (Keys.type == #Static) then \
		Keys.data = func Keys.data
	else
	(
		local data = Keys.data
		local IsSmooth =  isSmoothAnim data.type
		for k in data.keys do
		(
			k.value = func k.value
			if IsSmooth then
			(
				k.intan = func k.intan
				k.outtan = func k.outtan
			)
		)
	)
	Keys
)

fn IOFixColor val =
(
	if (val.type == #Static) then \
		val.data = [val.data.r,val.data.g,val.data.b]
)

fn rotateCCW v1 =
(
	local tmp = -v1.y
	v1.y = v1.x
	v1.x = tmp
)

fn transformExtents extents =
(
	local v1 = extents.maxExtent
	local v2 = extents.minExtent
	rotateCCW v1
	rotateCCW v2
	if v1.x < v2.x then
	(
		local tmp = v1.x
		v1.x = v2.x
		v2.x = tmp
	)
	if v1.y < v2.y then
	(
		local tmp = v1.y
		v1.y = v2.y
		v2.y = tmp
	)
	extents.maxExtent = v1
	extents.minExtent = v2
)

---------------------------------------------------------------------------------------------------------------------------------
-- 												MDL Write Functions											  		  ---
---------------------------------------------------------------------------------------------------------------------------------

fn swrite outstring = format outstring to:outstream

fn PairToString x y =
(
	"{ " + (x as string) + ", " + (y as string) + " }"
)

fn TripletToString x y z =
(
	"{ " + (x as string) + ", " + (y as string) +  ", " + (z as string) +" }"
)

fn Point2ToString p = PairToString p.x p.y
fn Point3ToString p = TripletToString p.x p.y p.z
fn IntegerPoint3ToString p = TripletToString (p.x as integer) (p.y as integer) (p.z as integer)
fn PositionToString p = TripletToString (-p.y) p.x p.z
fn ScaleToString p = TripletToString p.y p.x p.z
fn ColorToString p = TripletToString (p.z/255) (p.y/255) (p.x/255)
fn ColorToString2 c = TripletToString (c.b/255) (c.g/255) (c.r/255)
fn TvertexToString p = PairToString p.x (1-p.y)

-- This functions are used by geosets writter, in order to increase performance
fn writeMDLTVertex p =
	format "\t\t{ %, % },\n" p.x (1-p.y) to:outstream
fn writeMDLVertex p =
	format "\t\t{ %, %, % },\n" (-p.y) p.x p.z to:outstream
fn writeMDLSingleValue v =
	format "\t\t%,\n" v to:outstream
fn writeMDLFace p =
	format "\t\t\t{ %, %, % },\n" (p.x - 1) (p.y - 1) (p.z-1) to:outstream

fn QuatToString q = 
(
	"{ " + ((q.y*(-1)) as string) + \
	", " + (q.x as string) + \
	", " + (q.z as string) + \
	", " + (q.w as string) + \
	" }"
)

fn TrianglesToString tris =
(
	fn tmp a = ((a as integer - 1) as string)
	local result = "{ " + (tmp tris[1].x) + ", " + (tmp tris[1].y) + ", " + (tmp tris[1].z)
	for i = 2 to tris.count do result = result +  ", " + (tmp tris[i].x) + ", " + (tmp tris[i].y) + ", " + (tmp tris[i].z)
	result + " }"
)

--used only by gmax
fn writeMDLTriangles tris =
(
	fn tmp a = ((a as integer - 1) as string)
	format "\t\t\t" to:outstream
	local result = "{ " + (tmp tris[1].x) + ", " + (tmp tris[1].y) + ", " + (tmp tris[1].z)
	for i = 2 to tris.count do
	(
		result = result +  ", " + (tmp tris[i].x) + ", " + (tmp tris[i].y) + ", " + (tmp tris[i].z)
		if (mod i 500 == 0) then
		(
			format "%" result to:outstream
			result = ""
		)
	)
	result = result + " },"
	format "%\n" result to:outstream
)

fn ArrayToString array &func =
(
	if (array[1] != undefined) then
	(
		local result = "{ " + (func array[1])
		for i = 2 to array.count do result = result +  ", " + (func array[i])
		result + " }"
	) else "{ }"
)

fn BillboardToString b =
(
	case (b) of
	(
		#Full: "Billboarded"
		#X: "BillboardedLockX"
		#Y: "BillboardedLockY"
		#Z: "BillboardedLockZ"
		default: "Billboarded"
	)
)

fn ControlToString c =
(
	case (c) of
	(
		#Bezier: "Bezier"
		#Hermite: "Hermite"
		#Linear: "Linear"
		#NonInterp: "DontInterp"
		default: "DontInterp"
	)
)

fn FilterModeToString f =
(
	case (f) of
	(
		1: "None"
		2: "Transparent"
		3: "Blend"
		4: "Additive"
		5: "AddAlpha"
		6: "Modulate"
		7: "Modulate2x"
		default: "None"
	)
)

fn ParticleFiltermodeToString i =
(
	case i of
	(
		1: "Blend"
		2: "Additive"
		3: "Modulate"
		4: "Modulate2x"
		5: "AlphaKey"
	)
)

fn ParticleTypeToString i =
(
	case i of
	(
		1:"Head"
		2:"Tail"
		3:"Both"
	)
)

fn StringWithColomms s = "\"" + s + "\""

fn writeMDLSection label data =
(
	swrite (Tabs[IOLevel] + label + " " + (data as string) + " {\n")
	IOLevel = IOLevel + 1
)

fn newIOLevel =
(
	IOLevel = IOLevel + 1
)

fn CloseIOLevel =
(
	IOLevel = IOLevel - 1
)


fn writeMDLValue label data =
(
	swrite (Tabs[IOLevel] + label + " " + (data as string) + ",\n")
)

fn writeMDLComment label =
(
	swrite (Tabs[IOLevel] + "// " + (label as string) + "\n")
)

fn writeMDLPoint data =
(
	swrite (Tabs[IOLevel] + (data as string) + ",\n")
)

fn writeMDLSectionEnd =
(
	IOLevel = IOLevel - 1
	swrite (Tabs[IOLevel] + "}\n")
)

fn writeMDLFlag label =
(
	swrite (Tabs[IOLevel]+ label + ",\n")
)

fn writeMDLKey t p =
(
	swrite (Tabs[IOLevel]+ (t as string) + ": " + p + ",\n")
)

fn writeMDLExtents extents =
(
	local tmp2 = extents.maxExtent.y
	extents.maxExtent.y = extents.minExtent.y
	extents.minExtent.y = tmp2
	local tmp = Tabs[IOLevel]
	swrite ( \
		tmp + "MinimumExtent " + (PositionToString extents.minExtent) + ",\n" + \
		tmp + "MaximumExtent " + (PositionToString extents.maxExtent) + ",\n" + \
		tmp + "BoundsRadius " + (extents.bound as string) + ",\n" \
	)
)

fn resetLevel = IOLevel = 1

fn floatToString f = (f as string)

--------------------------------------------------------------------------------------------------------------------------------
--												MDX TAGS																 ---
--------------------------------------------------------------------------------------------------------------------------------

fn IntAsTag i = (
	fn getP a s = bit.and (bit.shift a (-s)) 0xFF
	(bit.intAsChar (getP i 0)) + (bit.intAsChar (getP i 8)) + (bit.intAsChar (getP i 16)) + (bit.intAsChar (getP i 24))
)

local tag_MDLX = 0x584c444d
local tag_VERS = 0x53524556
local tag_MODL = 0x4c444f4d
local tag_SEQS = 0x53514553
local tag_GLBS = 0x53424c47
local tag_TEXS = 0x53584554
local tag_LAYS = 0x5359414c
local tag_KMTF = 0x46544d4b
local tag_KMTA = 0x41544d4b
local tag_MTLS = 0x534c544d
local tag_SNDS = 0x53444e53
local tag_TXAN = 0x4e415854
local tag_KTAT = 0x5441544b
local tag_KTAR = 0x5241544b
local tag_KTAS = 0x5341544b
local tag_GEOS = 0x534f4547
local tag_VRTX = 0x58545256
local tag_NRMS = 0x534d524e
local tag_PTYP = 0x50595450
local tag_PCNT = 0x544e4350
local tag_PVTX = 0x58545650
local tag_GNDX = 0x58444e47
local tag_MTGC = 0x4347544d
local tag_MATS = 0x5354414d
local tag_UVAS = 0x53415655
local tag_UVBS = 0x53425655
local tag_GEOA = 0x414f4547
local tag_KGAO = 0x4f41474b
local tag_KGAC = 0x4341474b
local tag_KGTR = 0x5254474b
local tag_KGRT = 0x5452474b
local tag_KGSC = 0x4353474b
local tag_BONE = 0x454e4f42
local tag_LITE = 0x4554494c
local tag_KLAS = 0x53414c4b
local tag_KLAE = 0x45414c4b
local tag_KLAC = 0x43414c4b
local tag_KLAI = 0x49414c4b
local tag_KLBI = 0x49424c4b
local tag_KLBC = 0x43424c4b
local tag_KLAV = 0x56414c4b
local tag_HELP = 0x504c4548
local tag_ATCH = 0x48435441
local tag_KATV = 0x5654414b
local tag_PIVT = 0x54564950
local tag_PREM = 0x4d455250
local tag_KPEE = 0x4545504b
local tag_KPEG = 0x4745504b
local tag_KPLN = 0x4e4c504b
local tag_KPEL = 0x4c45504b
local tag_KPES = 0x5345504b
local tag_KPEV = 0x5645504b
local tag_PRE2 = 0x32455250
local tag_KP2S = 0x5332504b
local tag_KP2R = 0x5232504b
local tag_KP2L = 0x4c32504b
local tag_KP2G = 0x4732504b
local tag_KP2E = 0x4532504b
local tag_KP2N = 0x4e32504b
local tag_KP2W = 0x5732504b
local tag_KP2V = 0x5632504b
local tag_SNEM = 0x4d454e53
local tag_KESK = 0x4b53454b
local tag_RIBB = 0x42424952
local tag_KRHA = 0x4148524b
local tag_KRHB = 0x4248524b
local tag_KRAL = 0x4c41524b
local tag_KRCO = 0x4f43524b
local tag_KRTX = 0x5854524b
local tag_KRVS = 0x5356524b
local tag_EVTS = 0x53545645
local tag_KEVT = 0x5456454b
local tag_CAMS = 0x534d4143
local tag_KCTR = 0x5254434b
local tag_KTTR = 0x5254544b
local tag_KCRL = 0x4c52434b
local tag_CLID = 0x44494c43


---------------------------------------------------------------------------------------------------------------------------------
-- 												MDX Write Functions											  		  ---
---------------------------------------------------------------------------------------------------------------------------------

fn ControlToInt c =
(
	case (c) of
	(
		#Bezier: 3
		#Hermite: 2
		#Linear: 1
		#NonInterp: 0
		default: 0
	)
)

fn writeMDXHead string=
(
	writeByte outstream (bit.charAsInt string[1])
	writeByte outstream (bit.charAsInt string[2])
	writeByte outstream (bit.charAsInt string[3])
	writeByte outstream (bit.charAsInt string[4])
)

fn writeMDXString string n=
(
	for i=1 to n do
	(
		if i <= string.count then writeByte outstream (bit.charAsInt string[i])
		else writeByte outstream 0
	)
)

fn writeMDXFlags bitArr = 
(
	result = 0
	for b in bitArr do result += b
	writeLong outstream result
)

fn writeMDXLong val = writeLong outstream val
fn writeMDXFloat val = writeFloat outstream val
fn writeMDXByte val = writeByte outstream val
fn writeMDXShort val = writeShort outstream val

fn writeMDXFace val =
(
	writeShort outstream ((val.x as integer) - 1)
	writeShort outstream ((val.y as integer) - 1)
	writeShort outstream ((val.z as integer) - 1)
)

fn writeMDXPoint3Long val =
(
	writeLong outstream (val.x as integer)
	writeLong outstream (val.y as integer)
	writeLong outstream (val.z as integer)
)

fn writeMDXPoint3 val =
(
	writeFloat outstream val.y
	writeFloat outstream val.x
	writeFloat outstream val.z
)

fn writeMDXPosition val =
(
	writeFloat outstream (-val.y)
	writeFloat outstream val.x
	writeFloat outstream val.z
)

fn writeMDXQuaternion val =
(
	writeFloat outstream (-val.y)
	writeFloat outstream val.x
	writeFloat outstream val.z
	writeFloat outstream val.w
)

fn writeMDXTVertex val =
(
	writeFloat outstream val.x
	writeFloat outstream (1-val.y)
)

fn writeMDXColor val =
(
	writeFloat outstream (val.x/255.0)
	writeFloat outstream (val.y/255.0)
	writeFloat outstream (val.z/255.0)
)

fn writeMDXExtents extents =
(
	--local tmp = extents.maxExtent.y
	--extents.maxExtent.y = extents.minExtent.y
	--extents.minExtent.y = tmp
	transformExtents extents
	writeFloat outstream extents.bound
	writeFloat outstream extents.minExtent.x
	writeFloat outstream extents.minExtent.y
	writeFloat outstream extents.minExtent.z
	writeFloat outstream extents.maxExtent.x
	writeFloat outstream extents.maxExtent.y
	writeFloat outstream extents.maxExtent.z
	--writeMDXPosition extents.minExtent
	--writeMDXPosition extents.maxExtent
)

fn posStamp =  (ftell outstream)

--returns the start of the chunk
fn InitiateChunk name includeName:true =
(
	if includeName then writeMDXHead name
	local pos = (ftell outstream)
	writeLong outstream 0
	pos
)

--takes the start of the chunk
fn FinishChunk oldPos inclusive:4 =
(
	local currentPos = ftell outstream
	fseek outstream oldPos #seek_set
	writeMDXLong (currentPos - oldPos - inclusive)
	fseek outstream currentPos #seek_set
)

---------------------------------------------------------------------------------------------------------------------------------
-- 												MDX Read Functions											  		  ---
---------------------------------------------------------------------------------------------------------------------------------

fn readMDXByte = readByte instream #unsigned
fn readMDXLong = readLong instream #unsigned
fn readMDXShort = readShort instream #unsigned
fn readMDXFloat = readFloat instream
fn readMDXString size =
(
	result = ""
	for i = 1 to size do
		result += bit.intaschar (readByte instream #unsigned)
	result
)

fn readMDXPosition =
(
	local y = -(readFloat instream)
	local x = readFloat instream
	local z = readFloat instream
	[x,y,z]
)

fn readMDXVector3 =
(
	local x = readFloat instream
	local y = readFloat instream
	local z = readFloat instream
	[x,y,z]
)

fn readMDXExtents =
(
	ex = Wc3Extent ()
	ex.bound = readMDXFloat ()
	ex.minExtent = readMDXVector3 ()
	ex.maxExtent = readMDXVector3 ()
)

fn readMDXVector3Long =
(
	local x = readLong instream #unsigned
	local y = readLong instream #unsigned
	local z = readLong instream #unsigned
	[x,y,z]
)

fn readMDXScale =
(
	local y = readFloat instream
	local x = readFloat instream
	local z = readFloat instream
	[x,y,z]
)

fn readMDXQuaternion =
(
	local y = -(readFloat instream)
	local x = readFloat instream
	local z = readFloat instream
	local w = readFloat instream
	quat x y z w
)

fn readMDXTvertex =
(
	local u = readFloat instream
	local v = (1 - (readFloat instream))
	[u,v]
)

fn readMDXFace =
(
	local i1 = readShort instream #unsigned
	local i2 = readShort instream #unsigned
	local i3 = readShort instream #unsigned
	[i1,i2,i3]
)

fn readMDXColor =
(
	local r = round ((readFloat instream)*255.0)
	local g = round ((readFloat instream)*255.0)
	local b = round ((readFloat instream)*255.0)
	[r,g,b]
)

fn readMDXTime =
(
	local t = (readLong instream #unsigned)*framerate
	t = round (t as float/100.0)
	(t/10.0 as time)
)

fn peekMDXHeader =
(
	local i = readLong instream #unsigned
	fseek instream -4 #seek_cur
	i
)

fn jumpMDXReading bytes =
(
	fseek instream bytes #seek_cur
)

fn readMDXArray size &readFunc =
(
	local result = #()
	result[size] = undefined
	for i=1 to size do result[i] = readFunc ()
	result
)

fn MDXSkipChunk =
(
	local chunksize = ReadMDXLong ()
	fseek instream chunksize #seek_cur
)

fn MDXReadPos = ftell instream

local priv_chunkSize = 0
fn DebugMDxChunk = print priv_chunkSize
fn MDXStartReadingChunk = priv_chunkSize = (readMDXLong ())
fn MDXHasChunkEnded = priv_chunkSize <= 0
fn MDXStartChunkElem specialCase:0 =priv_chunkSize -= ((readMDXLong ()) + specialCase)
fn MDXFixChunkSize rest =priv_chunkSize -= rest

fn IntToFlagsMax num mx =
(
	result = #()
	result[mx] = false
	for i = 1 to mx do
		result[i] = I2B (bit.and (bit.shift num (1-i)) 0x01)
	result
)

fn IntToFlagsGMAX num mx =
(
	result = #()
	result[mx] = false
	local n = 1
	for i = 1 to mx do
	(
		result[i] = I2B (bit.and (num/n) 0x01)
		n = n*2
	)
	result
)

if gmax then IntToFlags = IntToFlagsGMAX else IntToFlags = IntToFlagsMax
