
fn writeMDLAnimation label data &transformationFunc =
(
	local keys = data.keys
	local isSmooth = isSmoothAnim data.type
	writeMDLSection label keys.count
		writeMDLFlag (ControlToString data.type)
		if data.globalSequenceId != undefined then writeMDLValue "GlobalSeqId" (data.globalSequenceId - 1)
		if isSmooth then
		(
			for k in keys do
			(
				writeMDLKey (F2WS k.time) (transformationFunc k.value)
				newIOLevel ()
				writeMDLValue "InTan" (transformationFunc k.intan)
				writeMDLValue "OutTan" (transformationFunc k.outtan)
				CloseIOLevel ()
			)
		)
		else \
			for k in keys do writeMDLKey (F2WS k.time) (transformationFunc k.value)
	writeMDLSectionEnd ()
)

fn writeMDLFloatParameter label val =
(
	if (val.type == #Static) then \
		writeMDLValue ("static " + label) val.data
	else \
		writeMDLAnimation label val.data floatToString
)

fn writeMDLColorParameter label val =
(
	if (val.type == #Static) then \
		writeMDLValue ("static " + label) (ColorToString2 val.data)
	else \
		writeMDLAnimation label val.data ColorToString
)

fn writeMDLTransformations trans =
(
	if trans.translation != undefined then \
		writeMDLAnimation "Translation" trans.translation PositionToString
	if trans.rotation != undefined then \
		writeMDLAnimation "Rotation" trans.rotation QuatToString
	if trans.scale != undefined then \
		writeMDLAnimation "Scaling" trans.scale ScaleToString 
)

fn writeMDLHeader model =
(
	writeMDLSection "Version" "" 
		writeMDLValue "FormatVersion" model.version 
	writeMDLSectionEnd () 
	writeMDLSection "Model" (StringWithColomms model.name) 
		if (model.geosets.count != 0) then writeMDLValue "NumGeosets" model.geosets.count 
		if (model.geosetAnims.count != 0) then writeMDLValue "NumGeosetAnims" model.geosetAnims.count 
		if (model.helpers.count != 0) then writeMDLValue "NumHelpers" model.helpers.count 
		if (model.lights.count != 0) then writeMDLValue "NumLights" model.lights.count 
		if (model.bones.count != 0) then writeMDLValue "NumBones" model.bones.count 
		if (model.attachments.count != 0) then writeMDLValue "NumAttachments" model.attachments.count 
		if (model.particleEmitters1.count != 0) then writeMDLValue "NumParticleEmitters" model.particleEmitters1.count 
		if (model.particleEmitters2.count != 0) then writeMDLValue "NumParticleEmitters2" model.particleEmitters2.count 
		if (model.ribbonEmitters.count != 0) then writeMDLValue "NumRibbonEmitters" model.ribbonEmitters.count 
		if (model.CollisionShapes.count != 0) then writeMDLValue "NumEvents" model.CollisionShapes.count 
		writeMDLValue "BlendTime" 150 
		if checkExtents model.extents then writeMDLExtents model.extents 
	writeMDLSectionEnd () 
)

fn writeMDLSequences seq =
(
	writeMDLSection "Sequences" seq.count 
		for s in seq do
		(
			writeMDLSection "Anim" (StringWithColomms s.name) 
				writeMDLValue "Interval" (PairToString (F2WS s.startTime) (F2WS s.endTime) ) 
				if s.loop then writeMDLFlag "NonLooping" 
				if s.movespeed != undefined then writeMDLValue "Rarity" s.movespeed 
				if s.rarity != undefined then writeMDLValue "MoveSpeed" s.rarity 
				if checkExtents s.extents then writeMDLExtents s.extents 
			writeMDLSectionEnd () 
		)
	writeMDLSectionEnd () 
)

fn writeMDLGlobalSequences seq =
(
	if (seq.count > 0) then
	(
		writeMDLSection "GlobalSequences" seq.count 
			for s in seq do writeMDLValue "Duration" (F2WS s) 
		writeMDLSectionEnd () 
	)
)

fn writeMDLTextures tex =
(
	if (tex.count > 0) then
	(
		writeMDLSection "Textures" tex.count 
			for t in tex do
			(
				writeMDLSection "Bitmap" "" 
					if t.replaceableId ==0 then \
						writeMDLValue "Image" (StringWithColomms t.path) 
					else
					(
						writeMDLValue "Image" (StringWithColomms "") 
						writeMDLValue "ReplaceableId" t.replaceableId 
					)
					if t.wrapWidth then writeMDLFlag "WrapWidth" 
					if t.wrapHeight then writeMDLFlag "WrapHeight" 
				writeMDLSectionEnd () 
			)
		writeMDLSectionEnd ()
	)
)

fn writeMDLMaterials materials =
(
	if (materials.count > 0) then
	(
		writeMDLSection "Materials" materials.count 
		for m in materials do
		(
			writeMDLSection "Material" "" 
			for l in m.layers do
			(
				writeMDLSection "Layer" "" 
					writeMDLValue "FilterMode" (FilterModeToString l.filtermode) 
					if l.unshaded then writeMDLFlag "Unshaded" 
					if l.twosided then writeMDLFlag "TwoSided" 
					if l.unfogged then writeMDLFlag "Unfogged" 
					if l.nodephtest then writeMDLFlag "NoDepthTest" 
					if l.nodephset then writeMDLFlag "NoDepthSet" 
					writeMDLFloatParameter "TextureID" l.textures
					if l.coordId != 0 then writeMDLValue "CoordId" l.coordId
					if l.tvertexid != undefined then writeMDLValue "TVertexAnimId" (l.tvertexid.id) 
					writeMDLFloatParameter "Alpha" (MaxMaterialAlphaToWc3Alpha l.alpha)
				writeMDLSectionEnd () 
			)
			writeMDLSectionEnd () 
		)
		writeMDLSectionEnd ()
	)
)

fn writeMDLTextureAnims tvanims =
(
	if (tvanims.count != 0) then
	(
		writeMDLSection "TextureAnims" tvanims.count 
		for tv in tvanims do
		(
			writeMDLSection "TVertexAnim" "" 
				writeMDLTransformations tv.transformation 
			writeMDLSectionEnd () 
		)
		writeMDLSectionEnd () 
	)
)

fn writeMDLGeosets meshes =
(
	if (meshes.count != 0) then
	(
		for m in meshes do
		(
			writeMDLComment (StringWithColomms m.mesh.name) 
			writeMDLSection "Geoset" "" 
				writeMDLSection "Vertices" m.vertices.count 
					for v in m.vertices do
						writeMDLVertex v
				writeMDLSectionEnd () 
				writeMDLSection "Normals" m.normals.count 
					for v in m.normals do
						writeMDLVertex v
				writeMDLSectionEnd () 
				for g in m.tvertices do
				(
					writeMDLSection "TVertices" g.count 
						for v in g do
							writeMDLTVertex v
					writeMDLSectionEnd ()
				)
				writeMDLSection "VertexGroup" "" 
					for v in m.matrixindices do
						writeMDLSingleValue v
				writeMDLSectionEnd () 

					writeMDLSection "Faces 1" (m.faces.count * 3)
						writeMDLSection "Triangles" ""
							if (not gmax) then
								writeMDLFlag (TrianglesToString m.faces)
							else
								writeMDLTriangles (m.faces)
						writeMDLSectionEnd ()
				writeMDLSectionEnd () 
				local tmp = ArrayFold m.matrices (fn tmp2 a b = a + b.count) 0
				writeMDLSection "Groups" ((m.matrices.count as string) + " " + (tmp as string)) 
					for g in m.matrices do
						writeMDLValue "Matrices" (ArrayToString g (fn tmp3 a = (a as string))) 
				writeMDLSectionEnd () 
				writeMDLExtents m.mainExtents 
				if (m.extents.count != 0) then
				(
					for n in m.extents do
					(
						writeMDLSection "Anim" "" 
							writeMDLExtents n 
						writeMDLSectionEnd () 
					)
				)
				if m.material != undefined then writeMDLValue "MaterialID" m.material.id else writeMDLValue "MaterialID" 0 
				writeMDLValue "SelectionGroup" 0 
				if m.unselectable then
					writeMDLFlag "Unselectable"
			writeMDLSectionEnd () 
		)
	)
)

fn writeMDLGeosetAnims geoAnims =
(
	if (geoAnims.count != 0) then
	(
		for g in geoAnims do
		(
			writeMDLSection "GeosetAnim" "" 
				if g.dropShadow != undefined then writeMDLFlag "DropShadow" 
				if g.alpha != undefined then writeMDLFloatParameter "Alpha" g.alpha 
				if g.color != undefined then writeMDLColorParameter "Color" g.color 
				writeMDLValue "GeosetId" g.geoset.id 
			writeMDLSectionEnd () 
		)
	)
)

fn writeMDLMDLObj obj =
(
	writeMDLValue "ObjectId" obj.objid
	if obj.parent != undefined then writeMDLValue "Parent" obj.parent.objid
	if obj.billboardType != #None then writeMDLFlag (BillboardToString obj.billboardType)
	if obj.cameraAnchored then writeMDLFlag "CameraAnchored"
	local tmp = #()
	if obj.dontInheriteTranslation then append tmp "Translation"
	if obj.dontInheriteRotation then append tmp "Rotation"
	if obj.dontInheriteScale then append tmp "Scaling"
	if tmp.count != 0 then writeMDLValue "DontInherit" (ArrayToString tmp (fn tmp2 a = a))
)

fn writeMDLBones bones =
(
	if (bones.count != 0) then
	(
		for b in bones do
		(
			writeMDLSection "Bone" (StringWithColomms b.name) 
				writeMDLMDLObj b 
				if (b.sub.geosets.count == 1) then
				(
					local g = b.sub.geosets[1]
					writeMDLValue "GeosetId" g.id 
					if g.geosetAnimId != undefined then 
						writeMDLValue "GeosetAnimId" g.geosetAnimId 
					else
						writeMDLValue "GeosetAnimId" "None" 
				)
				else
				(
					writeMDLValue "GeosetId" "Multiple" 
					writeMDLValue "GeosetAnimId" "None" 
				)
				writeMDLTransformations b.transformations 
			writeMDLSectionEnd () 
		)
	)
)

fn writeMDLLights lights =
(
	if (lights.count != 0) then
	(
		for l in lights do
		(
			writeMDLSection "Light" (StringWithColomms l.name) 
				writeMDLMDLObj l 
				local sub = l.sub
				case (sub.lightType) of
				(
					#Omnidirectional: writeMDLFlag "Omnidirectional" 
					#Directional: writeMDLFlag "Directional" 
					default: writeMDLFlag "Ambient" 
				)
				writeMDLFloatParameter "AttenuationStart" sub.attenuationStart 
				writeMDLFloatParameter "AttenuationEnd" sub.attenuationEnd 
				writeMDLFloatParameter "Intensity" sub.intensity 
				writeMDLColorParameter "Color" sub.color
				if (sub.ambIntensity != undefined) then writeMDLFloatParameter "AmbIntensity" sub.ambIntensity 
				if (sub.ambColor != undefined) then writeMDLColorParameter "AmbColor" sub.ambColor 
				if sub.visibility != undefined then writeMDLFloatParameter "Visibility" sub.visibility 
				writeMDLTransformations l.transformations 
			writeMDLSectionEnd () 
		)
	)
)

fn writeMDLHelpers helpers =
(
	if (helpers.count != 0) then
	(
		for h in helpers do
		(
			writeMDLSection "Helper" (StringWithColomms h.name) 
				writeMDLMDLObj h 
				writeMDLTransformations h.transformations 
			writeMDLSectionEnd () 
		)
	)
)

fn writeMDLAttachments attachments =
(
	if (attachments.count != 0) then
	(
		for a in attachments do
		(
			local sub = a.sub
			writeMDLSection "Attachment" (StringWithColomms sub.name )
				writeMDLMDLObj a
				if sub.attachmentId != undefined then writeMDLValue "AttachmentID" sub.attachmentId
				if sub.outPath != undefined then writeMDLValue "Path" (StringWithColomms sub.outPath)
				writeMDLTransformations a.transformations
				if sub.visibility != undefined then writeMDLFloatParameter "Visibility" sub.visibility
			writeMDLSectionEnd () 
		)
	)
)

fn writeMDLPivotPoints model =
(
	local pv = model.bones.count + \
		model.lights.count + \
		model.helpers.count + \
		model.attachments.count + \
		model.particleEmitters1.count + \
		model.particleEmitters2.count + \
		model.ribbonEmitters.count + \
		model.events.count + \
		model.CollisionShapes.count
	if pv != 0 then
	(
		writeMDLSection "PivotPoints" pv 
			for o in model.bones do writeMDLPoint (PositionToString o.pivot) 
			for o in model.lights do writeMDLPoint (PositionToString o.pivot) 
			for o in model.helpers do writeMDLPoint (PositionToString o.pivot) 
			for o in model.attachments do writeMDLPoint (PositionToString o.pivot) 
			for o in model.particleEmitters1 do writeMDLPoint (PositionToString o.pivot) 
			for o in model.particleEmitters2 do writeMDLPoint (PositionToString o.pivot) 
			for o in model.ribbonEmitters do writeMDLPoint (PositionToString o.pivot) 
			for o in model.events do writeMDLPoint (PositionToString o.pivot) 
			for o in model.CollisionShapes do writeMDLPoint (PositionToString o.pivot) 
		writeMDLSectionEnd () 
	)
)

fn writeMDLParticleEmitters particleEmitters =
(
	if (particleEmitters.count != 0) then
	(
		for p in particleEmitters do
		(
			writeMDLSection "ParticleEmitter" (StringWithColomms p.name)
				writeMDLMDLObj p
				local sub = p.sub
				writterTransformKeys sub.longitude degToRad
				writterTransformKeys sub.latitude degToRad
				if sub.lifespan.type == #Static then
					writterTransformKeys sub.lifespan MxLife2W3Life
				else
					writterTransformKeys sub.lifespan (fn tmp3 a = a/framerate)
				writterTransformKeys sub.gravity (fn tmp a = a*100.0)
				writterTransformKeys sub.initVelocity (fn tmp2 a = a*10.0)
				writeMDLFloatParameter "EmissionRate" sub.emissionRate
				writeMDLFloatParameter "Gravity" sub.gravity
				writeMDLFloatParameter "Longitude" sub.longitude
				writeMDLFloatParameter "Latitude" sub.latitude
				if sub.visibility != undefined then writeMDLFloatParameter "Visibility" sub.visibility 
				writeMDLSection "Particle" ""
					writeMDLFloatParameter "LifeSpan" sub.lifespan
					writeMDLFloatParameter "InitVelocity" sub.initVelocity
					writeMDLValue "Path" (StringWithColomms sub.path)
				writeMDLSectionEnd ()
				writeMDLTransformations p.transformations
			writeMDLSectionEnd ()
		)
	)
)

fn writeMDLParticleEmitters2 particleEmitters2 =
(
	if (particleEmitters2.count != 0) then
	(
		for p in particleEmitters2 do
		(
			writeMDLSection "ParticleEmitter2" (StringWithColomms p.name) 
				writeMDLMDLObj p 
				local sub = p.sub
				if sub.sortPrimesFarZ then writeMDLFlag "SortPrimsFarZ" 
				if sub.unshaded then writeMDLFlag "Unshaded" 
				if sub.lineEmitter then writeMDLFlag "LineEmitter" 
				if sub.unfogged then writeMDLFlag "Unfogged" 
				if sub.modelSpace then writeMDLFlag "ModelSpace" 
				if sub.xyQuad then writeMDLFlag "XYQuad" 
				writeMDLFloatParameter "Speed" sub.speed 
				writeMDLFloatParameter "Variation" sub.variation 
				writeMDLFloatParameter "Latitude" sub.latitude 
				writeMDLFloatParameter "Gravity" sub.gravity
				if sub.visibility != undefined then writeMDLFloatParameter "Visibility" sub.visibility
				if sub.squirt then writeMDLFlag "Squirt"
				writeMDLValue "LifeSpan" sub.lifeSpan
				writeMDLFloatParameter "EmissionRate" sub.emissionRate
				writeMDLFloatParameter "Width" sub.width
				writeMDLFloatParameter "Length" sub.length
				writeMDLFlag (ParticleFiltermodeToString sub.filtermode)
				writeMDLValue "Rows" sub.rows
				writeMDLValue "Columns" sub.columns
				writeMDLFlag (ParticleTypeToString sub.particleType)
				writeMDLValue "TailLength" sub.tailLength
				writeMDLValue "Time" sub.time
				writeMDLSection "SegmentColor" ""
					writeMDLValue "Color" (ColorToString sub.startColor)
					writeMDLValue "Color" (ColorToString sub.midColor)
					writeMDLValue "Color" (ColorToString sub.endColor)
				-- Fix for magos
				CloseIOLevel ()
				writeMDLFlag "}"
				-- writeMDLSectionEnd ()
				writeMDLValue "Alpha" (IntegerPoint3ToString sub.alpha)
				writeMDLValue "ParticleScaling" (Point3ToString sub.scaling) 
				writeMDLValue "LifeSpanUVAnim" (IntegerPoint3ToString sub.lifeSpanUVAnim) 
				writeMDLValue "DecayUVAnim" (IntegerPoint3ToString sub.decayUVAnim)
				writeMDLValue "TailUVAnim" (IntegerPoint3ToString sub.TailUVAnim)
				writeMDLValue "TailDecayUVAnim" (IntegerPoint3ToString sub.TailDecayUVAnim) 
				writeMDLValue "TextureID" sub.texture.id 
				if (sub.replaceableId != 1) then writeMDLValue "ReplaceableId" (sub.replaceableId - 1)
				if (sub.priorityPlane != 0) then writeMDLValue "PriorityPlane" sub.priorityPlane
				writeMDLTransformations p.transformations
			writeMDLSectionEnd ()
		)
	)
)

fn writeMDLRibbonEmitters ribbonEmitters =
(
	if (ribbonEmitters.count != 0) then
	(
		for r in ribbonEmitters do
		(
			writeMDLSection "RibbonEmitter" (StringWithColomms r.name)
				writeMDLMDLObj r
				local sub = r.sub
				writeMDLFloatParameter "HeightAbove" sub.heightAbove
				writeMDLFloatParameter "HeightBelow" sub.heightBelow
				writeMDLFloatParameter "Alpha" sub.alpha
				writeMDLColorParameter "Color" sub.color
				writterTransformKeys sub.textureSlot (fn tmp a = (a as integer))
				if sub.textureSlot.type != #Static then
					sub.textureSlot.data.type = #NonInterp
				writeMDLFloatParameter "TextureSlot" sub.textureSlot
				if sub.visibility != undefined then writeMDLFloatParameter "Visibility" sub.visibility 
				writeMDLValue "EmissionRate" sub.emissionRate
				writeMDLValue "LifeSpan" sub.lifeSpan
				writeMDLValue "Gravity" sub.gravity
				writeMDLValue "Rows" sub.rows
				writeMDLValue "Columns" sub.columns
				if sub.material != undefined then writeMDLValue "MaterialID" sub.material.id else writeMDLValue "MaterialID" 0 
				writeMDLTransformations r.transformations
			writeMDLSectionEnd ()
		)
	)
)

fn writeMDLCameras cameras =
(
	if (cameras.count != 0) then
	(
		for c in cameras do
		(
			writeMDLSection "Camera" (StringWithColomms c.name)
				writeMDLValue "Position" (PositionToString c.pivot)
				writeMDLTransformations c.transformations
				writeMDLValue "FieldOfView" (degToRad c.fieldOfView)
				writeMDLValue "FarClip" c.farClip
				writeMDLValue "NearClip" c.nearClip
				writeMDLSection "Target" ""
					writeMDLValue "Position" (PositionToString c.targetPos)
					writeMDLTransformations c.targetTransformations
				writeMDLSectionEnd ()
			writeMDLSectionEnd ()
		)
	)
)

fn writeMDLEvents events =
(
	if (events.count != 0) then
	(
		for o in events do
		(
			local tmp = substring o.name 5 o.name.count
			writeMDLSection "EventObject" (StringWithColomms tmp)
				writeMDLMDLObj o
				if (o.sub.eventTrack != undefined and o.sub.eventTrack.count != 0) then
				(
					writeMDLSection "EventTrack" o.sub.eventTrack.count
						for n in o.sub.eventTrack do
							writeMDLFlag ((F2WS n) as string)
					writeMDLSectionEnd ()
				)
				writeMDLTransformations o.transformations
			writeMDLSectionEnd ()
		)
	)
)

fn writeMDLCollisionShapes shapes =
(
	if (shapes.count != 0) then
	(
		for s in shapes do
		(
			writeMDLSection "CollisionShape" (StringWithColomms s.name)
				writeMDLMDLObj s
				local sub = s.sub
				case (sub.shape) of
				(
					#Box: writeMDLFlag "Box"
					#Sphere: writeMDLFlag "Sphere"
					default: writeMDLFlag "Box"
				)
				writeMDLSection "Vertices" sub.vertices.count
					for v in sub.vertices do
						writeMDLFlag (PositionToString v)
				writeMDLSectionEnd ()
				if sub.boundsRadius != undefined then writeMDLValue "BoundsRadius" sub.boundsRadius 
				writeMDLTransformations s.transformations
			writeMDLSectionEnd ()
		)
	)
)

fn writeMDLFile model =
(
	progressDownLevel 100 19
	resetLevel ()
	progressShowItem "Writting Header"
	writeMDLHeader model
	progressShowItem "Writting Animations"
	writeMDLSequences model.sequences
	progressShowItem "Writting Global Sequences"
	writeMDLGlobalSequences model.globalSequences
	progressShowItem "Writting Textures"
	writeMDLTextures model.textures
	progressShowItem "Writting Materials"
	writeMDLMaterials model.materials
	progressShowItem "Writting TVertex Animations"
	writeMDLTextureAnims model.tvertexAnims
	progressShowItem "Writting Geosets"
	writeMDLGeosets model.geosets
	progressShowItem "Writting Geoset Animations"
	writeMDLGeosetAnims model.geosetAnims
	progressShowItem "Writting Bones"
	writeMDLBones model.bones
	progressShowItem "Writting Lights"
	writeMDLLights model.lights
	progressShowItem "Writting Helpers"
	writeMDLHelpers model.helpers
	progressShowItem "Writting Attachments"
	writeMDLAttachments model.attachments
	progressShowItem "Writting Pivot Points"
	writeMDLPivotPoints model
	progressShowItem "Writting Particle Emitters 1"
	writeMDLParticleEmitters model.particleEmitters1
	progressShowItem "Writting Particle Emitters 2"
	writeMDLParticleEmitters2 model.particleEmitters2
	progressShowItem "Writting Ribbon Emitters"
	writeMDLRibbonEmitters model.ribbonEmitters
	progressShowItem "Writting Cameras"
	writeMDLCameras model.cameras
	progressShowItem "Writting Events"
	writeMDLEvents model.events
	progressShowItem "Writting Colission Shapes"
	writeMDLCollisionShapes model.CollisionShapes
	progressUpLevel ()
)

fn OutputMDL model outfile_name =
(
	gc()
	foutstream = createfile outfile_name					
	SetOutStream foutstream
	writeMDLFile model
	close foutstream
	gc ()
	edit outfile_name
)

-- For Gmax use
fn OutputMDLGmax model =
(
	clearListener ()
	SetOutStream listener
	writeMDLFile model
)