
--local lastReadObj

fn readMDXAnimation &func =
(
	jumpMDXReading 4
	local animdata = AnimationData ()
	local numTracks = readMDXLong ()
	local interType = readMDXLong ()
	local globalSeq = readMDXLong ()
	case interType of
	(
		1: animdata.type = #Linear
		2: animdata.type = #Hermite
		3: animdata.type = #Bezier
		default: animdata.type = #NonInterp
	)
	if globalSeq != -1 then animdata.globalSequenceId = globalSeq
	local result = #()
	result[numTracks] = undefined
	local isSmooth = interType == 2 or interType == 3
	for i = 1 to numTracks do
	(
		local t = readMDXTime ()
		local val = func ()
		if isSmooth then
		(
			local intan = func ()
			local outtan = func ()
			result[i] = SmoothKeyData t val intan outtan
		)
		else result[i] = KeyData t val
	)
	animdata.keys = result
	animdata
)

fn readMDXAnimateableParameter param &func =
(
	param.type = #Animated
	param.data = readMDXAnimation func
)

fn readMDXParameter &func =
(
	local paramVal = ParameterValue ()
	paramVal.type = #Static
	paramVal.data = func ()
	paramVal
)

fn createEmptyWc3Model =
(
	local model = Wc3Model ()
	model.version = 800
	model.name = ""
	model.geosetAnims = #()
	model.bones = #()
	model.helpers = #()
	model.lights = #()
	model.attachments = #()
	model.particleEmitters1 = #()
	model.particleEmitters2 = #()
	model.ribbonEmitters = #()
	model.events = #()
	model.collisionShapes = #()
	model.cameras = #()
	model.tvertexAnims = #()
	model
)

-- 	MDXStartReadingChunk ()
-- 	while not (MDXHasChunkEnded ()) do
-- 	(
-- 		MDXStartChunkElem ()
-- 		
-- 	)

fn readMDXSequences =
(
	local numSeqs = (readMDXLong ()) / 132
	local result = #()
	result[numSeqs] = undefined
	for i = 1 to numSeqs do
	(
		local newAnim = Wc3Sequence ()
		newAnim.name = readMDXString 80
		newAnim.startTime = readMDXTime ()
		newAnim.endTime = readMDXTime ()
		local movespeed = readMDXFloat ()
		if movespeed != 0.0 then newAnim.movespeed = movespeed
		newAnim.loop = I2B (readMDXLong ())
		local rarity = readMDXFloat ()
		if rarity != 0.0 then newAnim.rarity = rarity
		-- we don't give a shit about the rest
		jumpMDXReading 32
		result[i] = newAnim
	)
	result
)

fn readMDXTextures =
(
	local numTextures = (readMDXLong ()) / 268
	local result = #()
	result[numTextures] = undefined
	for i = 1 to numTextures do
	(
		local tex = Wc3Texture ()
		tex.replaceableId = readMDXLong () + 1
		tex.path = readMDXString 260
		local flags = IntToFlags (readMDXLong()) 2
		tex.wrapWidth = flags[1]
		tex.wrapHeight = flags[2]
		tex.id = i
		result[i] = tex
	)
	result
)

fn readMDXLayers =
(
	local tag = readMDXLong ()
	if tag != tag_LAYS then
		throw "Expected a layer when reading material, got something unknown"
	local numLayers = readMDXLong ()
	local result = #()
	result[numLayers] = undefined
	for i = 1 to numLayers do
	(
		jumpMDXReading 4
		local l = Wc3Layer ()
		l.filtermode = (readMDXLong ()) + 1
		local flags = IntToFlags (readMDXLong()) 8
		l.unshaded = flags[1]
		l.unknown1 = flags[3]
		l.unknown2 = flags[4]
		l.twosided = flags[5]
		l.unfogged = flags[6]
		l.nodephtest = flags[7]
		l.nodephset = flags[8]
		l.textures = readMDXParameter readMDXLong
		local tvertexid = readMDXLong ()
		if tvertexid != -1 then l.tvertexid = tvertexid + 1
		l.coordId = (readMDXLong ()) + 1
		l.alpha = readMDXParameter readMDXFloat
		--Now look for possible animations
		do (
			local Anim_tag = peekMDXHeader ()
			local notDone = true
			case Anim_tag of
			(
				tag_KMTF: readMDXAnimateableParameter l.textures readMDXLong
				tag_KMTA: readMDXAnimateableParameter l.alpha readMDXFloat
				default: notDone = false
			)
		) while (notDone)
		writterTransformKeys l.alpha (fn tmp a = a*100.0)
		writterTransformKeys l.textures (fn tmp a = a + 1)
		result[i] = l
	)
	result
)

fn readMDXMaterials =
(
	local result = #()
	MDXStartReadingChunk ()
	local i = 0
	while not (MDXHasChunkEnded ()) do
	(
		MDXStartChunkElem ()
		local mat = Wc3Material ()
		jumpMDXReading 8
		mat.layers = readMDXLayers ()
		i += 1
		mat.id = i
		append result mat
	)
	result
)

fn readMDXTVertexAnimations =
(
	local result = #()
	MDXStartReadingChunk ()
	local i = 0
	while not (MDXHasChunkEnded ()) do
	(
		MDXStartChunkElem ()
		local tv = Wc3TVertexAnim ()
		tv.transformation = TransformationData ()
		do (
			local Anim_tag = peekMDXHeader ()
			local notDone = true
			case Anim_tag of
			(
				tag_KTAT: tv.transformation.translation = readMDXAnimation readMDXPosition
				tag_KTAR: tv.transformation.rotation = readMDXAnimation readMDXQuaternion
				tag_KTAS: tv.transformation.scale = readMDXAnimation readMDXScale
				default: notDone = false
			)
		) while (notDone)
		i += 1
		tv.id = i
		append result tv
	)
	result
)

fn readMDXGeosets =
(
	local result = #()
	MDXStartReadingChunk ()
	local id = 0
	while not (MDXHasChunkEnded ()) do
	(
		MDXStartChunkElem ()
		local geo = Wc3Geoset ()
		local vertices, tvertices = #(), normals, vertgroups, faces = #(), matrices = #()
		local faceTypes, faceGroups
		do (
			local Geo_tag = readMDXLong ()
			local notDone = true
			case Geo_tag of
			(
				tag_VRTX:
				(
					local num = readMDXLong ()
					vertices = readMDXArray num readMDXposition
				)
				tag_NRMS:
				(
					local num = readMDXLong ()
					normals = readMDXArray num readMDXposition
				)
				tag_PTYP:
				(
					local num = readMDXLong ()
					faceTypes = readMDXArray num readMDXLong
				)
				tag_PCNT:
				(
					local num = readMDXLong ()
					faceGroups = readMDXArray num readMDXLong
				)
				tag_PVTX:
				(
					local num = readMDXLong ()
					for i = 1 to faceGroups.count do
					(
						--assumes all are 4
						local numFaces = faceGroups[i]/3
						for j = 1 to numFaces do
						(
							local face = Point3 0 0 0
							face[1] = readMDXShort () + 1
							face[2] = readMDXShort () + 1
							face[3] = readMDXShort () + 1
							append faces face
						)
					)
				)
				tag_GNDX:
				(
					local num = readMDXLong ()
					vertgroups = readMDXArray num (fn tmp = readMDXByte() + 1)
				)
				tag_MTGC:
				(
					local num = readMDXLong ()
					matrices[num] = undefined
					for i = 1 to num do
					(
						local size = readMDXLong()
						matrices[i] = #()
						matrices[i][size] = undefined
					)
				)
				tag_MATS:
				(
					local num = readMDXLong ()
					for m in matrices do
						for i = 1 to m.count do
							m[i] = readMDXLong () + 1
				)
				default: notDone = false
			)
		) while (notDone)
		jumpMDXReading -4
		geo.vertices = vertices
		geo.normals = normals
		geo.matrixindices = vertgroups
		geo.faces = faces
		geo.matrices = matrices
		geo.material = readMDXLong () + 1
		jumpMDXReading 4
		local flags = bit.and (readMDXLong ()) 0x04
		geo.unselectable = (flags != 0)
		geo.mainExtents = readMDXExtents ()
		local numext = readMDXLong ()
		jumpMDXReading (28*numext)
		do (
			local Geo_tag = readMDXLong ()
			local notDone = true
			case Geo_tag of
			(
				tag_UVAS:
					local num = readMDXLong ()
				tag_UVBS:
				(
					local num = readMDXLong ()
					append tvertices (readMDXArray num readMDXTvertex)
				)
				default: notDone = false
			)
		) while (notDone)
		jumpMDXReading -4
		geo.tvertices = tvertices
		geo.numUVWmaps = tvertices.count
		id += 1
		geo.id = id
		append result geo
	)
	result
)

fn readMDXGeosetAnimations =
(
	local result = #()
	MDXStartReadingChunk ()
	local id = 0
	while not (MDXHasChunkEnded ()) do
	(
		MDXStartChunkElem ()
		local ga = Wc3GeosetAnim ()
		ga.alpha = readMDXParameter readMDXFloat
		local flags = IntToFlags (readMDXLong()) 2
		ga.dropShadow = flags[1]
		ga.usesColor = flags[2]
		ga.color = readMDXParameter readMDXColor
		ga.geoset = (readMDXLong ()) + 1
		--Now look for possible animations
		do (
			local Anim_tag = peekMDXHeader ()
			local notDone = true
			case Anim_tag of
			(
				tag_KGAC: readMDXAnimateableParameter ga.color readMDXColor
				tag_KGAO: readMDXAnimateableParameter ga.alpha readMDXFloat
				default: notDone = false
			)
		) while (notDone)
		append result ga
	)
	result
)

fn readMDXNode =
(
	local obj = MDLObject ()
	obj.name = readMDXString 80
	obj.objId = readMDXLong() + 1
	local parent = readMDXLong ()
	if parent != -1 then obj.parent = parent + 1
	local flags = IntToFlags (readMDXLong()) 21
	obj.billboardType = #None
	if flags[4] then obj.billboardType = #Full
	if flags[5] then obj.billboardType = #X
	if flags[6] then obj.billboardType = #Y
	if flags[7] then obj.billboardType = #Z
	obj.transformations = TransformationData ()
	do (
		local Anim_tag = peekMDXHeader ()
		local notDone = true
		case Anim_tag of
		(
			tag_KGTR: obj.transformations.translation = readMDXAnimation readMDXPosition
			tag_KGRT: obj.transformations.rotation = readMDXAnimation readMDXQuaternion
			tag_KGSC: obj.transformations.scale = readMDXAnimation readMDXScale
			default: notDone = false
		)
	) while (notDone)
	#(obj,flags)
)

fn readMDXBones =
(
	local result = #()
	MDXStartReadingChunk ()
	while not (MDXHasChunkEnded ()) do
	(
		MDXStartChunkElem specialCase:8
		local bon = Wc3Bone ()
		local tmp = readMDXNode ()
		local mdlobj = tmp[1]
		bon.mdlObj = mdlobj 
		mdlobj.sub = bon
		jumpMDXReading 8
		append result mdlobj
	)
	result
)

fn readMDXLights =
(
	local result = #()
	MDXStartReadingChunk ()
	while not (MDXHasChunkEnded ()) do
	(
		MDXStartChunkElem ()
		readMDXLong ()
		local lite = Wc3Light ()
		local tmp = readMDXNode ()
		local mdlobj = tmp[1]
		lite.mdlObj = mdlobj 
		mdlobj.sub = lite
		lite.lightType = readMDXLong ()
		case (lite.lightType) of
		(
			0: lite.lightType=#Omnidirectional
			1: lite.lightType=#Directional
			2: lite.lightType=#Ambient
			default: lite.lightType=#Omnidirectional
		)
		lite.attenuationStart = readMDXParameter readMDXFloat
		lite.attenuationEnd = readMDXParameter readMDXFloat
		lite.color = readMDXParameter readMDXColor
		lite.intensity = readMDXParameter readMDXFloat
		lite.ambColor = readMDXParameter readMDXColor
		lite.ambIntensity = readMDXParameter readMDXFloat
		lite.visibility = ParameterValue ()
		do (
			local Anim_tag = peekMDXHeader ()
			local notDone = true
			case Anim_tag of
			(
				tag_KLAV: readMDXAnimateableParameter lite.visibility readMDXFloat
				tag_KLAC: readMDXAnimateableParameter lite.color readMDXColor
				tag_KLAI: readMDXAnimateableParameter lite.intensity readMDXFloat
				tag_KLBC: readMDXAnimateableParameter lite.ambColor readMDXColor
				tag_KLBI: readMDXAnimateableParameter lite.ambIntensity readMDXFloat
				tag_KLAS: readMDXAnimateableParameter lite.attenuationStart readMDXFloat
				tag_KLAE: readMDXAnimateableParameter lite.attenuationEnd readMDXFloat
				default: notDone = false
			)
		) while (notDone)
		append result mdlobj
	)
	result
)

fn readMDXHelpers =
(
	local result = #()
	MDXStartReadingChunk ()
	while not (MDXHasChunkEnded ()) do
	(
		MDXStartChunkElem ()
		local help = Wc3Helper ()
		local tmp = readMDXNode ()
		local mdlobj = tmp[1]
		help.mdlObj = mdlobj
		mdlobj .sub = help
		append result mdlobj 
	)
	result
)

fn readMDXAttachments =
(
	local result = #()
	MDXStartReadingChunk ()
	while not (MDXHasChunkEnded ()) do
	(
		MDXStartChunkElem ()
		readMDXLong ()
		local tmp = readMDXNode ()
		local mdlobj = tmp[1]
		local att = Wc3Attachment ()
		att.mdlObj = mdlobj
		mdlobj.sub = att
		att.outPath = readMDXString 260
		att.attachmentId = readMDXLong ()
		att.visibility = ParameterValue ()
		local notDone = true
		do (
			local Anim_tag = peekMDXHeader ()
			case Anim_tag of
			(
				tag_KATV: readMDXAnimateableParameter att.visibility readMDXFloat
				default: notDone = false
			)
		) while (notDone)
		--lastReadObj = mdlobj
		append result mdlobj
	)
	result
)

fn readPivotPoints =
(
	local num = (readMDXLong ()) /12
	readMDXArray num readMDXPosition
)

fn readMDXParticleEmitter =
(
	local result = #()
	MDXStartReadingChunk ()
	while not (MDXHasChunkEnded ()) do
	(
		MDXStartChunkElem ()
		readMDXLong ()
		local part = Wc3ParticleEmitter ()
		local tmp = readMDXNode ()
		local mdlobj = tmp[1]
		local flags = tmp[2]
		part.mdlObj = mdlobj 
		mdlobj.sub = part
		part.emissionRate = readMDXParameter readMDXFloat
		part.gravity = readMDXParameter readMDXFloat
		part.longitude = readMDXParameter readMDXFloat
		part.latitude = readMDXParameter readMDXFloat
		part.path = readMDXString 260
		part.lifespan = readMDXParameter readMDXFloat
		part.initVelocity = readMDXParameter readMDXFloat
		part.visibility = ParameterValue ()
		do (
			local Anim_tag = peekMDXHeader ()
			local notDone = true
			case Anim_tag of
			(
				tag_KPEE: readMDXAnimateableParameter part.emissionRate readMDXFloat
				tag_KPEG: readMDXAnimateableParameter part.gravity readMDXFloat
				tag_KPLN: readMDXAnimateableParameter part.longitude readMDXFloat
				tag_KPEL: readMDXAnimateableParameter part.latitude readMDXFloat
				tag_KPES: readMDXAnimateableParameter part.lifespan readMDXFloat
				tag_KPEV: readMDXAnimateableParameter part.visibility readMDXFloat
				default: notDone = false
			)
		) while (notDone)
		writterTransformKeys part.longitude radToDeg
		writterTransformKeys part.latitude radToDeg
		if part.lifespan.type == #Static then
			writterTransformKeys part.lifespan W3Life2MxLife
		else
			writterTransformKeys part.lifespan (fn tmp3 a = a*framerate)
		writterTransformKeys part.gravity (fn tmp a = a/100.0)
		writterTransformKeys part.initVelocity (fn tmp2 a = a/10.0)
		--lastReadObj = mdlobj
		append result mdlobj
	)
	result
)

fn readMDXParticleEmitter2 =
(
	local result = #()
	MDXStartReadingChunk ()
	while not (MDXHasChunkEnded ()) do
	(
		MDXStartChunkElem ()
		readMDXLong ()
		local part = Wc3ParticleEmitter2 ()
		local tmp = readMDXNode ()
		local mdlobj = tmp[1]
		local flags = tmp[2]
		part.mdlObj = mdlobj
		mdlobj.sub = part
		part.unshaded =  flags[16]
		part.sortPrimesFarZ = flags[17]
		part.lineEmitter =  flags[18]
		part.unfogged = flags[19]
		part.modelSpace = flags[20]
		part.xyQuad = flags[21]
		part.speed = readMDXParameter readMDXFloat
		part.variation = readMDXParameter readMDXFloat
		part.latitude = readMDXParameter readMDXFloat
		part.gravity = readMDXParameter readMDXFloat
		part.lifeSpan = readMDXFloat ()
		part.emissionRate = readMDXParameter readMDXFloat
		part.length = readMDXParameter readMDXFloat
		part.width =readMDXParameter readMDXFloat
		part.filtermode = readMDXLong () + 1
		part.rows = readMDXLong ()
		part.columns = readMDXLong ()
		part.particleType = readMDXLong () + 1
		part.tailLength = readMDXFloat ()
		part.time = readMDXFloat ()
		part.startColor = readMDXColor ()
		part.midColor = readMDXColor ()
		part.endColor = readMDXColor ()
		local alpha = point3 0 0 0
		alpha.x = readMDXByte ()
		alpha.y = readMDXByte ()
		alpha.z = readMDXByte ()
		part.alpha = alpha
		part.scaling = readMDXVector3 ()
		part.lifeSpanUVAnim = readMDXVector3Long ()
		part.decayUVAnim = readMDXVector3Long ()
		part.TailUVAnim = readMDXVector3Long ()
		part.TailDecayUVAnim = readMDXVector3Long ()
		part.texture = readMDXLong () + 1
		part.squirt = readMDXLong ()
		if part.squirt == 1 then part.squirt = true else part.squirt = false
		part.priorityPlane = readMDXLong ()
		part.replaceableId = readMDXLong () + 1
		part.visibility = ParameterValue ()
		do (
			local Anim_tag = peekMDXHeader ()
			local notDone = true
			case Anim_tag of
			(
				tag_KP2S: readMDXAnimateableParameter part.speed readMDXFloat
				tag_KP2L: readMDXAnimateableParameter part.latitude readMDXFloat
				tag_KP2R: readMDXAnimateableParameter part.variation readMDXFloat
				tag_KP2G: readMDXAnimateableParameter part.gravity readMDXFloat
				tag_KP2E: readMDXAnimateableParameter part.emissionRate readMDXFloat
				tag_KP2N: readMDXAnimateableParameter part.length readMDXFloat
				tag_KP2W: readMDXAnimateableParameter part.width readMDXFloat
				tag_KP2V: readMDXAnimateableParameter part.visibility readMDXFloat
				default: notDone = false
			)
		) while (notDone)
		--lastReadObj = mdlobj
		append result mdlobj
	)
	result
)

fn readMDXRibbonEmitter =
(
	local result = #()
	MDXStartReadingChunk ()
	while not (MDXHasChunkEnded ()) do
	(
		MDXStartChunkElem ()
		readMDXLong ()
		local rib = Wc3RibbonEmitter ()
		local tmp = readMDXNode ()
		local mdlobj = tmp[1]
		local flags = tmp[2]
		rib.mdlObj = mdlobj
		mdlobj.sub = rib
		rib.heightAbove = readMDXParameter readMDXFloat
		rib.heightBelow = readMDXParameter readMDXFloat
		rib.alpha = readMDXParameter readMDXFloat
		rib.color = readMDXParameter readMDXColor
		rib.lifeSpan = readMDXFloat ()
		rib.textureSlot = readMDXParameter readMDXLong
		rib.emissionRate = readMDXLong ()
		rib.rows = readMDXLong ()
		rib.columns = readMDXLong ()
		rib.material = readMDXLong () + 1
		rib.gravity = readMDXFloat ()

		rib.visibility = ParameterValue ()
		do (
			local Anim_tag = peekMDXHeader ()
			local notDone = true
			case Anim_tag of
			(
				tag_KRHA: readMDXAnimateableParameter rib.heightAbove readMDXFloat
				tag_KRHB: readMDXAnimateableParameter rib.heightBelow readMDXFloat
				tag_KRAL: readMDXAnimateableParameter rib.alpha readMDXFloat
				tag_KRCO: readMDXAnimateableParameter rib.color readMDXColor
				tag_KRTX: readMDXAnimateableParameter rib.textureSlot readMDXLong
				tag_KRVS: readMDXAnimateableParameter rib.visibility readMDXFloat
				default: notDone = false
			)
		) while (notDone)
		--lastReadObj = mdlobj
		append result mdlobj
	)
	result
)

fn readMDXEventObject =
(
	fn readMDXEventTrack =
	(
		jumpMDXReading 4
		local animdata = AnimationData ()
		local numTracks = readMDXLong ()
		local globalSeq = readMDXLong ()
		animdata.type = #NonInterp
		local tracks = #()
		tracks[numTracks] = undefined
		for i = 1 to numTracks do
			tracks[i] = KeyData (readMDXTime ()) true
		animdata.keys = tracks
		MDXFixChunkSize (12 + 4*numTracks)
		animdata
	)
	local result = #()
	MDXStartReadingChunk ()
	while not (MDXHasChunkEnded ()) do
	(
		MDXStartChunkElem ()
		local tmp = readMDXNode ()
		local mdlobj = tmp[1]
		local flags = tmp[2]
		local ev = Wc3Event ()
		mdlobj.sub = ev
		ev.mdlObj = mdlobj
		do (
			local Anim_tag = peekMDXHeader ()
			local notDone = true
			case Anim_tag of
			(
				tag_KEVT: ev.eventTrack = readMDXEventTrack ()
				default: notDone = false
			)
		) while (notDone)
		append result mdlobj
	)
	result
)

fn readMDXCollisionShapes =
(
	local result = #()
	MDXStartReadingChunk ()
	while not (MDXHasChunkEnded ()) do
	(
		MDXStartChunkElem ()
		local tmp = readMDXNode ()
		local mdlobj = tmp[1]
		local col = Wc3CollisionShape ()
		col.mdlObj = mdlobj
		mdlobj.sub = col
		col.shape = readMDXLong ()
		if col.shape == 2 then
			MDXFixChunkSize 20
		else
			MDXFixChunkSize 28
		col.vertices = #()
		append col.vertices (readMDXPosition ())
		if (col.shape == 2) then
			col.boundsRadius = readMDXFloat ()
		else
			append col.vertices (readMDXPosition ())
		append result mdlobj
	)
	result
)

fn readMDXCameras =
(
	local result = #()
	MDXStartReadingChunk ()
	while not (MDXHasChunkEnded ()) do
	(
		MDXStartChunkElem ()
		local cam = Wc3Camera ()
		cam.name = readMDXString 80
		cam.pivot = readMDXPosition ()
		cam.fieldOfView = radToDeg (readMDXFloat ())
		cam.farClip = readMDXFloat ()
		cam.nearClip = readMDXFloat ()
		cam.targetPos = readMDXPosition ()
		cam.transformations = TransformationData ()
		cam.targetTransformations = TransformationData ()
		do (
			local Anim_tag = peekMDXHeader ()
			local notDone = true
			case Anim_tag of
			(
				tag_KCTR: cam.transformations.translation = readMDXAnimation readMDXPosition
				tag_KTTR: cam.targetTransformations.translation = readMDXAnimation readMDXPosition
				tag_KCRL: cam.rotation = readMDXAnimation readMDXFloat
				default: notDone = false
			)
		) while (notDone)
		append result cam
	)
	result
)

fn restorePivots model pivots =
(
	for n in model.bones do
		n.pivot = pivots[n.objId]
	for n in model.helpers do
		n.pivot = pivots[n.objId]
	for n in model.lights do
		n.pivot = pivots[n.objId]
	for n in model.attachments do
		n.pivot = pivots[n.objId]
	for n in model.particleEmitters1 do
		n.pivot = pivots[n.objId]
	for n in model.particleEmitters2 do
		n.pivot = pivots[n.objId]
	for n in model.ribbonEmitters do
		n.pivot = pivots[n.objId]
	for n in model.events do
		n.pivot = pivots[n.objId]
	for n in model.CollisionShapes do
		n.pivot = pivots[n.objId]
)

fn readMDXFile =
(
	local chunk = readMDXLong ()
	if chunk != tag_MDLX then
		return undefined
	chunk = readMDXLong ()
	local model = createEmptyWc3Model ()
	local pivots
	while (chunk != undefined) do
	(
		case chunk of
		(
			tag_VERS:
				MDXSkipChunk ()
			tag_MODL:
				MDXSkipChunk ()
			tag_SEQS:
				model.sequences = readMDXSequences ()
			tag_GLBS:
				MDXSkipChunk ()
			tag_TEXS:
				model.textures = readMDXTextures ()
			tag_MTLS:
				model.materials = readMDXMaterials ()
			tag_SNDS:
				MDXSkipChunk ()
			tag_TXAN:
				model.tvertexAnims = readMDXTVertexAnimations ()
			tag_GEOS:
				model.geosets = readMDXGeosets ()
			tag_GEOA:
				model.geosetAnims = readMDXGeosetAnimations ()
			tag_BONE:
				model.bones = readMDXBones ()
			tag_LITE:
				model.lights = readMDXLights ()
			tag_HELP:
				model.helpers = readMDXHelpers ()
			tag_ATCH:
				model.attachments = readMDXAttachments ()
			tag_PIVT:
				pivots = readPivotPoints ()
			tag_PREM:
				model.particleEmitters1 = readMDXParticleEmitter ()
			tag_PRE2:
				model.particleEmitters2 = readMDXParticleEmitter2 ()
			tag_SNEM:
				MDXSkipChunk ()
			tag_RIBB:
				model.ribbonEmitters = readMDXRibbonEmitter ()
			tag_EVTS:
				model.events = readMDXEventObject ()
			tag_CAMS:
				model.cameras = readMDXCameras ()
			tag_CLID:
				model.collisionShapes = readMDXCollisionShapes ()
			default:
			(
				if (not gmax) then
				(
					print "Debug Information:"
					print lastReadObj
					print (bit.intAsHex chunk)
					print "Tag: " + (IntAsTag chunk)
					print "Position: " + (bit.intAsHex (MDXReadPos ()))
				)
				--throw "Found An Unknown Tag"
			)
		)
		chunk = readMDXLong ()
	)
	restorePivots model pivots
	model
)

fn InputMDX infile_name =
(
	finstream = fopen infile_name "rb"
	SetInStream finstream
	--local size = posStamp ()
	model = readMDXFile ()
	--size = size - (posStamp ())
	fclose finstream
	gc()
	model
)