

struct animChunk (name, data, writeFunc)

fn writeMDXAnimateableParameter val defaultVal name &anims &writeFunc =
(
	if (val.type == #Static) then
		writeFunc val.data
	else 
	(
		local a = animChunk name:name data:val.data writeFunc:writeFunc
		writeFunc defaultVal
		append anims a
	)
)

fn writeMDXAnimation name data &writeFunc =
(
	writeMDXHead name
	local keys = data.keys
	local isSmooth = isSmoothAnim data.type
	writeMDXLong keys.count
	writeMDXLong (ControlToInt data.type)
	if data.globalSequenceId == undefined then
		writeMDXLong -1
	else
		writeMDXLong (data.globalSequenceId - 1)
	if isSmooth then
		for k in keys do
		(
			writeMDXLong (F2WI k.time)
			writeFunc k.value
			writeFunc k.intan
			writeFunc k.outtan
		)
	else
		for k in keys do
		(
			writeMDXLong (F2WI k.time)
			writeFunc k.value
		)
)

fn writeMDXAnimations anims =
(
	for a in anims do
		writeMDXAnimation a.name a.data a.writeFunc
)

-- Chunk functions

fn writeMDXVersion =
(
	local pos = InitiateChunk "VERS"
		writeMDXLong 800 -- Version
	FinishChunk pos
)

fn writeMDXHeader model =
(
	local pos = InitiateChunk "MODL"
		writeMDXString model.name 80 -- name
		writeMDXString "" 260 -- Animation File name
		writeMDXExtents model.extents
		writeMDXLong 150 -- Blend Time
	FinishChunk pos
)

fn writeMDXSequences seq =
(
	local pos = InitiateChunk "SEQS"
		for s in seq do
		(
			writeMDXString s.name 80
			writeMDXLong  (F2WI s.startTime)
			writeMDXLong  (F2WI s.endTime)
			if s.movespeed == undefined then 
				writeMDXFloat 0.0
			else
				writeMDXFloat (s.movespeed as float)
			writeMDXLong (B2I s.loop)
			if s.rarity == undefined then
				writeMDXFloat 0.0
			else
				writeMDXFloat (s.rarity as float)
			writeMDXLong 0 --Sync Point
			writeMDXExtents s.extents
		)
	FinishChunk pos
)

fn writeMDXGlobalSequences seq =
(
	if (seq.count > 0) then
	(
		local pos = InitiateChunk "GLBS"
			for s in seq do writeMDXLong (F2WI s)
		FinishChunk pos
	)
)

fn writeMDXTextures tex =
(
	fn Flags t =
	(
		b = #{}
		b[1] = t.wrapWidth
		b[2] = t.wrapHeight
		b
	)
	if (tex.count > 0) then
	(
		local pos = InitiateChunk "TEXS"
			for t in tex do
			(
				writeMDXLong t.replaceableId 
				writeMDXString t.path 260
				writeMDXFlags (Flags t)
			)
		FinishChunk pos
	)
)

fn writeMDXLayers layers =
(
	fn Flags l =
	(
		b = #{}
		b[1] = l.unshaded
		b[4] = l.unknown1
		b[8] = l.unknown2
		b[16] = l.twosided
		b[32] = l.unfogged
		b[64] = l.nodephtest
		b[128] = l.nodephset
		b
	)
	writeMDXHead "LAYS"
	writeMDXLong layers.count
	for l in layers do
	(
		local pos = InitiateChunk ""  includeName:false --inclusive size
		local animations = #()
		writeMDXLong (l.filtermode-1)
		writeMDXFlags (Flags l)
		writterTransformKeys l.alpha (fn tmp a = a/100.0)
		writeMDXAnimateableParameter l.textures 0 "KMTF" animations writeMDXLong
		if l.tvertexid != undefined then 
			writeMDXLong  l.tvertexid.id
		else
			writeMDXLong -1
		writeMDXLong l.coordId
		writeMDXAnimateableParameter l.alpha 1.0 "KMTA" animations writeMDXFloat
		writeMDXAnimations animations
		FinishChunk pos inclusive:0
	)
)

fn writeMDXMaterials materials =
(
	if (materials.count > 0) then
	(
		local pos = InitiateChunk "MTLS"
		for m in materials do
		(
			local pos2 = InitiateChunk ""  includeName:false --inclusive size
			writeMDXLong 0
			writeMDXLong 0
			writeMDXLayers m.layers
			FinishChunk pos2 inclusive:0
		)
		FinishChunk pos
	)
)

fn writeMDXTextureAnims tvanims =
(
	if (tvanims.count != 0) then
	(
		local pos = InitiateChunk "TXAN"
		for tv in tvanims do
		(
			local pos2 = InitiateChunk ""  includeName:false --inclusive size
			local trans = tv.transformation
			if trans != undefined then
			(
				if trans.translation != undefined then \
					writeMDXAnimation "KTAT" trans.translation writeMDXPosition
				if trans.rotation != undefined then \
					writeMDXAnimation "KTAR" trans.rotation writeMDXQuaternion
				if trans.scale != undefined then \
					writeMDXAnimation "KTAS" trans.scale writeMDXPoint3 
			)
			FinishChunk pos2 inclusive:0
		)
		FinishChunk pos
	)
)

fn writeMDXGeosets meshes =
(
	if (meshes.count != 0) then
	(
		local pos = InitiateChunk "GEOS"
		for m in meshes do
		(
			local pos2 = InitiateChunk ""  includeName:false --inclusive size
			-- Vertex
			writeMDXHead "VRTX"
			writeMDXLong m.vertices.count
			for v in m.vertices do
				writeMDXPosition v
			-- Normals
			writeMDXHead "NRMS"
			writeMDXLong m.normals.count
			for v in m.normals do
				writeMDXPosition v
			-- FaceGroups Type
			writeMDXHead "PTYP"
			writeMDXLong 1
			writeMDXLong 4 -- Triangles
			-- FaceGroups
			writeMDXHead "PCNT"
			writeMDXLong 1
			writeMDXLong (m.faces.count*3)
			-- Faces
			writeMDXHead "PVTX"
			writeMDXLong (m.faces.count*3)
			for f in m.faces do
				writeMDXFace f
			-- Vertex Groups
			writeMDXHead "GNDX"
			writeMDXLong m.matrixindices.count
			for v in m.matrixindices do
				writeMDXByte (v as integer)
			-- Matrices
			writeMDXHead "MTGC"
			writeMDXLong m.matrices.count
			local tmp = 0
			for g in m.matrices do
			(
				writeMDXLong g.count
				tmp += g.count
			)
			--Matrix Groups
			writeMDXHead "MATS"
			writeMDXLong tmp
			for g in m.matrices do
				for i in g do
					writeMDXLong i
			if m.material != undefined then writeMDXLong m.material.id else writeMDXLong 0
			writeMDXLong 0 --SectionGroup
			if m.unselectable then
				writeMDXLong 4
			else
				writeMDXLong 0 --Flags
			writeMDXExtents m.mainExtents
			writeMDXLong m.extents.count
			for n in m.extents do
				writeMDXExtents n
			-- Tvertex Groups
			writeMDXHead "UVAS"
			writeMDXLong m.tvertices.count
			-- Tvertices
			for tvertexGroup in m.tvertices do
			(
				writeMDXHead "UVBS"
				writeMDXLong tvertexGroup.count
				for v in tvertexGroup do
					writeMDXTVertex v
			)
			FinishChunk pos2 inclusive:0
		)
		FinishChunk pos
	)
)

fn writeMDXGeosetAnims geoAnims =
(
	fn Flags g =
	(
		b =#{}
		b[1] = g.dropShadow != undefined
		b[2] = g.color != undefined
		b
	)
	if (geoAnims.count != 0) then
	(
		local pos = InitiateChunk "GEOA"
		for g in geoAnims do
		(
			local pos2 = InitiateChunk ""  includeName:false --inclusive size
			local animations = #()
			if g.alpha != undefined then
				writeMDXAnimateableParameter g.alpha 1.0 "KGAO" animations writeMDXFloat
			else
				writeMDXFloat 1.0
			writeMDXFlags (Flags g)
			if g.color != undefined then
			(
				IOFixColor g.color
				writeMDXAnimateableParameter g.color [255,255,255] "KGAC" animations writeMDXColor
			)
			else
				writeMDXColor [255,255,255]
			writeMDXLong g.geoset.id
			writeMDXAnimations animations
			FinishChunk pos2 inclusive:0
		)
		FinishChunk pos
	)
)

fn writeMDXNode obj =
(
	fn Flags obj =
	(
		fn typeId sub =
		(
			case (sub.type) of
			(
				#Wc3Bone:256
				#Wc3Light:512
				#Wc3Attachment:2048
				#Wc3ParticleEmitter:4096
				#Wc3ParticleEmitter2:4096
				#Wc3RibbonEmitter:16384
				#Wc3Event:1024
				#Wc3CollisionShape:8192
				default:0
			)
		)
		local sub = obj.sub
		local type = typeId sub
		local b = #{}
		if type != 0 then b[type] = true
		if (sub.type == #Wc3ParticleEmitter) then
			b[32768] = (sub.emissionType == #Mdl)
		if (sub.type == #Wc3ParticleEmitter2) then
		(
			b[65536] = sub.sortPrimesFarZ
			b[32768] = sub.unshaded
			b[131072] = sub.lineEmitter
			b[262144] = sub.unfogged
			b[524288] = sub.modelSpace
			b[1048576] = sub.xyQuad
		)
		b[1] = obj.dontInheriteTranslation
		b[2] = obj.dontInheriteRotation
		b[4] = obj.dontInheriteScale
		if obj.billboardType != #None then
		(
			b[8] = obj.billboardType == #Full
			b[16] = obj.billboardType == #X
			b[32] = obj.billboardType == #Y
			b[64] = obj.billboardType == #Z
		)
		b[128] = obj.cameraAnchored
		b
	)
	local sub = obj.sub
	local pos = InitiateChunk ""  includeName:false --inclusive size
	writeMDXString obj.name 80
	writeMDXLong obj.objid
	if obj.parent != undefined then 
		writeMDXLong obj.parent.objid
	else
		writeMDXLong -1
	writeMDXFlags (Flags obj)
	--writeMDXLong 256
	local trans = obj.transformations
	if trans != undefined then
	(
		if trans.translation != undefined then \
			writeMDXAnimation "KGTR" trans.translation writeMDXPosition
		if trans.rotation != undefined then \
			writeMDXAnimation "KGRT" trans.rotation writeMDXQuaternion
		if trans.scale != undefined then \
			writeMDXAnimation "KGSC" trans.scale writeMDXPoint3 
	)
	FinishChunk pos inclusive:0
)

fn writeMDXBones bones =
(
	if (bones.count != 0) then
	(
		local pos = InitiateChunk "BONE"
		for b in bones do
		(
			writeMDXNode b
			if (b.sub.geosets.count == 1) then
			(
				local g = b.sub.geosets[1]
				writeMDXLong g.id 
				if g.geosetAnimId != undefined then 
					writeMDXLong g.geosetAnimId 
				else
					writeMDXLong -1
			)
			else
			(
				writeMDXLong -1
				writeMDXLong -1
			)
		)
		FinishChunk pos
	)
)

fn writeMDXLights lights =
(
	if (lights.count != 0) then
	(
		local pos = InitiateChunk "LITE"
		for l in lights do
		(
			local pos2 = InitiateChunk ""  includeName:false --inclusive size
			writeMDXNode l
			local sub = l.sub
			local animations = #()
			case (sub.lightType) of
			(
				#Omnidirectional: writeMDXLong 0
				#Directional: writeMDXLong 1
				default: writeMDXLong 2
			)
			writeMDXAnimateableParameter sub.attenuationStart 0.0 "KLAS" animations writeMDXFloat
			writeMDXAnimateableParameter sub.attenuationEnd 0.0 "KLAE" animations writeMDXFloat
			IOFixColor sub.color
			writeMDXAnimateableParameter sub.color [255,255,255] "KLAC" animations writeMDXColor
			writeMDXAnimateableParameter sub.intensity 0.0 "KLAI" animations writeMDXFloat
			writeMDXAnimateableParameter sub.ambIntensity 0.0 "KLBI" animations writeMDXFloat
			IOFixColor sub.ambColor
			writeMDXAnimateableParameter sub.ambColor [255,255,255] "KLBC" animations writeMDXColor
			if sub.visibility != undefined then
				writeMDXAnimation "KLAV" sub.visibility.data writeMDXFloat
			writeMDXAnimations animations
			FinishChunk pos2 inclusive:0
		)
		FinishChunk pos
	)
)

fn writeMDXHelpers helpers =
(
	if (helpers.count != 0) then
	(
		local pos = InitiateChunk "HELP"
		for h in helpers do
			writeMDXNode h
		FinishChunk pos
	)
)

fn writeMDXAttachments attachments =
(
	if (attachments.count != 0) then
	(
		local pos = InitiateChunk "ATCH"
		local id = 0
		for a in attachments do
		(
			local pos2 = InitiateChunk ""  includeName:false --inclusive size
			local sub = a.sub
			a.name = sub.name
			writeMDXNode a
			local path = ""
			if sub.outPath != undefined then path = sub.outPath
			writeMDXString path 260
			writeMDXLong id
			id += 1
			if sub.visibility != undefined then
				writeMDXAnimation "KATV" sub.visibility.data writeMDXFloat
			FinishChunk pos2 inclusive:0
		)
		FinishChunk pos
	)
)

fn writeMDXPivotPoints model =
(
	local pos = InitiateChunk "PIVT"
	for o in model.bones do writeMDXPosition o.pivot 
	for o in model.lights do writeMDXPosition o.pivot 
	for o in model.helpers do writeMDXPosition o.pivot 
	for o in model.attachments do writeMDXPosition o.pivot 
	for o in model.particleEmitters1 do writeMDXPosition o.pivot 
	for o in model.particleEmitters2 do writeMDXPosition o.pivot 
	for o in model.ribbonEmitters do writeMDXPosition o.pivot 
	for o in model.events do writeMDXPosition o.pivot 
	for o in model.CollisionShapes do writeMDXPosition o.pivot 
	FinishChunk pos
)

fn writeMDXParticleEmitters particleEmitters =
(
	if (particleEmitters.count != 0) then
	(
		local pos = InitiateChunk "PREM"
		for p in particleEmitters do
		(
			local pos2 = InitiateChunk ""  includeName:false --inclusive size
			writeMDXNode p
			local sub = p.sub
			local animations = #()
			writterTransformKeys sub.longitude degToRad
			writterTransformKeys sub.latitude degToRad
			if sub.lifespan.type == #Static then
				writterTransformKeys sub.lifespan MxLife2W3Life
			else
				writterTransformKeys sub.lifespan (fn tmp3 a = a/framerate)
			writterTransformKeys sub.gravity (fn tmp a = a*100.0)
			writterTransformKeys sub.initVelocity (fn tmp2 a = a*10.0)
			writeMDXAnimateableParameter sub.emissionRate 0.0 "KPEE" animations writeMDXFloat
			writeMDXAnimateableParameter sub.gravity 0.0 "KPEG" animations writeMDXFloat
			writeMDXAnimateableParameter sub.longitude 0.0 "KPLN" animations writeMDXFloat
			writeMDXAnimateableParameter sub.latitude 0.0 "KPLT" animations writeMDXFloat
			writeMDXString sub.path 260
			writeMDXAnimateableParameter sub.lifespan 0.0 "KPEL" animations writeMDXFloat
			writeMDXAnimateableParameter sub.initVelocity 0.0 "KPES" animations writeMDXFloat
			if sub.visibility != undefined then
				writeMDXAnimation "KPEV" sub.visibility.data writeMDXFloat
			writeMDXAnimations animations
			FinishChunk pos2 inclusive:0
		)
		FinishChunk pos
	)
)

fn writeMDXParticleEmitters2 particleEmitters2 =
(
	if (particleEmitters2.count != 0) then
	(
		local pos = InitiateChunk "PRE2"
		for p2 in particleEmitters2 do
		(
			local pos2 = InitiateChunk ""  includeName:false --inclusive size
			writeMDXNode p2
			local sub = p2.sub
			local animations = #()
			writeMDXAnimateableParameter sub.speed  0.0 "KP2S" animations writeMDXFloat
			writeMDXAnimateableParameter sub.variation 0.0 "KP2R" animations writeMDXFloat
			writeMDXAnimateableParameter sub.latitude 0.0 "KP2L" animations writeMDXFloat
			writeMDXAnimateableParameter sub.gravity 0.0 "KP2G" animations writeMDXFloat
			
			writeMDXFloat sub.lifeSpan
			writeMDXAnimateableParameter sub.emissionRate 0.0 "KP2E" animations writeMDXFloat
			writeMDXAnimateableParameter sub.length 0.0 "KP2N" animations writeMDXFloat
			writeMDXAnimateableParameter sub.width 0.0 "KP2W" animations writeMDXFloat
			
			writeMDXLong (sub.filtermode-1)
			writeMDXLong sub.rows
			writeMDXLong sub.columns
			writeMDXLong (sub.particleType - 1)
			writeMDXFloat sub.tailLength
			writeMDXFloat sub.time
			
			writeMDXColor (sub.startColor as point3)
			writeMDXColor (sub.midColor as point3)
			writeMDXColor (sub.endColor as point3)
			writeMDXByte sub.alpha.x
			writeMDXByte sub.alpha.y
			writeMDXByte sub.alpha.z
			writeMDXFloat sub.scaling.x
			writeMDXFloat sub.scaling.y
			writeMDXFloat sub.scaling.z
			writeMDXPoint3Long sub.lifeSpanUVAnim
			writeMDXPoint3Long sub.decayUVAnim
			writeMDXPoint3Long sub.TailUVAnim
			writeMDXPoint3Long sub.TailDecayUVAnim
			
			writeMDXLong sub.texture.id 
			if sub.squirt then 
				writeMDXLong 1 
			else 
				writeMDXLong 0
			writeMDXLong sub.priorityPlane
			writeMDXLong (sub.replaceableId-1)
			
			if sub.visibility != undefined then
				writeMDXAnimation "KP2V" sub.visibility.data writeMDXFloat
			writeMDXAnimations animations
			FinishChunk pos2 inclusive:0
		)
		FinishChunk pos
	)
)


fn writeMDXRibbonEmitters ribbonEmitters =
(
	if (ribbonEmitters.count != 0) then
	(
		local pos = InitiateChunk "RIBB"
		for r in ribbonEmitters do
		(
			local pos2 = InitiateChunk ""  includeName:false --inclusive size
			writeMDXNode r
			local sub = r.sub
			local animations = #()
			writeMDXAnimateableParameter sub.heightAbove 0.0 "KRHA" animations writeMDXFloat
			writeMDXAnimateableParameter sub.heightBelow 0.0 "KRHB" animations writeMDXFloat
			writeMDXAnimateableParameter sub.alpha 0.0 "KRAL" animations writeMDXFloat
			IOFixColor sub.color
			writeMDXAnimateableParameter sub.color [255,255,255] "KRCO" animations writeMDXColor
			writeMDXFloat sub.lifeSpan
			writterTransformKeys sub.textureSlot (fn tmp a = (a as integer))
			if sub.textureSlot.type != #Static then
				sub.textureSlot.data.type = #NonInterp
			writeMDXAnimateableParameter sub.textureSlot 0 "KRTX" animations writeMDXLong
			writeMDXLong sub.emissionRate
			writeMDXLong sub.rows
			writeMDXLong sub.columns
			if sub.material != undefined then 
				writeMDXLong sub.material.id 
			else writeMDXLong 0
			writeMDXFloat sub.gravity
			if sub.visibility != undefined then
				writeMDXAnimation "KRVS" sub.visibility.data writeMDXFloat
			writeMDXAnimations animations
			FinishChunk pos2 inclusive:0
		)
		FinishChunk pos
	)
)

fn writeMDXEvents events =
(
	if (events.count != 0) then
	(
		local pos = InitiateChunk "EVTS"
		for o in events do
		(
			o.name = substring o.name 5 o.name.count
			writeMDXNode o
			local track = o.sub.eventTrack
			if (track != undefined and track.count != 0) then
			(
				writeMDXHead "KEVT"
				writeMDXLong track.count
				writeMDXLong -1
				for t in track do
					writeMDXLong (F2WI t)
			)
		)
		FinishChunk pos
	)
)

fn writeMDXCameras cameras =
(
	if (cameras.count != 0) then
	(
		
		local pos = InitiateChunk "CAMS"
		for c in cameras do
		(
			local pos2 = InitiateChunk ""  includeName:false --inclusive size
			writeMDXString c.name 80
			writeMDXPosition c.pivot
			writeMDXFloat (degToRad c.fieldOfView)
			writeMDXFloat c.farClip
			writeMDXFloat c.nearClip
			writeMDXPosition c.targetPos
			local trans = c.transformations
			if trans != undefined then
				if trans.translation != undefined then 
					writeMDXAnimation "KCTR" trans.translation writeMDXPosition
			trans = c.targetTransformations
			if trans != undefined then
				if trans.translation != undefined then 
					writeMDXAnimation "KTTR" trans.translation writeMDXPosition
			FinishChunk pos2 inclusive:0
		)
		FinishChunk pos
	)
)

fn writeMDXCollisionShapes shapes =
(
	if (shapes.count != 0) then
	(
		local pos = InitiateChunk "CLID"
		for s in shapes do
		(
			writeMDXNode s
			local sub = s.sub
			case (sub.shape) of
			(
				#Box: writeMDXLong 0
				#Sphere: writeMDXLong 2
				default: writeMDXLong 1
			)
			for v in sub.vertices do
				writeMDXPosition v
			if sub.boundsRadius != undefined then
				writeMDXFloat sub.boundsRadius
		)
		FinishChunk pos
	)
)

fn writeMDXFile model =
(
	progressDownLevel 100 20
	writeMDXHead "MDLX" -- Initial flag
	progressShowItem "Writting Version"
	writeMDXVersion ()
	progressShowItem "Writting Header"
	writeMDXHeader model
	progressShowItem "Writting Animations"
	writeMDXSequences model.sequences
	progressShowItem "Writting Global Sequences"
	writeMDXGlobalSequences model.globalSequences
	progressShowItem "Writting Materials"
	writeMDXMaterials model.materials
	progressShowItem "Writting Textures"
	writeMDXTextures model.textures
	progressShowItem "Writting TVertex Animations"
	writeMDXTextureAnims model.tvertexAnims
	progressShowItem "Writting Geosets"
	writeMDXGeosets model.geosets
	progressShowItem "Writting Geoset Animations"
	writeMDXGeosetAnims model.geosetAnims
	progressShowItem "Writting Bones"
	writeMDXBones model.bones
	progressShowItem "Writting Lights"
	writeMDXLights model.lights
	progressShowItem "Writting Helpers"
	writeMDXHelpers model.helpers
	progressShowItem "Writting Attachments"
	writeMDXAttachments model.attachments
	progressShowItem "Writting Pivot Points"
	writeMDXPivotPoints model
	progressShowItem "Writting Particle Emitters 1"
	writeMDXParticleEmitters model.particleEmitters1
	progressShowItem "Writting Particle Emitters 2"
	writeMDXParticleEmitters2 model.particleEmitters2
	progressShowItem "Writting Ribbon Emitters"
	writeMDXRibbonEmitters model.ribbonEmitters
	progressShowItem "Writting Cameras"
	writeMDXCameras model.cameras
	progressShowItem "Writting Events"
	writeMDXEvents model.events
	progressShowItem "Writting Colission Shapes"
	writeMDXCollisionShapes model.CollisionShapes
	progressUpLevel ()
)
	
fn OutputMDX model outfile_name =
(
	gc()
	foutstream = fopen outfile_name "wb"
	SetOutStream foutstream
	local size = posStamp ()
	writeMDXFile model
	size = size - (posStamp ())
	fclose foutstream
	gc()
)