
/*
include "NeoDexModules\\Wc3Model.ms"
include "NeoDexModules\\UtilityFunctions.ms"
include "NeoDexModules\\Wc3Animation.ms"
*/
global sceneObjects

--Lists for the final model data
global Wc3GlobalSequences -- gathered elsewhere
global Wc3Textures = #()
global Wc3Materials = #()
global Wc3TvertexAnims = #()
global Wc3Sequences = #()
global Meshes = #()
global BonesMeshRef = #()
global GeosetAnims = #()
global Wc3Objects = #() -- all the objets
global Wc3Bones = #()
global Wc3Helpers = #()
global Wc3Attachments = #()
global Wc3Lights = #()
global Wc3PartEmitters1 = #()
global Wc3PartEmitters2 = #()
global Wc3RibbonEmitters = #()
global Wc3Events = #()
global Wc3Cameras = #()
global Wc3Collisions = #()
global Wc3Extents

-- Needed specificaly
global GeosetCounts = 0
global BoneCounts = 0
global ObjectIdCount = 0
global MaterialCount = 0
global TextureCount = 0
global TvertexAnimCount = 0
global GeosetAnimCount = 0

struct FakeObject (
	name,
	pos,
	parent,
	ref
)

struct FakeMesh (
	name,
	material,
	modifiers
)

struct FakeMaterial (
	texture,
	alpha = 100.0,
	filtermode = 1,
	unshaded = false,
	twosides = false,
	unfogged = false,
	unselectable = false,
	nodephtest = false,
	nodephset = false,
	unknown1 = false,
	unknown2 = false,
	retexture = 1,
	coordId = 0,
	path = "",
	UseCColor = false,
	MatColor = black
)
-- Defaults are used for undefined properties on model, like skinning or materials
global DefaultBone = FakeObject "Default Bone" [0,0,0] undefined undefined
global DefaultAnimation = (Wc3Sequence "Stand" (10 as time) (40 as time) false undefined undefined undefined)
global DefaultMaterial = (FakeMaterial ())

---------------------------------------------------------------------------------------------------------------------------------
-- 												Wc3Sequences Initial Loading											  ---
---------------------------------------------------------------------------------------------------------------------------------

fn GetSequenceData =
(
	local result = #()
	if HasNoteTracks Rootnode then
	(
		local Nt = GetNoteTrack Rootnode 1
		local keys = Nt.keys
		for k in keys do
		(
			local notFound = true
			for i = 1 to result.count where notFound do
			(
				if result[i][1].value == k.value then
				(
					append result[i] k
					notFound = false
				)
			)
			if notFound then
				append result #(k)
		)
	)
	for i = result.count to 1 by -1 where result[i].count != 2 do deleteItem result i
	result
)

fn readSequenceData key key2 =
(
	local data = filterString key.value "\"\n\t\x0D"
	local animName = data[1]
	local noLoop = false
	local rarity, movespeed
	for i = 2 to data.count do 
	(
		local c = data[i]
		if stricmp c "Nonlooping" == 0 then NoLoop = true
		if stricmp (substring c 1 6) "Rarity" == 0 then rarity = ((substring c 7 -1) as integer)
		if stricmp (substring c 1 9) "Movespeed" == 0 then movespeed = ((substring c 10 -1) as integer)
	)
	Wc3Sequence animName key.time key2.time NoLoop Rarity Movespeed undefined
)

fn LoadSequences =
(
	progressDownLevel 10 1
	progressShowItem "Loading Animations"
	local keys = GetSequenceData ()
	if keys.count > 0 then
	(
		Wc3Sequences[keys.count] = undefined
		for i = 1 to keys.count do
			Wc3Sequences[i] = readSequenceData keys[i][1] keys[i][2] 
	)
	else (
		Wc3Sequences[1] = DefaultAnimation
		reportError "No Sequences" "RootNode" undefined
	)
	progressUpLevel ()
)

---------------------------------------------------------------------------------------------------------------------------------
-- 												Objects Initial Loading												  	  ---
---------------------------------------------------------------------------------------------------------------------------------

fn IsGeosetAnimated geo = geo.visObj != undefined or (geo.mesh.modifiers[#Wc3_Vertex_Color] != undefined)

fn IsIgnorable obj = obj.isHidden --or obj.isFrozen

----------------------------------------------------------------------------------------------------------------------------
-- 										Bones Section																  --


fn ObjectChecker a b = a.maxObj == b

fn ObjectChecker2 a b = a.mdlObj.maxObj == b

fn LoadMDLObject obj =
(
	local index = ArrayFindItemEx Wc3Objects obj ObjectChecker
	if index != 0 then
	(
		Wc3Objects[index]
	)
	else
	(
		mdl = MDLObject ()
		mdl.maxObj = obj
		if classOf obj == Biped_Object then
			mdl.pivot = obj.transform.pos
		else
			mdl.pivot = obj.pos
		mdl.name = obj.name
		-- a terrible way to fix the issue, but meh no time to think more
		if classOf obj == FakeObject then obj = obj.ref
		if obj != undefined then
		(
			mdl.transformations = GetTransformations obj
			-- IK Affected bones may have troubles with inheritance
			if not (isIKAffected obj) then
			(
				local n = getInheritanceFlags obj
				if n[1] == false and n[2] == false and n[3] == false then mdl.dontInheriteTranslation = true
				if n[4] == false and n[5] == false and n[6] == false then mdl.dontInheriteRotation = true
				if n[7] == false and n[8] == false and n[9] == false then mdl.dontInheriteScale = true
			)
			mdl.billboardType = getBillboard obj
			mdl.cameraAnchored = getCameraAnchored obj
		)
		else mdl.transformations = TransformationData ()
		append Wc3Objects mdl
		mdl
	)
)

fn LoadBone obj geoset =
(
	local index = ArrayFindItemEx BonesMeshRef obj ObjectChecker2
	if index != 0 then
	(
		local index2 = findItem BonesMeshRef[index].geosets geoset
		if index2 == 0 then
			append BonesMeshRef[index].geosets geoset
	)
	else
	(
		local mdl = LoadMDLObject obj
		local mdlBone = Wc3Bone ()
		mdlBone.mdlObj = mdl
		mdl.sub = mdlBone
		append mdlBone.geosets geoset
		append BonesMeshRef mdlBone
		BoneCounts = BoneCounts + 1
		index = BoneCounts
	)
	index
)
---------------------------------------------------------------------------------------------------------------------------------
-- 												Geosets Initial Loading											            ---
---------------------------------------------------------------------------------------------------------------------------------

local exportSmoothgroups = false

fn LoadGeoset obj =
(
	fn checkWeights weights =
	(
		local minWeight = weights[1] - 0.02
		local maxWeight = weights[1] + 0.02
		local result = false
		local balance = 0.0
		for w in weights do
		(
			result = result or (minWeight > w or w > maxWeight)
			balance += w
		)
		(result or (balance < 0.98 or balance > 1.02))
	)
	fn generateFakeBone geo =
	(
		fn IsAnyAnimated o = o.pos.isAnimated or o.rotation.IsAnimated or o.scale.isanimated
		local m = geo.mesh
		if ((IsAnyAnimated m) or (IsGeosetAnimated geo)) then
		(
			local tmp = FakeObject ("Mesh " + m.name) m.pos m.parent m
			tmp
		)
		else if m.parent != undefined then m.parent
		else DefaultBone
	)
	fn getVisObject m =
	(
		local tmp = #(m)
		join tmp (GetParents m)
		local i = 1
		local count = tmp.count
		local result
		while (i <= count and result == undefined) do
		(
			if tmp[i].visibility.IsAnimated then result = tmp[i]
			i += 1
		)
		result
	)
	local geoset = Wc3Geoset ()
	local verts = #()
	local normals = #()
	local tverts = #()
	local matrixindices = #()
	local matrices = #()
	local tex = (getNumTVerts obj) > 0
	local s = undefined
	geoset.mesh = obj
	geoset.visObj = getVisObject obj
	local replaceBone = generateFakeBone geoset
	geoset.id = GeosetCounts
	GeosetCounts = GeosetCounts + 1
	append Meshes geoset
	local loadName = "Loading " + obj.name + ": "
	progressUpdateItem (loadName + "Skinning")
	-- Loading Skinning
	max modify mode
	if IsSkinned obj then
	(
		s = obj.skin
		modPanel.setCurrentObject s
	)
	local tmpIndices
	local tmpBones = #()
	local numBones
	-- load functions to cache:
	local funcBoneName = skinOps.GetBoneName
	local funcNodename = getNodeByName
	local funcBoneId = skinOps.GetVertexWeightBoneID
	local funcVertexCount = skinOps.GetVertexWeightCount
	local funcGetWeight = skinOps.GetVertexWeight
	local validSkin = s != undefined and (skinOps.GetNumberVertices s) >= obj.numverts
	-- finish loading
	if validSkin then
	(
		numBones = skinops.GetNumberBones s
		tmpBones[numBones+1] = undefined
		for i = 1 to numBones do
			tmpBones[i] = funcNodename (funcBoneName s i 0) exact:true
		numBones = numBones + 1
	) else numBones = 1
	tmpBones[numBones] = replaceBone
	if validSkin then
	(
		local weightsUnbalanced = false
		tmpIndices = #()
		tmpIndices[obj.numverts] = undefined
		for v = 1 to obj.numverts do
		(
			-- matrix finding
			local matrixindex = undefined
			local bonegroup = #()
			local count = funcVertexCount s v
			if count > 0 then
			(
				local bonenodes = #()
				local weights = #()
				local count = (funcVertexCount s v)
				bonegroup[count] = undefined
				for w = 1 to count do
				(
					--local boneId = funcBoneId s v w
					bonegroup[w] = funcBoneId s v w
					weights[w] = funcGetWeight s v w
				)
				sort bonegroup
				weightsUnbalanced = weightsUnbalanced or (checkWeights weights)
			)
			else
				bonegroup = #(numBones) --we give it the last, which is the fake bone
			local found = ArrayFindItem matrices bonegroup
			if found == 0 then
			(
				append matrices bonegroup
				matrixindex = matrices.count - 1
			)
			else matrixindex = found - 1
			tmpIndices[v] = matrixindex
		)
		if weightsUnbalanced then
			reportError "Weights" obj.name undefined
	)
	else (
		if (s != undefined) then
			reportError "Bad Skinning" obj.name undefined
		tmpBones[1] = replaceBone
		matrices = #(#(1))
		tmpIndices = ArrayRepeat obj.numverts 0
	)
	max create mode
	if suspendEditing != undefined then suspendEditing ()
	progressUpdateItem (loadName + "Related Bones")
	for m in matrices do
		for i = 1 to m.count do m[i] = LoadBone tmpBones[m[i]] geoset
	progressUpdateItem (loadName + "Smoothgroups")
	-- Loading Smoothgroups
	local faces
	local preCalculatedNormals
	local facegroupMapper
	if exportSmoothgroups then
	(
		local tmp = gatherSmoothGroupData obj
		faces = tmp[1]
		preCalculatedNormals = tmp[2]
		facegroupMapper = tmp[3]
		-- Safer in case no normals were generated
		for i = 1 to facegroupMapper.count where preCalculatedNormals [i] == undefined do
			preCalculatedNormals [i] = getNormal obj facegroupMapper[i]
	)
	else
	(
		local numVerts = getNumVerts obj
		preCalculatedNormals = #()
		preCalculatedNormals[numVerts] = undefined
		for i = 1 to numVerts do
			preCalculatedNormals[i] = getNormal obj i
	)
	progressUpdateItem (loadName + "Unwrap")
	-- Loading Unwrap
	fn canSimplifyUnwrapCalculation o =
	(
		local numberOfMaps = 0
		for i = 1 to (meshop.getNumMaps o) do
			if (meshop.getMapSupport o i) then
				numberOfMaps += 1
		numberOfMaps == 1
	)
	local hasUnwrap = (getNumTVerts obj) > 0
	if hasUnwrap then
	(
		if ((getNumTVerts obj) < obj.vertices.count) then
			reportError "Bad Unwrap" obj.name undefined
		if (canSimplifyUnwrapCalculation obj) and (not exportSmoothgroups) then
		(
			-- In this case we ignore partitions so calculations are faster.
			local tmp = breakUnwrapFast obj
			faces = tmp[2]
			local map = tmp[1]
			local vertexIndices = tmp[3]
			local count = vertexIndices.count
			verts[count] = undefined
			tverts[count] = undefined
			normals[count] = undefined
			matrixindices[count] = undefined
			for i = 1 to count do
			(
				local v = map[vertexIndices[i]]
				verts[i] = getVert obj v
				normals[i] = preCalculatedNormals[v]
				tverts[i] = getTVert obj vertexIndices[i]
				matrixindices[i] = tmpIndices[v]
			)
			geoset.tvertices = #(tverts)
		)
		else
		(
			if (exportSmoothgroups) then
			(
				local tmp = breakUnwrap obj faces:faces numVerts:facegroupMapper.count
				faces = tmp[1]
				tverts = tmp[2]
				local mapper = tmp[3]
				local count = mapper.count
				verts[count] = undefined
				normals[count] = undefined
				matrixindices[count] = undefined
				for i = 1 to count do
				(
					local v = facegroupMapper[mapper[i]]
					verts[i] = getVert obj v
					normals[i] = preCalculatedNormals[mapper[i]]
					matrixindices[i] = tmpIndices[v]
				)
			) else (
				local tmp = breakUnwrap obj
				faces = tmp[1]
				tverts = tmp[2]
				local mapper = tmp[3]
				local count = mapper.count
				verts[count] = undefined
				normals[count] = undefined
				matrixindices[count] = undefined
				for i = 1 to count do
				(
					local v = mapper[i]
					verts[i] = getVert obj v
					normals[i] = preCalculatedNormals[v]
					matrixindices[i] = tmpIndices[v]
				)
			)
			geoset.tvertices = tverts
		)
	)
	else (
		if exportSmoothgroups then
		(
			local count = facegroupMapper.count
			verts[count] = undefined
			tverts[count] = undefined
			normals[count] = undefined
			matrixindices[count] = undefined
			for i = 1 to count do
			(
				local v = facegroupMapper[i]
				verts[i] = getVert obj v
				normals[i] = preCalculatedNormals[i]
				tverts[i] = [0,0]
				matrixindices[i] = tmpIndices[v]
			)
		)
		else
		(
			faces = getFaces obj
			local count = obj.vertices.count
			verts[count] = undefined
			tverts[count] = undefined
			normals[count] = undefined
			matrixindices[count] = undefined
			for i = 1 to count do
			(
				verts[i] = getVert obj i
				normals[i] = preCalculatedNormals[i]
				tverts[i] = [0,0]
				matrixindices[i] = tmpIndices[i]
			)
		)
		geoset.tvertices = #(tverts)
		reportError "No Unwrap" obj.name undefined
	)
	-- Final Results
	geoset.vertices = verts
	geoset.matrixindices = matrixindices
	geoset.faces = faces
	geoset.normals = normals
	geoset.matrices = matrices
	geoset.unselectable = getUnselectable obj
	if resumeEditing != undefined then resumeEditing ()
	OK
)


---------------------------------------------------------------------------------------------------------------------------------
-- 												Wc3Objects Initial Loading											       ---
---------------------------------------------------------------------------------------------------------------------------------

fn LoadLight obj =
(
	local tmp = Wc3Light ()
	if ClassOf obj == Omnilight then tmp.lightType = #Omnidirectional
	else if classOf obj == freeSpot or classOf obj == Directionallight then tmp.lightType = #Directional
	else tmp.lightType = #Ambient
	tmp.attenuationStart = FloatParameterProcess obj.farAttenStart.controller obj.farAttenStart
	tmp.attenuationEnd = FloatParameterProcess obj.farAttenEnd.controller obj.farAttenEnd
	tmp.intensity = FloatParameterProcess obj.multiplier.controller obj.multiplier
	tmp.color = ColorParameterProcess obj.color.controller obj.color
	tmp.ambIntensity = FloatParameterProcess lightLevelController lightLevel
	tmp.ambColor = ColorParameterProcess ambientColorController ambientColor
	if obj.visibility.isAnimated then tmp.visibility = VisibilityParameterProcess obj
	tmp
)

fn LoadHelper obj = 
(
	local sub = Wc3Helper ()
	sub.MDLObj = LoadMDLObject obj
	sub.MDLObj.sub = sub
	sub
)

fn LoadAttachment obj =
(
	local tmp = Wc3Attachment ()
	if (obj.usesExternalModel) then
	(
		tmp.name = obj.name
		tmp.outPath = obj.externalModelPath
	)
	else (
		local s = attachmentList[obj.nameFirst]
		if obj.nameAdd1 != 1 then s = s + " " + attachmentAdd1List[obj.nameAdd1]
		if obj.nameAdd2 != 1 then s = s + " " + attachmentAdd2List[obj.nameAdd2]
		tmp.name = s + " Ref"
	)
	if (obj.usesAttachmentId) then
		tmp.attachmentId= obj.attachmentId
	if obj.visibility.isAnimated then tmp.visibility = VisibilityParameterProcess obj
	tmp
)

fn getGravityByBinding theBinding =
(
	for s in SpaceWarps do 
	(
		if classof s == gravity then for o in refs.dependents s where o == theBinding do return s
	)
	return undefined
)

fn LoadParticleEmitter obj =
(
	local tmp = Wc3ParticleEmitter ()
	local g = getGravityByBinding obj
	tmp.emissionRate = FloatParameterProcess obj.Birth_Rate.controller obj.Birth_Rate
	if g != undefined then 
		tmp.gravity = FloatParameterProcess  g.strength.controller g.strength
	else
		tmp.gravity = ParameterValue #Static 0
	tmp.longitude = FloatParameterProcess  obj.Plane_Spread.controller obj.Plane_Spread
	tmp.latitude = FloatParameterProcess  obj.Axis_Spread.controller obj.Axis_Spread
	if obj.visibility.isAnimated then tmp.visibility = VisibilityParameterProcess obj
	tmp.lifespan =  FloatParameterProcess obj.life.controller obj.life
	tmp.initVelocity = FloatParameterProcess  obj.speed.controller obj.speed
	local path = getUserProp obj "ParticlePath"
	if path == undefined then path = "" else
	(
		local h = filterString path "."
		path = h[1]
	)
	tmp.path = path + ".MDL"
	tmp
)

fn LoadParticleEmitter2 obj =
(
	local tmp = Wc3ParticleEmitter2 ()
	tmp.sortPrimesFarZ = obj.SortZ
	tmp.unshaded = obj.Unshaded
	tmp.lineEmitter = obj.LineEmitter
	tmp.unfogged = obj.Unfogged
	tmp.modelSpace = obj.ParticleSpace
	tmp.xyQuad = obj.XYQuads
	tmp.speed = FloatParameterProcess obj.speed.controller obj.speed
	tmp.variation = FloatParameterProcess obj.variation.controller obj.variation
	tmp.latitude = FloatParameterProcess obj.coneangle.controller obj.coneangle
	tmp.gravity = FloatParameterProcess obj.gravity.controller obj.gravity
	if obj.visibility.isAnimated then tmp.visibility = VisibilityParameterProcess obj
	tmp.squirt = obj.Squirt
	tmp.lifeSpan = obj.Life
	tmp.emissionRate = FloatParameterProcess obj.PartsEmit.controller obj.PartsEmit
	tmp.width = FloatParameterProcess obj.width.controller obj.width
	tmp.length = FloatParameterProcess obj.length.controller obj.length
	tmp.filtermode = obj.filtermode
	tmp.rows = obj.rows
	tmp.columns = obj.cols
	tmp.particleType = obj.particletype
	tmp.tailLength = obj.taillength
	tmp.time = obj.midtime
	tmp.startColor = (obj.startcolor  as point3)
	tmp.midColor = (obj.middlecolor  as point3)
	tmp.endColor = (obj.endcolor  as point3)
	tmp.alpha = [obj.startalpha, obj.middlealpha, obj.endalpha]
	tmp.scaling = [obj.startscale, obj.middlescale, obj.endscale]
	tmp.lifeSpanUVAnim = [obj.startLifespanHead, obj.endLifespanHead, obj.repeatLifespanHead]
	tmp.decayUVAnim = [obj.startDecayHead, obj.endDecayHead, obj.repeatDecayHead]
	tmp.TailUVAnim = [obj.startLifespanTail, obj.endLifespanTail, obj.repeatLifespanTail]
	tmp.TailDecayUVAnim = [obj.startLifespanTail, obj.endLifespanTail, obj.repeatLifespanTail]
	tmp.texture = undefined
	tmp.replaceableId = obj.rtexture
	tmp.priorityPlane = obj.priorityplane
	tmp
)

fn LoadRibbonEmitter obj =
(
	local tmp = Wc3RibbonEmitter ()
	tmp.heightAbove = FloatParameterProcess obj.above.controller obj.above
	tmp.heightBelow = FloatParameterProcess obj.below.controller obj.below
	tmp.alpha = FloatParameterProcess obj.alpha.controller obj.alpha
	tmp.color = ColorParameterProcess obj.vertexcolor.controller obj.vertexcolor
	tmp.textureSlot = FloatParameterProcess obj.slots.controller obj.slots
	--tmp.textureSlot = obj.slots -- Unanimated for the time being
	if obj.visibility.isAnimated then tmp.visibility = VisibilityParameterProcess obj
	tmp.emissionRate = obj.emission
	tmp.lifeSpan = obj.life
	tmp.gravity = obj.gravity
	tmp.rows = obj.rows
	tmp.columns = obj.columns
	tmp.material = undefined
	tmp
)

fn LoadEvent obj =
(
	local tmp = Wc3Event ()
	tmp.eventTrack = EventProcess obj
	tmp
)

fn LoadCollisionShape obj =
(
	local tmp = Wc3CollisionShape ()
	if (classOf obj == Wc3CollisionBox or classOf obj == Box) then
	(
		tmp.shape = #Box
		tmp.vertices = #(obj.min,obj.max)
	)
	else (
		tmp.shape = #Sphere
		tmp.boundsRadius = obj.radius
		tmp.vertices = #(obj.center)
	)
	tmp
)

fn LoadCamera obj =
(
	local tmp = Wc3Camera ()
	tmp.name = obj.name
	tmp.pivot = obj.pos
	local trans = TransformationData ()
	--tmp.transformations = GetTransformations obj --some weird shit broke
	if obj.pos.isAnimated then trans.translation = GetAnimationData obj obj.pos.controller PositionKeys PositionType
	tmp.transformations = trans
	tmp.fieldOfView = obj.fov
	tmp.farClip = obj.farclip
	tmp.nearClip = obj.nearclip
	tmp.targetPos = obj.target.pos
	tmp.targetTransformations = GetTransformations obj.target
	tmp.targetTransformations.rotation = undefined
	tmp.targetTransformations.scale = undefined
	tmp
)

fn LoadWc3Object obj =
(
	local tmp = true
	local sub
	if IsAttachment obj then
	(
		sub = LoadAttachment obj
	)
	else if IsEventObject obj then
	(
		sub = LoadEvent obj
	)
	else if IsLight obj then
	(
		sub = LoadLight obj
	)
	else if IsParticleEmitter obj then
	(
		sub = LoadParticleEmitter obj
	)
	else if IsParticleEmitter2 obj then
	(
		sub = LoadParticleEmitter2 obj
	)
	else if IsRibbonEmitter obj then
	(
		sub = LoadRibbonEmitter obj
	)
	else if IsCollisionShape obj then
	(
		sub = LoadCollisionShape obj
	)
	else
		tmp = false
	if tmp then
	(
		local mdl = LoadMDLObject obj
		sub.MDLObj = mdl
		mdl.sub = sub
	)
)

fn AssignObjectId objCollection =
(
	for obj in objCollection do (
		obj.objId = ObjectIdCount
		ObjectIdCount =  ObjectIdCount + 1
	)
)

fn LoadObjects = 
(
	sceneObjects = objects
	max modify mode
	-- Load Geosets and Mdl Objects
	local progressAdder = 100.0/(sceneObjects.count as float)
	local progress = 0.0
	--progressStart "Loading objects and meshes"
	progressDownLevel 20 sceneObjects.count
	for obj in sceneObjects do
	(
		progressShowItem ("Loading: " + obj.name)
		if (not (IsIgnorable obj)) then
		(
			if IsMesh obj then
			(
				select obj
				LoadGeoset obj
				deselect obj
			)
			else
			(
				LoadWc3Object obj
			)
		)
	)
	max create mode
	progressUpLevel ()
	progressDownLevel 10 1
	progressShowItem "Building Object ids"
	-- Fix for Geoset generated bones that are linked to other geosets
	local FaultyBones = ArrayFilter Wc3Objects (fn func a = classOf a.maxObj == FakeObject and classOf a.maxObj.parent == Editable_mesh)
	for o in FaultyBones do
	(
		local m = ArrayFindItemEx Meshes o.maxObj.parent (fn func2 a b = a.mesh == b)
		o.maxObj.parent = BonesMeshRef[Meshes[m].matrices[1][1]].mdlObj.maxObj
	)
	--progressEnd()
	-- Load Helpers
	local helpersArray = #()
	for obj in Wc3Objects do
	(
		join helpersArray (GetParents obj.maxObj)
	)
	helpersArray = (ArrayMakeUnique helpersArray)
	-- Abusing lambda calculus for the win.
	fn tmp elem = (
		(ArrayFindItemEx Wc3Objects elem ObjectChecker) == 0
	)
	helpersArray = (ArrayFilter helpersArray tmp)
	for h in helpersArray do
		LoadHelper h
	-- Assign Parents
	for obj in Wc3Objects do
	(
		local index = ArrayFindItemEx Wc3Objects obj.maxObj.parent ObjectChecker
		if index != 0 then obj.parent = Wc3Objects[index]
	)
	-- Produce Hierarchy and assign Ids
	local h = Hierarchy ()
	for obj in Wc3Objects do
		HierarchyAddNode h obj
	Wc3Bones = HierarchyFilter h (fn isMdlBone a = ( a.sub.type == #Wc3Bone))
	AssignObjectId Wc3Bones
	Wc3Lights = HierarchyFilter h (fn isMdlHelper a = ( a.sub.type == #Wc3Light))
	AssignObjectId Wc3Lights
	Wc3Helpers = HierarchyFilter h (fn isMdlHelper a = ( a.sub.type == #Wc3Helper))
	AssignObjectId Wc3Helpers
	Wc3Attachments = HierarchyFilter h (fn isMdlHelper a = ( a.sub.type == #Wc3Attachment))
	AssignObjectId Wc3Attachments
	Wc3PartEmitters1 = HierarchyFilter h (fn isMdlHelper a = ( a.sub.type == #Wc3ParticleEmitter))
	AssignObjectId Wc3PartEmitters1
	Wc3PartEmitters2 = HierarchyFilter h (fn isMdlHelper a = ( a.sub.type == #Wc3ParticleEmitter2))
	AssignObjectId Wc3PartEmitters2
	Wc3RibbonEmitters = HierarchyFilter h (fn isMdlHelper a = ( a.sub.type == #Wc3RibbonEmitter))
	AssignObjectId Wc3RibbonEmitters
	Wc3Events = HierarchyFilter h (fn isMdlHelper a = ( a.sub.type == #Wc3Event))
	AssignObjectId Wc3Events
	Wc3Collisions = HierarchyFilter h (fn isMdlHelper a = ( a.sub.type == #Wc3CollisionShape))
	AssignObjectId Wc3Collisions
	-- Fix Geoset Matrices
	for m in Meshes do
	(
		for n in m.matrices do
		(
			for i = 1 to n.count do n[i] = BonesMeshRef[n[i]].mdlobj.objid
		)
	)
	-- Load Cameras
	Wc3Cameras = #()
	for c in $cameras do
	(
		if (not (IsIgnorable c)) then
			if classOf c == Targetcamera then append Wc3Cameras (LoadCamera c)
	)
	progressUpLevel ()
)

---------------------------------------------------------------------------------------------------------------------------------
-- 												Extents Initial Loading											      	 ---
---------------------------------------------------------------------------------------------------------------------------------
fn CalculateBoundsRadius minExtent maxExtent =
(
	if (dot maxExtent maxExtent) > (dot minExtent minExtent) then
		(length maxExtent)--*0.5
	else
		(length minExtent)--*0.5
)

fn CalculateExtents m start end interp =
(
	local MinExtent = undefined
	local MaxExtent = undefined
	for i = start to end by interp do
	(
		at time i
		(
			if m.visibility == true then
			(
				if MinExtent == undefined then MinExtent = m.min else
				(
					if MinExtent.x > m.min.x then MinExtent.x = m.min.x
					if MinExtent.y > m.min.y then MinExtent.y = m.min.y
					if MinExtent.z > m.min.z then MinExtent.z = m.min.z
				)
				if MaxExtent == undefined then MaxExtent = m.max else
				(
					if MaxExtent.x < m.max.x then MaxExtent.x = m.max.x
					if MaxExtent.y < m.max.y then MaxExtent.y = m.max.y
					if MaxExtent.z < m.max.z then MaxExtent.z = m.max.z
				)
			)
		)
	)
	if MinExtent == undefined then MinExtent = [0,0,0]
	if MaxExtent == undefined then MaxExtent = [0,0,0]
	Wc3Extent MinExtent MaxExtent (CalculateBoundsRadius MinExtent MaxExtent)
)

fn CalculateExtentsFromExtentsList extentList =
(
	local minExtent = undefined
	local maxExtent = undefined
	for m in extentList do
	(
		if minExtent == undefined then minExtent = copy m.minExtent else
		(
			if minExtent.x > m.minExtent.x then minExtent.x = m.minExtent.x
			if minExtent.y > m.minExtent.y then minExtent.y = m.minExtent.y
			if minExtent.z > m.minExtent.z then minExtent.z = m.minExtent.z
		)
		if maxExtent == undefined then maxExtent = copy m.maxExtent else
		(
			if maxExtent.x < m.maxExtent.x then maxExtent.x = m.maxExtent.x
			if maxExtent.y < m.maxExtent.y then maxExtent.y = m.maxExtent.y
			if maxExtent.z < m.maxExtent.z then maxExtent.z = m.maxExtent.z
		)
	)
	if minExtent == undefined then minExtent = [0,0,0]
	if maxExtent == undefined then maxExtent = [0,0,0]
	Wc3Extent minExtent maxExtent (CalculateBoundsRadius minExtent maxExtent)
)

fn LoadExtents extentsCalculation:1 extentsPrecision:5 =
(
	progressDownLevel 20 Wc3Sequences.count
	--progressShowItem "Calculating Extents"
	local SeqInterp = (extentsCalculation == 1)
	local interp = 1
	if (not SeqInterp) then interp = (((extentsPrecision*framerate)/30) as integer)
	if interp < 1 then interp = 1
	local tmp = #()
	if (Wc3Sequences.count != 0) then
	(
		for s in Wc3Sequences do
		(
			progressShowItem ("Calculating Extents: "+ s.name)
			if SeqInterp then interp = (((extentsPrecision*(s.endTime-s.startTime))/framerate) as integer)
			local tmp2 = #()
			for geo in Meshes do
			(
				local m = CalculateExtents geo.mesh s.startTime s.endTime interp
				append geo.extents m
				if not geo.unselectable then
					append tmp2 m
			)
			s.extents = CalculateExtentsFromExtentsList tmp2
			if not( (matchpattern s.name pattern:"death*") or (matchpattern s.name pattern:"decay*") or (matchpattern s.name pattern:"dissipate*")) then
				append tmp s.extents
		)
		Wc3Extents = CalculateExtentsFromExtentsList tmp
		for geo in Meshes do
			geo.mainExtents = CalculateExtentsFromExtentsList geo.extents
	)
	else
	(
		local tmp = #()
		for geo in Meshes do
		(
			local m = CalculateExtents geo.mesh 0 10
			geo.mainExtents = m
			append tmp m
		)
		Wc3Extents = CalculateExtentsFromExtentsList tmp
	)
	progressUpLevel ()
)

---------------------------------------------------------------------------------------------------------------------------------
-- 												GeosetsTweaks Initial Loading											  ---
---------------------------------------------------------------------------------------------------------------------------------

fn mergeExtents ex1 ex2 =
(
	local result = Wc3Extent ()
	result.minExtent = minPoint ex1.minExtent ex2.minExtent
	result.maxExtent = maxPoint ex1.maxExtent ex2.maxExtent
	result.bound = CalculateBoundsRadius result.minExtent result.maxExtent
	result
)

fn compareVisibility v1 v2 =
(
	if (v1 != undefined and v2 != undefined) then
		compareControllers v1.visibility.controller v2.visibility.controller
	else (v1 == undefined and v2 == undefined)
)

fn areGeosetsSimilar geo1 geo2 =
(
	local mesh1 = geo1.mesh
	local mesh2 = geo2.mesh
	mesh1.material == mesh2.material and (compareVisibility geo1.visObj geo2.visObj)
)

fn mergeGeosets geo1 geo2 =
(
	--Returns new index or current index
	fn expandMatrix &matrix newValue =
	(
		local found = ArrayFindItem matrix newValue
		if found == 0 then
		(
			append matrix newValue
			found = matrix.count - 1
		) else found = found - 1
		found
	)
	local matrices = deepCopy geo1.matrices
	local newGeo = Wc3Geoset ()
	local RefittedIndices = ArrayRepeat geo2.matrixindices.count undefined
	for j = 1 to  geo2.matrices.count do
	(
		local oldId = j - 1
		local newId = expandMatrix matrices geo2.matrices[j]
		for i = 1 to RefittedIndices.count do
			if geo2.matrixindices[i] == oldId then
				RefittedIndices[i] = newId
	)
	local faces = deepCopy geo1.faces
	local maxVert = geo1.vertices.count
	for f in geo2.faces do 
		append faces (f + maxVert)
	newGeo.vertices = #()
	join newGeo.vertices geo1.vertices
	join newGeo.vertices geo2.vertices
	newGeo.normals = #()
	join newGeo.normals geo1.normals
	join newGeo.normals geo2.normals
	newGeo.tvertices = #()
	local maxVert = geo1.tvertices.count
	if maxVert < geo2.tvertices.count then maxVert = geo2.tvertices.count
	local maxer1 = geo1.tvertices.count +1, maxer2 = geo2.tvertices.count + 1
	for i = 1 to maxVert do
	(
		local tmp = #()
		join tmp geo1.tvertices[mod i maxer1]
		join tmp geo2.tvertices[mod i maxer2]
		append newGeo.tvertices tmp
	)
	newGeo.matrixindices = #()
	join newGeo.matrixindices geo1.matrixindices
	join newGeo.matrixindices RefittedIndices
	newGeo.faces = faces
	newGeo.matrices = matrices
	newGeo.mainExtents = mergeExtents geo1.mainExtents geo2.mainExtents
	newGeo.extents = #()
	for i = 1 to geo1.extents.count do
		append newGeo.extents (mergeExtents geo1.extents[i] geo2.extents[i])
	newGeo.visObj = geo1.visObj
	-- Finished data copying
	-- Resetting bones
	local tmpBones = #()
	for m in matrices do
		for b in m do
		(
			local sub = Wc3Bones[b+1].sub
			local i1 = findItem sub.geosets geo1
			if i1 != 0 then deleteItem sub.geosets i1
			local i2 = findItem sub.geosets geo2
			if i2 != 0 then deleteItem sub.geosets i2
			append tmpBones sub
		)
	tmpBones = ArrayMakeUnique tmpBones
	for b in tmpBones do append b.geosets newGeo
	newGeo.mesh = FakeMesh "Generated Geoset" geo1.mesh.material geo1.mesh.modifiers
	newGeo.unselectable = geo1.unselectable and geo2.unselectable
	local i1 = findItem Meshes geo1
	deleteItem Meshes i1
	local i2 = findItem Meshes geo2
	deleteItem Meshes i2
	append Meshes newGeo
	newGeo
)

fn mergeSimilarGeosets =
(
	local pendingGeosets = Meshes.count != 0
	while (pendingGeosets) do
	(
		pendingGeosets = false
		local i = 1
		local stayIn = true
		local count = Meshes.count
		local onMerge
		while (i != count and stayIn) do
		(
			local geo = Meshes[i]
			onMerge = #(geo)
			for j = (i+1) to count do
				if j != i then
					if (areGeosetsSimilar geo Meshes[j]) then
					(
						stayIn = false
						append onMerge Meshes[j]
					)
			i += 1
		)
		if (not stayIn) then
		(
			local newGeo = onMerge[1]
			deleteItem onMerge 1
			while (onMerge.count != 0) do
			(
				newGeo = mergeGeosets newGeo onMerge[1]
				deleteItem onMerge 1
			)
			pendingGeosets = true
		)
	)
)

fn cleanMatrices geo =
(
	local tmp = ArrayMakeUnique geo.matrixIndices
	sort tmp
	local newMatrices = #()
	local tmp2 = #()
	local i = 0
	local indices = geo.matrixIndices
	for t in tmp do
	(
		append newMatrices geo.matrices[t+1]
		tmp2[t+1] = i
		i += 1
	)
	for i = 1 to indices.count do
		indices[i] = tmp2[indices[i] + 1]
	geo.matrices = newMatrices
	OK
)

-- this is better than using an octree
fn GatherBoundingVertices geo1 geo2 thresshold =
(
	thresshold = thresshold*0.5
	local min2 = geo2.mesh.min - thresshold
	local max2 = geo2.mesh.max + thresshold
	local result = #()
	for i = 1 to geo1.vertices.count do
		if (IsPointInBounds min2 max2 geo1.vertices[i]) then
			append result i
	result
)

fn applySyntheticSkin geo1 geo2 thresshold fixNormals =
(
	fn mergeMatrices array array2 =
	(
		local result = deepCopy array
		join result array2
		result = ArrayMakeUnique result
		sort result
		result
	)
	fn addNewMatrix &array a =
	(
		local result = ArrayFindItem array a
		if (result == 0) then
		(
			append array a
			result = array.count 
		)
		(result - 1)
	)
	vertices1 = GatherBoundingVertices geo1 geo2 thresshold
	vertices2 = GatherBoundingVertices geo2 geo1 thresshold
	ts = thresshold*0.5
	fn tmp p1 t = ((abs p1.x) < t and (abs p1.y) < t and (abs p1.z) < t)
	for i = 1 to vertices1.count do
	(
		local v = vertices1[i]
		for j = 1 to vertices2.count do
		(
			local w = vertices2[j]
			if (tmp (geo1.vertices[v] - geo2.vertices[w]) ts) then
			(
				local matrix = mergeMatrices (geo2.matrices[geo2.matrixIndices[w]+1])  (geo1.matrices[geo1.matrixIndices[v]+1])
				geo1.matrixindices[v] = addNewMatrix geo1.matrices matrix
				geo2.matrixindices[w] = addNewMatrix geo2.matrices matrix
				if fixNormals then
				(
					local newNormal = (geo1.normals[v] + geo2.normals[w])*0.5
					geo1.normals[v] = newNormal
					geo2.normals[w] = newNormal
				)
			)
		)
	)
	cleanMatrices geo1
	cleanMatrices geo2
)

fn sortMatrices geo =
(
	for m in geo.matrices do sort m
)

fn DoSyntheticSkin FixNormals:true Thresshold:0.1 =
(
	-- Gather skinnable objects
	local objs = ArrayFilter Meshes (fn func a = (getWc3Skin a.mesh) == #ATSkin)
	-- fix Matrices
	for g in objs do sortMatrices g
	-- Generate Collision Matrix
	local colMatrix = collideMatrix (ArrayMap objs (fn tmp a = a.mesh)) 1
	for i = 1 to objs.count do
		for j = i + 1 to objs.count do
			if colMatrix[i][j] then
				applySyntheticSkin objs[i] objs[j] Thresshold FixNormals
)

fn DoGeosetTweaks mergeGeosets:true ATFixNormals:true ATThresshold:0.1 =
(
	progressDownLevel 10 2
	progressShowItem "Calculating AT Skin"
	DoSyntheticSkin FixNormals:ATFixNormals Thresshold:ATThresshold
	progressShowItem "Merging Geosets"
	if mergeGeosets then mergeSimilarGeosets ()
	-- Fix Ids after merge
	GeosetCounts = 0
	for m in Meshes do
	(
		m.id = GeosetCounts
		GeosetCounts += 1
	)
	progressUpLevel ()
)

---------------------------------------------------------------------------------------------------------------------------------
-- 												Materials Initial Loading											       ---
---------------------------------------------------------------------------------------------------------------------------------

fn UsesAnimatedTextures T =
(
	s = T.filename
	(s != undefined) and (s.count > 3) and (SubString s (s.count-3) -1 == ".ifl")
)

--Igni - allow other texture file extensions
fn GetImageFile s = --this function changes the extention of .xxx to .blp
(
	return (getFilenameFile s) + ".blp"
)

fn GetImageArray T =
(
	fn get_names name a = append a name
	files = #()
	enumerateFiles T get_names files --#localOnly
	sort files
	--OptimizeArray files
	local i = findItem files T.filename
	if i != 0 then deleteItem files i
	files
)

fn LoadTexture file replaceable wrapWidth wrapHeight =
(
	fn TextureChecker a b = (
		a.replaceableId == b.replaceableId and a.path == b.path and 
		a.wrapWidth == b.wrapWidth and a.wrapHeight == b.wrapHeight
	)
	local texture = Wc3Texture file replaceable wrapWidth wrapHeight
	local index = ArrayFindItemEx Wc3Textures texture TextureChecker
	if index == 0 then
	(
		texture.id = TextureCount
		TextureCount = TextureCount + 1
		append Wc3Textures texture
	) else texture = Wc3Textures[index]
	texture
)

fn LoadTextures texture wc3Path replaceable =
(
	local result
	local texPath
	if texture != undefined and texture.filename != undefined then
	(
		if (UsesAnimatedTextures texture) then
		(
			local files = ArrayMap (GetImageArray texture) GetImageFile
			local playBack = texture.playbackrate
			local ids = #()
			for f in files do 
			(
				local t = \
					(LoadTexture (wc3Path + f) replaceable texture.coords.U_Tile texture.coords.V_Tile)
				append ids t.id
			)
			result = GetTextureAnimation ids playBack
		)
		else
		(
			local t
			if (replaceable == 0) then
				t = LoadTexture (wc3Path + (GetImageFile texture.filename)) replaceable texture.coords.U_Tile texture.coords.V_Tile
			else
				t = LoadTexture "" replaceable texture.coords.U_Tile texture.coords.V_Tile
			result = ParameterValue #Static t.id
		)
	)
	else
	(
		local u_tile = false, v_tile = false
		if (texture != undefined) then
		(
			u_tile = texture.coords.U_Tile
			v_tile = texture.coords.V_Tile
		)
		local t
		if (replaceable == 0) then
			t = LoadTexture "Textures\white.blp" replaceable u_tile v_tile
		else
			t = LoadTexture "" replaceable u_tile v_tile
		result = ParameterValue #Static t.id
	)
	result
)

fn buildTvertexAnimation cu cv =
(
	tv = Wc3TvertexAnim ()
	local trans = TransformationData ()
	trans.translation = combineTVAnimation cu cv undefined
	tv.transformation = trans
	append Wc3TvertexAnims tv
	tv
)

fn LoadLayer l mat = 
(
	local layer = Wc3Layer ()
	layer.filtermode = l.filtermode
	if classOf l != FakeMaterial then
		layer.alpha = FloatParameterProcess l.alpha.controller l.alpha
	else
		layer.alpha = ParameterValue #Static l.alpha
	layer.unselectable = l.unselectable
	layer.nodephtest = l.nodephtest
	layer.nodephset = l.nodephset
	layer.twosided = l.twosides
	layer.unshaded = l.unshaded
	layer.unfogged = l.unfogged
	layer.unknown1 = l.unknown1
	layer.unknown2 = l.unknown2
	layer.textures = LoadTextures l.texture l.path (l.retexture - 1)
	if (l.texture != undefined and classOf l.texture == BitmapTexture) then
	(
		layer.coordId = l.texture.coords.mapChannel - 1
		if (l.texture.coords.U_offset.isAnimated or l.texture.coords.V_offset.isAnimated) then
			layer.tvertexid = buildTvertexAnimation l.texture.coords.V_offset.controller l.texture.coords.U_offset.controller
	)
	else
		layer.coordId = 0
	layer
)

fn LoadLayers mat =
(
	local layers
	if IsWc3Material mat.maxObj or classOf mat.maxObj == FakeMaterial then 
		layers = #(mat.maxObj) else layers = (ArrayFilter mat.maxObj.materialList IsWc3Material)
	local Wc3Layers = #()
	if gmax then layers = ArrayInvert layers
	for l in layers do append Wc3Layers (LoadLayer l mat)
	Wc3Layers
)

fn LoadMaterial mat =
(
	local index = ArrayFindItemEx Wc3Materials mat ObjectChecker
	if index == 0 then
	(
		local result = Wc3Material ()
		result.maxObj = mat
		result.id = MaterialCount
		result.layers = LoadLayers result
		MaterialCount = MaterialCount + 1
		append Wc3Materials result
		result
	)
	else 
	(
		Wc3Materials[index]
	)
)

fn LoadMaterials =
(
	progressDownLevel 10 1
	progressShowItem "Loading Materials and TvertexAnimations"
	for geo in Meshes do (
		local m = geo.mesh
		local mat = m.material
		local tvertexAnim
		--if HasTvertexAnimation m then tvertexAnim = LoadTvertexAnims m
		if mat != undefined and (IsWc3Material mat or IsMultiLayered mat) then
		(
			local wc3Mat = LoadMaterial mat
			geo.material = wc3Mat
		)
		else
		(
			local wc3Mat
			if mat != undefined and classOf mat == Standardmaterial then
				wc3Mat = LoadMaterial (FakeMaterial texture:mat.diffuseMap alpha:mat.opacity)
			else
				wc3Mat = LoadMaterial DefaultMaterial
			geo.material = wc3Mat
			reportError "Material" geo.mesh.name undefined
		)
	)
	for r in Wc3RibbonEmitters do
	(
		local mat = r.maxObj.rmaterial
		if mat != undefined and (IsWc3Material mat or IsMultiLayered mat) then
		(
			local wc3Mat = LoadMaterial mat
			r.sub.material = wc3Mat
		)
		else
		(
			local wc3Mat
			if mat != undefined and classOf mat == Standardmaterial then
				wc3Mat = LoadMaterial (FakeMaterial texture:mat.diffuseMap alpha:mat.opacity)
			else
				wc3Mat = LoadMaterial DefaultMaterial
			r.sub.material = wc3Mat
		)
	)
	for p in Wc3PartEmitters2 do
	(
		local obj = p.maxObj
		if (obj.texture != undefined) then
			p.sub.texture = LoadTexture (obj.path + (GetImageFile obj.texture)) 0 false false
		else
			p.sub.texture = LoadTexture "Textures\white.blp" 0 false false
	)
	TvertexAnimCount = 0
	for tv in Wc3TvertexAnims do
	(
		tv.id = TvertexAnimCount
		TvertexAnimCount += 1
	)
	progressUpLevel ()
)

---------------------------------------------------------------------------------------------------------------------------------
-- 												GeosetAnims Initial Loading											       ---
---------------------------------------------------------------------------------------------------------------------------------

fn LoadGeosetAnims =
(
	progressDownLevel 10 1
	progressShowItem "Loading Geoset Animations"
	for geo in Meshes do (
		if (IsGeosetAnimated geo) then
		(
			local m = geo.mesh
			local geoAnim = Wc3GeosetAnim ()
			geoAnim.id = GeosetAnimCount
			GeosetAnimCount = GeosetAnimCount + 1
			geoAnim.geoset = geo
			if geo.visObj != undefined then geoAnim.alpha = VisibilityParameterProcess geo.visObj
			local vc = m.modifiers[#Wc3_Vertex_Color]
			if (vc != undefined) then 
			(
				geoAnim.usesColor = true
				geoAnim.color = ColorParameterProcess vc.VertexColor.controller vc.VertexColor
			)
			append GeosetAnims geoAnim
			geo.geosetAnimId = geoAnim.id
		)
	)
	progressUpLevel ()
)

fn CleanUpSession =
(
	Wc3Extents = Wc3Extent ()
	Wc3GlobalSequences = #() -- gathered on animation file
	Wc3Textures = #()
	Wc3Materials = #()
	Wc3TvertexAnims = #()
	Wc3Sequences = #()
	Meshes = #()
	BonesMeshRef = #()
	GeosetAnims = #()
	Wc3Objects = #() -- all the objets
	Wc3Bones = #()
	Wc3Lights = #()
	Wc3Helpers = #()
	Wc3Attachments = #()
	Wc3PartEmitters1 = #()
	Wc3PartEmitters2 = #()
	Wc3RibbonEmitters = #()
	Wc3Events = #()
	Wc3Cameras = #()
	Wc3Collisions = #()

	-- Needed specificaly
	GeosetCounts = 0
	BoneCounts = 0
	ObjectIdCount = 0
	MaterialCount = 0
	TextureCount = 0
	TvertexAnimCount = 0
	GeosetAnimCount = 0
)

fn ParseScene mergeGeosets:true extentsCalculation:1 extentsPrecision:5 ATFixNormals:true ATThresshold:0.1 faceGroups:false =
(
	CleanUpSession ()
	LoadSequences ()
	exportSmoothgroups = faceGroups
	InitiateAnimationData (ArrayMap Wc3Sequences (fn tmp a = interval a.startTime a.endTime)) Wc3Sequences
	LoadObjects ()
	LoadExtents extentsCalculation:extentsCalculation extentsPrecision:(11 - extentsPrecision)
	DoGeosetTweaks mergeGeosets:mergeGeosets ATFixNormals:ATFixNormals ATThresshold:ATThresshold
	LoadMaterials ()
	LoadGeosetAnims ()
	Wc3GlobalSequences = GetGlobalSequences ()
	local model = Wc3Model ()
	model.version = 800
	model.name = "Just Another Model"
	model.extents = Wc3Extents
	model.globalSequences = Wc3GlobalSequences
	model.sequences = Wc3Sequences
	model.textures = Wc3Textures
	model.materials = Wc3Materials
	model.tvertexAnims = Wc3TvertexAnims
	model.geosets = Meshes
	model.geosetAnims = GeosetAnims
	model.bones = Wc3Bones
	model.lights = Wc3Lights
	model.helpers = Wc3Helpers
	model.attachments = Wc3Attachments
	model.particleEmitters1 = Wc3PartEmitters1
	model.particleEmitters2 = Wc3PartEmitters2
	model.ribbonEmitters = Wc3RibbonEmitters
	model.events = Wc3Events
	model.cameras = Wc3Cameras
	model.CollisionShapes = Wc3Collisions
	model
)