
local ModelPath
local mdlObjects
local AnimatedTextures = 0
local bezierConstant = ((3.0*framerate)/(TicksPerFrame as float))
local rotationTCBObjs

fn writeKeys keys controller smoothType =
(
	fn getDelta t1 t2 = (( t2 - t1 ) as float)
	case smoothType of
	(
		#Bezier:
		(
			for i = 1 to keys.count do
			(
				local k = keys[i]
				local key = addNewKey controller k.time
				local tl
				local kb, ka
				key.value = k.value
				key.inTangentType = #custom
				key.outTangentType = #custom
				if (i > 1) then
				(
					kb = keys[i-1]
					local delta = (getDelta kb.time k.time)
					key.inTangent = bezierConstant*(k.intan-k.value)/delta
				)
				if (i < keys.count) then
				(
					ka = keys[i+1]
					local delta = (getDelta k.time ka.time)
					key.outTangent = bezierConstant*(k.outtan-k.value)/delta
				)
				if (classof controller == bezier_float) then
					tl = 0.333333
				else
					tl = [0.333333,0.333333,0.333333]
				key.outTangentLength = tl
				key.inTangentLength =  tl
			)
		)
		#NonInterp:
		(
			for k in keys do
			(
				local key = addNewKey controller k.time
				key.value = k.value
				key.inTangentType = #step
				key.outTangentType = #step
			)
		)
		default:
		(
			for k in keys do
			(
				local key = addNewKey controller k.time
				key.value = k.value
			)
		)
	)
)

fn writeKeysColor keys controller t =
(
	for k in keys do
	(
		local key = addNewKey controller k.time
		key.value = k.value
		if t then
		(
			key.inTangentType = #linear
			key.outTangentType = #linear
		)
	)
)

fn applyRotationAnimations obj rotData =
(
	obj.rotation.controller = linear_rotation ()
	local controller = obj.rotation.controller
	writeKeys rotData.keys controller #Linear
	if rotData.globalSequenceId != undefined then
	(
		setAfterORT controller #cycle
		enableORTs controller true
	)
)

fn applyTranslationAnimations obj posData =
(
	fn typeToController t =
	(
		case t of
		(
			#linear: linear_position ()
			#bezier: bezier_position ()
			#hermite: tcb_position ()
			default: bezier_position ()
		)
	)
	local smoothType = posData.type
	obj.pos.controller = typeToController smoothType
	local controller = obj.pos.controller
	local subPos
	in coordsys parent subPos = obj.pos
	for k in posData.keys do
	(
		k.value = k.value + subPos
		if smoothType == #Bezier then
		(
			k.intan = k.intan + subPos
			k.outtan = k.outtan + subPos
		)
	)
	writeKeys posData.keys controller smoothType
	if posData.globalSequenceId != undefined then
	(
		setAfterORT controller #cycle
		enableORTs controller true
	)
)

fn applyScaleAnimations obj scaleData =
(
	fn typeToController t =
	(
		case t of
		(
			#linear: linear_scale ()
			#bezier: bezier_scale ()
			#hermite: tcb_scale ()
			default: bezier_scale ()
		)
	)
	local smoothType = scaleData.type
	obj.scale.controller =  typeToController smoothType
	local controller = obj.scale.controller
	local subScale
	in coordsys parent subScale = obj.scale
	for k in scaleData.keys do
	(
		k.value = k.value/subScale
		if smoothType == #Bezier then
		(
			k.intan = k.intan/subScale
			k.outtan = k.outtan/subScale
		)
	)
	writeKeys scaleData.keys controller smoothType
	if scaleData.globalSequenceId != undefined then
	(
		setAfterORT controller #cycle
		enableORTs controller true
	)
)

fn applyTransformAnimations obj =
(
	local n = obj.maxObj
	local trans = obj.transformations
	if trans.rotation != undefined and Settings.importRotAnim then
	(
		applyRotationAnimations n trans.rotation
		if trans.rotation.type == #Hermite or trans.rotation.type == #Bezier then
			append rotationTCBObjs obj
	)
	if trans.translation != undefined and Settings.importTransAnim then
		applyTranslationAnimations n trans.translation
	if trans.scale != undefined and Settings.importScaleAnim then
		applyScaleAnimations n trans.scale
)

fn applyFloatParamAnimations obj animData defaultValue param =
(
	fn typeToController t =
	(
		case t of
		(
			#linear: linear_float ()
			#bezier: bezier_float ()
			#hermite: tcb_float ()
			default: bezier_float ()
		)
	)
	local smoothType = animData.type
	local controller = typeToController smoothType
	for k in animData.keys do
	(
		k.value = k.value + defaultValue
		if smoothType == #Bezier then
		(
			k.intan = k.intan + defaultValue
			k.outtan = k.outtan + defaultValue
		)
	)
	writeKeys animData.keys controller smoothType
	if animData.globalSequenceId != undefined then
	(
		setAfterORT controller #cycle
		enableORTs controller true
	)
	setPropertyController obj param controller
)

fn applyVisibilityAnimations o control =
(
	if Settings.importVisAnim then
	(
		if control.type == #NonInterp then
		(
			o.visibility = On_Off ()
			o.visibility.controller = On_Off ()
			local c = o.visibility.controller
			local keys = control.keys
			if keys[1].value > 0.1 then
				addNewKey c 0f
			deleteItem keys 1
			for k in keys do
				addNewKey c k.time
		) else
			applyFloatParamAnimations o control 0.0 #visibility
	)
)

fn applyColorParamAnimations obj animData defaultValue param =
(
	fn typeToController t =
	(
		case t of
		(
			#linear: bezier_point3 ()
			#bezier: bezier_color ()
			#hermite: tcb_point3 ()
			default: bezier_point3 ()
		)
	)
	local smoothType = animData.type
	local controller = typeToController smoothType
	for k in animData.keys do
	(
		k.value = k.value/255
		if smoothType == #Bezier then
		(
			k.intan = k.intan/255
			k.outtan = k.outtan/255
		)
	)
	writeKeysColor animData.keys controller (animData.type == #linear)
	if animData.globalSequenceId != undefined then
	(
		setAfterORT controller #cycle
		enableORTs controller true
	)
	setPropertyController obj param controller
)

fn applyFloatParameter obj prop param =
(
	if param.type == #Static then
		setProperty obj prop param.data
	else
		if Settings.importParamAnim then
			applyFloatParamAnimations obj param.data 0.0 prop
)

fn applyColorParameter obj prop param =
(
	if param.type == #Static then
		setProperty obj prop param.data
	else
		if Settings.importColorAnim then 
			applyColorParamAnimations obj param.data [0,0,0] prop
)

fn getCompoundMaterial =
(
	if gmax then
		Multimaterial ()
	else
		CompositeMaterial ()
)

fn FindTexture fname =
(
	(ModelPath+fname+".tga")
)

fn createLayer layer data n =
(
	layer.filtermode = data.filtermode
	--layer.unselectable  = data.unselectable
	layer.unshaded      = data.unshaded
	layer.unfogged      = data.unfogged
	layer.twosides      = data.twosided
	layer.NoDephTest = data.nodephtest
	layer.NoDephSet  = data.nodephset
	applyFloatParameter layer "alpha" data.alpha
	local bm
	if data.textures.type == #Static then
	(
		local tex = data.textures.data
		layer.retexture = tex.replaceableId
		if (tex.replaceableId == 1) then
		(
			local filename = getFilenameFile tex.path
			local layPath = getFilenamePath tex.path
			if layPath == undefined then layPath = ""
			local finalFile = FindTexture filename
			layer.texture = Bitmaptexture fileName:finalFile
			layer.path = layPath
			layer.texture.coordinates.U_tile = tex.wrapHeight
			layer.texture.coordinates.V_tile = tex.wrapWidth
			showTextureMap n layer.texture true
		)
	)
	else if Settings.importTXAnim then
	(
		fn createIFL tex =
		(
			local texPath = (ModelPath + "Textures" + (AnimatedTextures as string) + ".ifl")
			local expFile = createfile texPath
			AnimatedTextures = AnimatedTextures + 1
			for t in tex do
			(
				local finalFile = FindTexture (getFilenameFile t.path)
				format "%\n" finalFile to:expFile
			)
			close expFile
			texPath
		)
		local tex = data.textures.data
		local k = ArrayMap tex.keys (fn tmp3 a = a.value)
		local k2 = ArrayMap tex.keys (fn tmp2 a = a.time)
		local startTime = k2[1]
		for i = k2.count to 2 by -1 do k2[i] = k2[i] - k2[i-1]
		local avgT = (ArrayFold k2 (fn tmp a b = a + b) 0f)/k2.count
		local layPath = getFilenamePath k[1].path
		if layPath == undefined then layPath = ""
		local wh = k[1].wrapHeight
		local ww = k[1].wrapWidth
		local finalFile = createIFL k
		layer.texture = Bitmaptexture fileName:finalFile
		layer.path = layPath
		layer.texture.starttime = startTime
		layer.texture.playbackrate = (TicksPerFrame/(avgT as float))
		layer.texture.endcondition = if tex.globalSequenceId != undefined then 0 else 2
		layer.texture.coordinates.U_tile = wh
		layer.texture.coordinates.V_tile = ww
		showTextureMap n layer.texture true
	)
	if (data.tvertexid != undefined and layer.texture != undefined and Settings.importUVAnim) then
	(
		local tv = data.tvertexid.transformation
		local c = layer.texture.coordinates
		if (tv.translation != undefined) then
		(
			applyFloatParamAnimations c tv.translation[2] 0.0 #U_Offset
			applyFloatParamAnimations c tv.translation[1] 0.0 #V_Offset
		)
	)
	if (not gmax) then layer.showinviewport = true
)

fn recreateMaterials mats =
(
	local matName = "Imported Material 0"
	for i = 1 to mats.count do
	(
		local m = mats[i]
		local m2
		if m.layers.count == 1 then
		(
			-- Single Layered
			m2 = Warcraft3 ()
			createLayer m2 m.layers[1] m2
		)
		else
		(
			-- MultiLayered
			if gmax then m.layers = ArrayInvert m.layers
			m2 = getCompoundMaterial ()
			if (not gmax) then 
				m2.mapEnables = #(false, false, false, false, false, false, false, false, false)
			else
				m2.mapEnabled = #(false, false, false, false, false, false, false, false, false)
			for j = 1 to m.layers.count do
			(
				m2.materialList[j] = Warcraft3()
				if (not gmax) then 
					m2.mapEnables[j] = true
				else
					m2.mapEnabled[j] = true
				createLayer m2.materialList[j] m.layers[j] m2
			)
		)
		m2.name = matName + (i as string)
		m.maxObj = m2
		if (i < 25) then meditmaterials[i] = m2 
	)
)

fn createMeshWc3 g =
(
	fn createVertexMap g tverts map =
	(
		local obj = g.mesh
		local chan = map
		local tvertices = tverts
		meshop.setMapSupport obj chan true
		meshop.setNumMapVerts obj chan tvertices.count
		local func = meshop.setMapVert
		for i = 1 to tvertices.count do
		(
			local tv = [tvertices[i].x,tvertices[i].y,0]
			func obj chan i tv
		)
		func = meshop.setMapFace
		for i = 1 to obj.faces.count do
			func obj chan i g.faces[i]
	)
	local obj = mesh vertices:g.vertices faces:g.faces
	g.mesh = obj
	-- Normals
	for i = 1 to g.normals.count do
		setNormal obj i g.normals[i]
	-- Load Maps
	for i = 1 to g.numUVWmaps do
		createVertexMap g g.tvertices[i] i
	--  materials
	obj.material = g.material.maxObj
	obj
)

fn createMeshMax g =
(
	fn createVertexMap obj tvfaces tverts map =
	(
		local tvertices = tverts
		meshop.setMapSupport obj map true
		meshop.setNumMapVerts obj map tvertices.count
		local func = meshop.setMapVert
		for i = 1 to tvertices.count do
		(
			local tv = [tvertices[i].x,tvertices[i].y,0]
			func obj map i tv
		)
		func = meshop.setMapFace
		for i = 1 to tvfaces.count do
			func obj map i tvfaces[i]
	)
	local obj = mesh vertices:g.vertices faces:g.faces
	g.mesh = obj
	-- Normals
	for i = 1 to g.normals.count do
		setNormal obj i g.normals[i]
	-- Load Maps
	for i = 1 to g.numUVWmaps do
	(
		createVertexMap g.mesh g.tvfaces[i] g.tvertices[i] i
	)
	--  materials
	obj.material = g.material.maxObj
	obj
)

fn createMesh g =
(
	if classof g == Wc3Geoset then 
		createMeshWc3 g 
	else 
		createMeshMax g
)

fn getBones geo =
(
	local bns = #()
	for m in geo.matrices do
		for i = 1 to m.count do
		(
			local b = mdlObjects[m[i]].maxObj
			local index = findItem bns b
			if index == 0 then
			(
				append bns b
				index = bns.count
			)
			m[i] = index
		)
	bns
)

fn applySkinning g =
(
	local obj = g.mesh
	g.mesh = obj
	-- Skinning
	select obj
	max modify mode
	local bns = getBones g
	local SkinMod = Skin ()
	addModifier obj SkinMod
	modPanel.setCurrentObject SkinMod
	local func = skinOps.addBone
	for b in bns do func SkinMod b 0
	update obj
	max views redraw
	func = skinOps.ReplaceVertexWeights
	for i = 1 to g.vertices.count do
	(
		local matrix = g.matrices[g.matrixIndices[i]]
		local weight = (1.0/(matrix.count as float))
		local weights = ArrayRepeat matrix.count weight
		func SkinMod i matrix weights
	)
	max create mode
)

fn recreateGeosetsGMAX geos =
(
	for g in geos do
		createMesh g
)

fn recreateGeosetsMAX geos =
(
	local geoObjLayer = LayerManager.newLayerFromName "Geosets"
	for g in geos do
		geoObjLayer.addNode (createMesh g)
)

local recreateGeosets
if gmax then 
	recreateGeosets = recreateGeosetsGMAX 
else
	recreateGeosets = recreateGeosetsMAX

fn setupNode obj n =
(
	setBillboard n obj.billboardType
)

fn createBone obj =
(
	if Settings.importBonesAsMaxBones  then
	(
		p = BoneSys.createBone [0,0,0] [0,0,0] [0,0,0.1]
		p.width 		= 0
		p.height		= 0
		p.showLinks	= true
		p.boneScaleType = #none
		p.pos = obj.pivot
		p.name = obj.name
	)
	else
		p = point pos:obj.pivot name:obj.name
	obj.maxObj = p
	setupNode obj p
	p
)

fn createHelper obj =
(
	local p
	if Settings.importHelpersAsMaxBones  then
	(
		p = BoneSys.createBone [0,0,0] [0,0,0] [0,0,0.1]
		p.width 		= 0
		p.height		= 0
		p.showLinks	= true
		p.boneScaleType = #none
		p.pos = obj.pivot
		p.name = obj.name
	)
	else
		p = point pos:obj.pivot name:obj.name
	p.wirecolor = green
	obj.maxObj = p
	setupNode obj p
	p
)

fn getAttachmentValues name =
(
	fn searchString sarr s =
	(
		local index = 1
		for i = 1 to sarr.count where matchPattern s pattern:sarr[i] do index = i
		index
	)
	local s1 = filterString name " "
	local n1 = searchString attachmentList s1[1]
	if (s1.count > 1) then
	(
		local n2 = searchString attachmentAdd1List s1[2]
		local n3 = searchString attachmentAdd2List s1[2]
		#(n1,n2,n3)
	)
	else #(n1)
)

fn createLightNormal obj =
(
	local l
	local sub = obj.sub
	case (sub.lightType) of
	(
		#Omnidirectional: l = Omnilight pos:obj.pivot name:obj.name
		#Directional: l = Directionallight pos:obj.pivot name:obj.name
		default: l = Omnilight pos:obj.pivot name:obj.name
	)
	obj.maxObj = l
	applyFloatParameter l "farAttenStart" sub.attenuationStart
	applyFloatParameter l "farAttenEnd" sub.attenuationEnd
	applyFloatParameter l "multiplier" sub.intensity
	applyColorParameter l "color" sub.color
	setupNode obj l
	l
)

fn createLightSc2 obj =
(
	local l
	local sub = obj.sub
	case (sub.lightType) of
	(
		#Omnidirectional: l = SC2OmniLight pos:obj.pivot name:obj.name
		#Directional: l = SC2OmniLight pos:obj.pivot name:obj.name
		default: l = SC2OmniLight pos:obj.pivot name:obj.name
	)
	obj.maxObj = l
	applyFloatParameter l "attenStart" sub.attenuationStart
	applyFloatParameter l "attenEnd" sub.attenuationEnd
	applyFloatParameter l "intensity" sub.intensity
	applyColorParameter l "color" sub.color
	l
)

fn createLight obj =
(
	createLightNormal obj
)

fn createAttachmentWc3 obj =
(
	local p = Wc3AttachPoint pos:obj.pivot name:obj.name wirecolor:red
	obj.maxObj = p
	local flags = getAttachmentValues obj.name
	if flags[1] != undefined then p.nameFirst = flags[1]
	if flags[2] != undefined then p.nameAdd1 = flags[2]
	if flags[3] != undefined then p.nameAdd2 = flags[3]
	setupNode obj p
	p
)

fn createAttachmentSc2 obj =
(
	local p = SC2AttachPt pos:obj.pivot name:obj.name
	obj.maxObj = p
	local s = "Ref "
	local flags = getAttachmentValues obj.name
	if flags[1] != undefined then s = s + " " + flags[1]
	if flags[2] != undefined then s = s + " " + flags[2]
	if flags[3] != undefined then s = s + " " + flags[3]
	obj.name = s
	p
)

fn createAttachment obj =
(
	createAttachmentWc3 obj
)


fn createParticleEmitter obj =
(
	if gmax then
		return OK
	local p = SuperSpray pos:obj.pivot name:obj.name
	obj.maxObj = p
	local sub = obj.sub
	applyFloatParameter p "Birth_Rate" sub.emissionRate
	applyFloatParameter p "Plane_Spread" sub.longitude 
	applyFloatParameter p "Axis_Spread" sub.latitude
	if not (sub.gravity.type == #Static and sub.gravity.data == 0) then
	(
		local g = Gravity pos:obj.pivot name:(obj.name + "_Gravity")
		applyFloatParameter g "strength" sub.gravity
		bindSpaceWarp p g
	)
	applyFloatParameter p "life" sub.lifespan
	applyFloatParameter p "speed" sub.initVelocity
	setUserProp p "ParticlePath" sub.path
	p.Off_Plane = 90.0
	p.Off_Axis = 0.0
	setupNode obj p
	p
)

fn createParticleEmitter2 obj =
(
	local p = BlizzParticle2 pos:obj.pivot name:obj.name
	obj.maxObj = p
	local sub = obj.sub
	applyFloatParameter p "speed" sub.speed
	applyFloatParameter p "variation" sub.variation
	applyFloatParameter p "coneangle" sub.latitude
	applyFloatParameter p "gravity" sub.gravity
	p.life = sub.lifeSpan
	applyFloatParameter p "PartsEmit" sub.emissionRate
	p.Squirt = sub.squirt
	applyFloatParameter p "width" sub.width
	applyFloatParameter p "length" sub.length
	p.LineEmitter = sub.lineEmitter
	p.rtexture = sub.replaceableId
	local tex = sub.texture
	if (tex.replaceableId == 1) then
	(
		local filename = getFilenameFile tex.path
		local layPath = getFilenamePath tex.path
		if layPath == undefined then layPath = ""
		local finalFile = FindTexture filename
		p.texture = finalFile
		p.path = layPath
	)
	p.filtermode = sub.filtermode
	p.rows = sub.rows
	p.cols = sub.columns
	p.Unshaded = sub.unshaded
	p.Unfogged = sub.unfogged
	p.SortZ = sub.sortPrimesFarZ
	p.particleType = sub.particleType
	p.taillength = sub.tailLength
	p.midtime = sub.time
	p.startcolor = sub.startColor
	p.middlecolor = sub.midColor
	p.endcolor = sub.endColor
	p.startalpha = sub.alpha.x
	p.middlealpha = sub.alpha.y
	p.endalpha = sub.alpha.z
	p.startscale = sub.scaling.x
	p.middlescale = sub.scaling.y
	p.endscale = sub.scaling.z
	p.startLifespanHead = sub.lifeSpanUVAnim.x
	p.endLifespanHead = sub.lifeSpanUVAnim.y
	p.repeatLifespanHead = sub.lifeSpanUVAnim.z
	p.startDecayHead = sub.decayUVAnim.x
	p.endDecayHead = sub.decayUVAnim.y
	p.repeatDecayHead = sub.decayUVAnim.z
	p.startLifespanTail = sub.TailUVAnim.x
	p.endLifespanTail = sub.TailUVAnim.y
	p.repeatLifespanTail = sub.TailUVAnim.z
	p.startDecayTail = sub.TailDecayUVAnim.x
	p.endDecayTail = sub.TailDecayUVAnim.y
	p.repeatDecayTail = sub.TailDecayUVAnim.z
	p.priorityplane = sub.priorityPlane
	p.ParticleSpace = sub.modelSpace
	p.XYQuads = sub.xyQuad
	setupNode obj p
	p
)

fn createRibbonEmitter obj =
(
	local r = BlizRibbon pos:obj.pivot name:obj.name
	obj.maxObj = r
	local sub = obj.sub
	applyFloatParameter r "above" sub.heightAbove
	applyFloatParameter r "below" sub.heightBelow
	applyFloatParameter r "alpha" sub.alpha
	applyFloatParameter r "slots" sub.textureSlot
	applyColorParameter r "VertexColor" sub.color
	r.emission = sub.emissionRate
	r.life = sub.lifeSpan
	r.gravity = sub.gravity
	r.rows = sub.rows
	r.columns = sub.columns
	r.rmaterial = sub.material.maxObj
	setupNode obj r
	r
)

local eventObjColor = color 224 198 87

fn createEventObject obj =
(
	local p = Point pos:obj.pivot name:("Obj:" + obj.name) wirecolor:eventObjColor
	obj.maxObj = p
	p.cross = off
	p.axistripod = on
	local NT = notetrack "Events"
	addNoteTrack p NT
	for k in obj.sub.eventTrack.keys do
		addNewNoteKey NT.keys k.time
	setupNode obj p
	p
)

fn createCollisionShapesMAX obj =
(
	local p
	local sub = obj.sub
	if sub.shape == 2 then
		p = Wc3CollisionSphere pos:sub.vertices[1] radius:sub.boundsRadius name:obj.name wirecolor:blue
	else
	(
		local center = (sub.vertices[1] + sub.vertices[2])*0.5
		center.z = minVal sub.vertices[1].z sub.vertices[2].z
		local l = abs (sub.vertices[1].y - sub.vertices[2].y)
		local w= abs (sub.vertices[1].x - sub.vertices[2].x)
		local h= abs (sub.vertices[1].z - sub.vertices[2].z)
		p = Wc3CollisionBox pos:center width:w length:l height:h name:obj.name wirecolor:blue
	)
	p.pivot = obj.pivot
	obj.maxObj = p
	setupNode obj p
	p
)

fn createCollisionShapesGMAX obj =
(
	local p
	local sub = obj.sub
	if sub.shape == 2 then
		p = Sphere pos:sub.vertices[1] radius:sub.boundsRadius name:obj.name wirecolor:blue
	else
	(
		local center = (sub.vertices[1] + sub.vertices[2])*0.5
		center.z = minVal sub.vertices[1].z sub.vertices[2].z
		local l = abs (sub.vertices[1].y - sub.vertices[2].y)
		local w= abs (sub.vertices[1].x - sub.vertices[2].x)
		local h= abs (sub.vertices[1].z - sub.vertices[2].z)
		p = Box pos:center width:w length:l height:h name:obj.name wirecolor:blue
	)
	p.pivot = obj.pivot
	p.xray = on
	freeze p
	obj.maxObj = p
	setupNode obj p
	p
)

local createCollisionShapes
if gmax then 
	createCollisionShapes = createCollisionShapesGMAX 
else
	createCollisionShapes = createCollisionShapesMAX

fn createCamera cam layer =
(
	local obj = Targetcamera fov:cam.fieldOfView nearclip:cam.nearClip farclip:cam.farClip  pos:cam.pivot target:(Targetobject transform:(matrix3 [1,0,0] [0,1,0] [0,0,1] cam.targetPos))
	obj.name = cam.name
	if (not gmax) then layer.addNode obj
	obj
)

fn recreateObjectsMAX =
(
	local animObjLayer = LayerManager.newLayerFromName "Bones & Helpers"
	local lightsObjLayer = LayerManager.newLayerFromName "Lights"
	local attachmentsObjLayer = LayerManager.newLayerFromName "Attachments"
	local eventsObjLayer = LayerManager.newLayerFromName "Event Objects"
	local effectsObjLayer = LayerManager.newLayerFromName "Effects"
	local collisionObjLayer = LayerManager.newLayerFromName "Collision Objects"
	for i = 1 to mdlObjects.count do
	(
		local obj = mdlObjects[i]
		if obj != undefined then
		(
			case obj.sub.type of
			(
				#Wc3Bone: 
					if Settings.importBones and (not obj.deleted) then 
						animObjLayer.addNode (createBone obj)
				#Wc3Helper: 
					if Settings.importHelpers and (not obj.deleted) then 
						animObjLayer.addNode (createHelper obj)
				#Wc3Light: 
					if Settings.importLights then lightsObjLayer.addNode (createLight obj)
				#Wc3Attachment: 
					if Settings.importAttachments then attachmentsObjLayer.addNode (createAttachment obj)
				#Wc3Event: 
					if Settings.importEO then eventsObjLayer.addNode (createEventObject obj)
				#Wc3CollisionShape: 
					if Settings.importCS then collisionObjLayer.addNode (createCollisionShapes obj)
				#Wc3ParticleEmitter: 
					if Settings.importPE1 then effectsObjLayer.addNode (createParticleEmitter obj)
				#Wc3ParticleEmitter2: 
					if Settings.importPE2 then effectsObjLayer.addNode (createParticleEmitter2 obj)
				#Wc3RibbonEmitter: 
					if Settings.importRE then effectsObjLayer.addNode (createRibbonEmitter obj)
			)
		)
	)
)

fn recreateObjectsGMAX =
(
	for i = 1 to mdlObjects.count do
	(
		local obj = mdlObjects[i]
		if obj != undefined  then
		(
			case obj.sub.type of
			(
				#Wc3Bone: 
					if Settings.importBones and (not obj.deleted) then createBone obj
				#Wc3Helper: 
					if Settings.importHelpers and (not obj.deleted) then createHelper obj
				#Wc3Light: 
					if Settings.importLights then createLight obj
				#Wc3Attachment: 
					if Settings.importAttachments then createAttachment obj
				#Wc3Event: 
					if Settings.importEO then createEventObject obj
				#Wc3CollisionShape: 
					if Settings.importCS then createCollisionShapes obj
				#Wc3ParticleEmitter: 
					if Settings.importPE1 then createParticleEmitter obj
				#Wc3ParticleEmitter2: 
					if Settings.importPE2 then createParticleEmitter2 obj
				#Wc3RibbonEmitter: 
					if Settings.importRE then createRibbonEmitter obj
			)
		)
	)
)

local recreateObjects
if gmax then 
	recreateObjects = recreateObjectsGMAX 
else
	recreateObjects = recreateObjectsMAX


fn relinkObjects =
(
	for i = 1 to mdlObjects.count do
	(
		local obj = mdlObjects[i]
		if obj != undefined and obj.maxObj != undefined then
		(
			local parent = obj.parent
			while (parent != undefined and parent.maxObj == undefined) do
				parent = parent.parent
			if parent != undefined then
				obj.maxObj.parent = parent.maxObj
		)
	)
)

fn reapplySkinning geos =
(
	for g in geos do
		applySkinning g
)

fn recreateSequencesMAX sequences =
(
	fn sequenceToString seq =
	(
		local newLine =  (bit.intAsChar 13)+(bit.intAsChar 10)
		local s = "\""+ seq.name + "\""
		if seq.loop then s += newLine + "NonLooping"
		if seq.rarity != undefined then s += (newLine + "Rarity " + ((seq.rarity as integer) as string))
		if seq.movespeed != undefined then s += (newLine + "MoveSpeed " + ((seq.movespeed as integer) as string))
		s
	)
	while( numNoteTracks rootNode != 0 ) do deleteNoteTrack rootNode (getNoteTrack rootNode 1)
	local NT = notetrack "animations"
	local ftm_cnt = FrameTagManager.CreateNewTag
	local ftm_gti = FrameTagManager.GetTagID
	addNoteTrack rootNode NT
	for s in Sequences do
	(
		local val = sequenceToString s
		local key1 = addNewNoteKey NT.keys s.startTime
		local key2 = addNewNoteKey NT.keys s.endTime
		key1.value = val
		key2.value = val
		local tag1 = ftm_cnt (s.name + " Start") s.startTime
		local id = ftm_gti tag1
		local tag2 = ftm_cnt (s.name + " End") s.endTime lockID:id
	)
)

fn recreateSequencesGMAX sequences =
(
	fn sequenceToString seq =
	(
		local newLine =  (bit.intAsChar 13)+(bit.intAsChar 10)
		local s = "\""+ seq.name + "\""
		if seq.loop then s += newLine + "NonLooping"
		if seq.rarity != undefined then s += (newLine + "Rarity " + ((seq.rarity as integer) as string))
		if seq.movespeed != undefined then s += (newLine + "MoveSpeed " + ((seq.movespeed as integer) as string))
		s
	)
	while( numNoteTracks rootNode != 0 ) do deleteNoteTrack rootNode (getNoteTrack rootNode 1)
	local NT = notetrack "animations"
	addNoteTrack rootNode NT
	for s in Sequences do
	(
		local val = sequenceToString s
		local key1 = addNewNoteKey NT.keys s.startTime
		local key2 = addNewNoteKey NT.keys s.endTime
		key1.value = val
		key2.value = val
	)
)

local recreateSequences
if gmax then 
	recreateSequences = recreateSequencesGMAX 
else
	recreateSequences = recreateSequencesMAX

mapped fn applyGeosetAnimations geoAnim =
(
	if geoAnim.alpha.type == #Animated then
		applyVisibilityAnimations geoAnim.geoset.mesh geoAnim.alpha.data
	if geoAnim.usesColor then
	(
		local m = Wc3VertexMod ()
		addModifier geoAnim.geoset.mesh m
		applyColorParameter m "VertexColor" geoAnim.color
	)
)

mapped fn applyObjectVisibilityAnimations obj =
(
	if (IsProperty obj.sub #visibility) and obj.sub.visibility.type == #Animated and obj.maxObj != undefined then
		applyVisibilityAnimations obj.maxObj obj.sub.visibility.data
)

fn animateObjects =
(
	for m in mdlObjects where m != undefined and m.maxObj != undefined do
		applyTransformAnimations m
)

mapped fn applyFunc node &func2 =
(
	func2 node
	applyFunc node.children func2
)

fn applyFuncToSelection &func =
(
	fn isselected sel nod =
	(
		for s in sel do
			if nod == s then return true
		false
	)
	rootnodes = #()
	for s in selection do
		if s.parent == undefined or isselected selection s.parent then
		append rootnodes s
	applyFunc rootnodes func
)

fn fixRotationTCB objs =
(
	deselect selection
	local arr = #()
	if objs.count != 0 then arr[objs.count] = undefined
	for i = 1 to objs.count do
		arr[i] = objs[i].maxObj
	select arr
	applyFuncToSelection (fn func a = a.rotation.controller = tcb_rotation())
	for o in objs where (o.transformations.rotation.globalSequenceId == undefined) do
	(
		local keys1 = o.maxObj.rotation.controller.keys
		local keys2 = o.transformations.rotation.keys
		local n = keys2.count
		for i = 1 to n where keys2[i].lock do
			if (keys1[i] != undefined) then
				keys1[i].continuity = 0.0
	)
	deselect selection
)

fn rebuildSceneModel model originalPath =
(
	rotationTCBObjs = #()
	undo off (
		with redraw off
		(
			AnimatedTextures = 0
			ModelPath = originalPath
			progressDownLevel 5 1
			progressShowItem "Building Materials"
			if Settings.importMaterials then recreateMaterials model.materials
			progressUpLevel ()
			progressDownLevel 5 1
			progressShowItem "Building Meshes"
			recreateGeosets model.geosets
			MdlObjects = model.ex
			progressUpLevel ()
			progressDownLevel 10 1
			progressShowItem "Building Objects"
			if Settings.importObjects then recreateObjects ()
			if Settings.importObjects then relinkObjects ()
			local camObjLayer = if (not gmax) then LayerManager.newLayerFromName "Cameras" else undefined
			if Settings.importCameras then for c in model.cameras do
				createCamera c camObjLayer
			progressUpLevel ()
		)
		progressDownLevel 15 1
		progressShowItem "Skinning the Meshes"
		if Settings.importSkinning then reapplySkinning model.geosets
		progressUpLevel ()
		with redraw off
		(
			progressDownLevel 15 1
			progressShowItem "Animating the Objects"
			if Settings.importAnimations then
			(
				recreateSequences model.sequences
				animateObjects ()
				applyGeosetAnimations model.geosetAnims
				applyObjectVisibilityAnimations MdlObjects
			)
			progressUpLevel ()
		)
		max select all
		-- Only once doesn't do a proper zoom
		for i = 1 to 3 do
			max zoomext sel all
		max views redraw
		ClearSelection()
		if Settings.importRotAnim then fixRotationTCB rotationTCBObjs
	)
)