
	fn F2I f = --this function transforms a frame to it's exact integer value.
	(
		(f as integer)/TicksPerFrame
	)
----------------------------------------------------------------------------------------------------------------------------
	fn F2R f = --this function transforms a frame to it's real value.
	(
		(f as float)/(TicksPerFrame as float)
	)
----------------------------------------------------------------------------------------------------------------------------
	--this function transforms a life frames to universal wc3 life.
	fn MxLife2W3Life l = F2R(l)/framerate
	fn W3Life2MxLife l = (l*framerate as time)
----------------------------------------------------------------------------------------------------------------------------
	fn F2WS f = --this function transforms a 3dsmax frame to a warcraft 3 frame
	(
		(((((f as float)/TicksPerFrame)*1000.0/framerate) as integer) as string)
	)
	
	fn F2WI f = --this function transforms a 3dsmax frame to a warcraft 3 frame
	(
		((((f as float)/TicksPerFrame)*1000.0/framerate) as integer)
	)
	
	fn B2I b = --takes a boolean and returns an integer
	(
		if b then 1 else 0
	)
	
	fn I2B i = --takes a integer, returns boolean
	(
		if i == 0 then false else true
	)
	
	fn round a = (ceil ((floor (a*2))/2) as integer)

	fn relativeFactor obj = (obj.max - obj.min)
	--fn relativePosition p2 p r = (p2-p)/r
----------------------------------------------------------------------------------------------------------------------------
-- Quat Algebra Sector
	fn QuatMult q s =
	(
		local p = q
		p.angle *= s
		p
	)

	fn QuatSub q p =
	(
		q * inverse(p)
	)
	
	fn DotQuat q1 q2 = (q1.w*q2.w + q1.x*q2.x + q1.y*q2.y + q1.z*q2.z)
	
	fn AngleBetweenQuat q1 q2 = acos(DotQuat q1 q2) --actually returns half the angle
	
	fn CheckLock q1 q2 = --this function checks if there can be squad interpolation errors between 2 quats
	(
		local a = (acos (DotQuat q1 q2))
		if a > 45 and a < 135 then true else false
	)
	
	fn PowQuat quatsamp t = (exp (LogN(quatsamp)*t)) --powers the quaternion to any value t
		
	fn SumQuat q1 q2 = --does an arithmetic plussing
	(
		(quat (q1.x+q2.x) (q1.y+q2.y) (q1.z+q2.z) (q1.w+q2.w))
	)
	
	fn RotatePoint p q =
	(
		local q2 = quat p.x p.y p.z 0
		q2 = q*q2*(inverse q)
		[q2.x,q2.y,q2.z]
	)
	
----------------------------------------------------------------------------------------------------------------------------
----- Array Utility Functions
	
	fn ArrayCompare a b =				-- returns true if two arrays are exactly the same
	(
		if a == b then return 0			-- == operator for arrays only returns true if they are at same address, as in c
		if (classOf a) != Array and (classOf b) != Array then if a < b then return -1 else if a > b then return 1
		for i = 1 to a.count do
		(
			if b.count < i then return 1
			local c = ArrayCompare a[i] b[i]
			if c < 0 then return -1
			else if c > 0 then return 1
		)
		if a.count < b.count then return -1
		return 0
	)
	
	fn ArrayFindItem array value =
	(
		for i = 1 to array.count do
		(
			if (ArrayCompare array[i] value) == 0 then return i
		)
		return 0
	)
	
	fn ArrayFindItemEx array value &func =
	(
		for i = 1 to array.count do
		(
			if (func array[i] value) then return i
		)
		return 0
	)
	
	fn ArrayMakeUnique array =
	(
		local result = #()
		for elem in array do
			if (findItem result elem) == 0 then append result elem
		result
	)
	
	fn ArrayMakeUniqueEx array &func =
	(
		local result = #()
		for elem in array do
			if (ArrayFindItemEx result elem func) == 0 then append result elem
		result
	)
	
	fn OptimizeArray A = --this function reduces optimizes repeated array values
	(
		for i = 1 to A.count do 
		for j = A.count to i+1 by -1 do 
		 if A[i] == A[j] do 
		  deleteItem A j
	)
	
	fn ArrayFold array &func init =
	(
		for elem in array do
			init = func init elem
		init
	)
	
	fn ArrayFilter array &func =
	(
		local result = #()
		for elem in array do
			if (func elem) then append result elem
		result
	)
	
	fn ArrayCheckProperty array &func =
	(
		for elem in array do
			if not (func elem) then return false
		true
	)
	
	fn ArrayMap array &func =
	(
		local result = #()
		if array.count != 0 then result[array.count] = undefined
		local i = 1
		for elem in array do
		(
			result[i] = (func elem)
			i += 1
		)
		result
	)
	
	fn ArrayRepeat size value =
	(
		local result = #()
		if size != 0 then result[size] = value
		for i = 1 to size do
			result[i] = value
		result
	)
	
	fn ArrayAppendUnique &array val =
	(
		local index = findItem array val
		if index == 0 then
		(
			append array val
			index = array.count
		)
		index
	)

	fn ArrayGroup arr func =
	(
		local result = #()
		for a in arr do
		(
			local noMatch = true
			for i = 1 to result.count where ((func result[i][1] a) and noMatch) do
			(
				noMatch = false
				append result[i] a
			)
			if noMatch then
				append result #(a)
		)
		result
	)
	
	fn ArrayConcat arr =
	(
		result = #()
		for a in arr do join result a
		result
	)
	
	fn ArrayInvert arr =
	(
		result = #()
		local n = arr.count
		for i = 1 to n do
			result[i] = arr[n - i + 1]
		result
	)
	
	fn ArrayInverse arr =
	(
		result = #()
		for i = 1 to arr.count do
		(
			local index = arr[i]
			if result[index] == undefined then
				result[index] = #()
			append result[index] i
		)
		result
	)
	
----------------------------------------------------------------------------------------------------------------------------
----- Mapping Functions
fn mapPoint3 p f s =
	f ( f ( f s p.x) p.y) p.z

fn mapPoint2 p f s =
	f ( f s p.x) p.y

fn mapQuat p f s =
	f (f ( f ( f s p.x) p.y) p.z) p.w

fn mapVal p f s = f p s

	
----------------------------------------------------------------------------------------------------------------------------
----- Hierarchy Functions
	
--  Not implemented as methods due to some errors on recursion
	struct Hierarchy (
		object,
		children = #(),
		parent
	)
	
	fn HierarchyFindElement h elem =
	(
		for c in h.children do 
		(
			if c.object == elem then return c
			local result = HierarchyFindElement c elem
			if result != undefined then return result
		)
		return undefined
	)
	
	fn HierarchyAddNode h obj =
	(
		local backup = HierarchyFindElement h obj
		if backup == undefined then (
			local parent
			if obj.parent != undefined then
			(
				parent = HierarchyAddNode h obj.parent
			)
			local result
			if parent == undefined then
			(
				result = (Hierarchy obj #() undefined )
				append h.children result
			)
			else
			(
				result = (Hierarchy obj #() obj.parent)
				append parent.children result
			)
			result
		)
		else (
			backup
		)
	)
	
	fn HierarchyFilter h &func =
	(
		local result = #()
		if h.object != undefined and (func h.object) then append result h.object
		for c in h.children do 
		(
			join result (HierarchyFilter c func)
		)
		result
	)

	struct DoubleV(val1, val2) --useful for paring 2 values

	-- Would be better if changed its inner structure for a hashtable
	struct Dictionary
	(
		entries = #(),
		dataList = #(),
		searchFunction = findItem
	)
	
	fn DictionaryAddEntry dictio entry =
	(
		append dictio.entries entry
		append dictio.dataList #()
	)
	
	fn DictionaryGetEntryData dictio entry =
	(
		local index = dictio.searchFunction dictio.entries entry
		if (index != 0) then dictio.dataList[index] else undefined
	)
	
	fn DictionaryAddEntryData dictio entry data =
	(
		local index = dictio.searchFunction dictio.entries entry
		if (index != 0) then append dictio.dataList[index] data else
		(
			DictionaryAddEntry dictio entry
			append dictio.dataList[dictio.dataList.count] data
		)
	)
	
	fn DictionaryOutputData dictio =
	(
		local count = dictio.entries.count
		local result = #()
		for i = 1 to count do append result #(dictio.entries[i],dictio.dataList[i])
		result
	)
	

----------------------------------------------------------------------------------------------------------------------------
fn IsWc3Material m = (classOf m == Warcraft3)
----------------------------------------------------------------------------------------------------------------------------
fn IsMultiLayeredMAX m = (
	fn tmp m = m == undefined or IsWc3Material m
	if (classOf m == compositematerial) then
		ArrayCheckProperty m.materialList tmp
	else
		false
)

fn IsMultiLayeredGMAX m = (
	local result = true
	if (classOf m == multimaterial) then
	(
		for i = 1 to m.materialList.count do
			result = result and ((not m.mapEnabled[i]) or (IsWc3Material m.materialList[i]))
		result
	)
	else
		false
)

local IsMultiLayered
if gmax then
	IsMultiLayered = IsMultiLayeredGMAX
else
	IsMultiLayered = IsMultiLayeredMAX

----------------------------------------------------------------------------------------------------------------------------
fn IsSkinned o = (o.modifiers[#Skin] != undefined)
----------------------------------------------------------------------------------------------------------------------------
fn IsMesh o = (classOf o == Editable_mesh)
----------------------------------------------------------------------------------------------------------------------------
fn IsCamera o = (classOf o == Targetcamera)
----------------------------------------------------------------------------------------------------------------------------
fn IsCollisionShapeGmax o = (classOf o == Box or classOf o == Sphere)
fn IsCollisionShapeMax o =((classof o) as string == "Wc3CollisionBox") or ((classof o) as string == "Wc3CollisionSphere")
global IsCollisionShape
if gmax then IsCollisionShape = IsCollisionShapeGmax else IsCollisionShape = IsCollisionShapeMax
----------------------------------------------------------------------------------------------------------------------------
fn IsAttachment o = (classOf o == Wc3AttachPoint)
----------------------------------------------------------------------------------------------------------------------------
fn IsEventObject o = ((substring o.name 1 4) == "Obj:")
----------------------------------------------------------------------------------------------------------------------------
fn IsParticleEmitter o = (classOf o == SuperSpray) --currently disable
----------------------------------------------------------------------------------------------------------------------------
fn IsParticleEmitter2 o = (classOf o == BlizzParticle2) --requires BlizPart2 plugin
----------------------------------------------------------------------------------------------------------------------------
fn IsRibbonEmitter o = (classOf o == BlizRibbon) --requires BlizRibbon plugin
----------------------------------------------------------------------------------------------------------------------------
fn IsLight o = (classOf o == OmniLight or classOf o == freeSpot or classOf o == Directionallight or classOf o == Skylight)
----------------------------------------------------------------------------------------------------------------------------

	-- returns #NormalSkin or #ATSkin
	fn getWc3Skin o =
	(
		fn InterfaceToSkin s = if s == 2 then #ATSkin else #NormalSkin
		local s = getUserProp o "Wc3Skin"
		if s == undefined then #NormalSkin else (InterfaceToSkin (s as integer))
	)

	fn setWc3Skin o s =
	(
		fn skinToInterface s =
		(
			case (s) of
			(
				#NormalSkin: 1
				#ATSkin: 2
				default: 0
			)
		)
		setUserProp o "Wc3Skin" (skinToInterface s)
	)

	-- can return #None, #Full, #X, #Y, #Z
	fn getBillboard o =
	(
		fn interfaceToBillboard b =
		(
			case (b) of
			(
				1: #None
				2: #Full
				3: #X
				4: #Y
				5: #Z
				default: #None
			)
		)
		local b = getUserProp o "Billboard"
		if b == undefined then #None else (interfaceToBillboard (b as integer))
	)

	fn setBillboard o b =
	(
		fn billboardToInterface b =
		(
			case (b) of
			(
				#None: 1
				#Full: 2
				#X: 3
				#Y: 4
				#Z: 5
				default: 0
			)
		)
		setUserProp o "Billboard" (billboardToInterface b)
	)

	-- returns true or false
	fn getCameraAnchored o =
	(
		local b = getUserProp o "CameraAnchored"
		if b == undefined then false else I2B (b as integer)
	)

	fn setCameraAnchored o c =
	(
		setUserProp o "CameraAnchored" (B2I c)
	)
	
	fn getUnselectable o =
	(
		local b = getUserProp o "Unselectable"
		if b == undefined then false else I2B (b as integer)
	)

	fn setUnselectable o c =
	(
		setUserProp o "Unselectable" (B2I c)
	)
	
	fn getUsesBake o =
	(
		local b = getUserProp o "UsesBake"
		if b == undefined then false else I2B (b as integer)
	)
	
	fn setUsesBake o c =
	(
		setUserProp o "UsesBake" (B2I c)
	)
	
	fn getBakeInterp o =
	(
		local b = getUserProp o "BakeInterp"
		if b == undefined then 5 else (b as integer)
	)
	
	fn setBakeInterp o c =
	(
		setUserProp o "BakeInterp" c
	)
	
global attachmentList = #("Head", "Overhead", "Origin", "Foot", "Chest", "Hand", "Weapon", "Sprite")
global attachmentAdd1List = #("None", "Right", "Left")
global attachmentAdd2List = #("None", "RallyPoint", "First", "Second", "Third", "Fourth", "Fifth", "Sixth")
	
--------------------------------------------------------------------------------------------------------------------------------------------------------------
-- Collision Detection
--------------------------------------------------------------------------------------------------------------------------------------------------------------
	fn IsPointInBounds min1 max1 p =
	(
		p.x > min1.x and p.x < max1.x and \
		p.y > min1.y and p.y < max1.y and \
		p.z > min1.z and p.z < max1.z
	)

	fn boundsCollide min1 max1 min2 max2 =
	(
		max1.x > min2.x and \
		min1.x < max2.x and \
		max1.y > min2.y and \
		min1.y < max2.y and \
		max1.z > min2.z and \
		min1.z < max2.z
	)

	fn objectBoundsCollide o p = boundsCollide o.min o.max p.min p.max

	fn collideMatrix objs threshold =
	(
		local count = objs.count
		local result = ArrayRepeat count #()
		for i = 1 to count do result[i] = ArrayRepeat count false
		local tmp = threshold*0.5
		local func = boundsCollide -- load to cache
		for i = 1 to count do
		(
			local min1 = objs[i].min - tmp
			local max1 = objs[i].max + tmp
			for j = i to count do
			(
				local r = (func min1 max1 (objs[j].min - tmp) (objs[j].max + tmp))
				result[i][j] = r
				result[j][i] = r
			)
		)
		result
	)
	fn minVal a b = if a > b then b else a
	fn maxVal a b = if a > b then a else b
	fn minPoint p1 p2 = [minVal p1.x p2.x, minVal p1.y p2.y, minVal p1.z p2.z]
	fn maxPoint p1 p2 = [maxVal p1.x p2.x, maxVal p1.y p2.y, maxVal p1.z p2.z]

	fn getAnimationIntervals =
	(
		local s = "\*"
		local v = undefined
		local result = #()
		if HasNoteTracks Rootnode then
		(
			Nt = GetNoteTrack Rootnode 1
			if Nt != undefined then
			(
				for k in Nt.keys do
				(
					if s == k.value and k.value != "\*" and v != undefined then
					(
						append result (interval v.time k.time)
						s = "\*"
						v = undefined
					)
					else
					(
						s = k.value
						v = k
					)
				)
			)
		)
		result
	)
	
	fn getKeysWithinInterval keys interv =
	(
		result = #()
		for k in keys do 
			if k.time > interv.start and k.time < interv.end then
				append result k
		result
	)
	
	fn getKeysOnSequences keys sequences inclusive:false =
	(
		local func1, func2
		if inclusive then
		(
			func1 = fn tmp1 a b = a <= b
			func2 = fn tmp2 a b = a >= b
		) else (
			func1 = fn tmp1 a b = a < b
			func2 = fn tmp2 a b = a > b
		)
		result = #()
		result[sequences.count] = undefined
		for j = 1 to sequences.count do result[j] = #()
		local i = 1
		local count = keys.count
		for j = 1 to sequences.count do
		(
			local s = sequences[j]
			while (i <= count and (func1 keys[i].time s.end)) do
			(
				if (func2 keys[i].time s.start) then
					append result[j] keys[i]
				i += 1
			)
		)
		result
	)
	
	fn getKeysOnSequences2 keys sequences inclusive:false =
	(
		local func1, func2
		if inclusive then
		(
			func1 = fn tmp1 a b = a <= b
			func2 = fn tmp2 a b = a >= b
		) else (
			func1 = fn tmp1 a b = a < b
			func2 = fn tmp2 a b = a > b
		)
		result = #()
		result[sequences.count] = undefined
		for j = 1 to sequences.count do result[j] = #()
		local i = 1
		local count = keys.count
		for j = 1 to sequences.count do
		(
			local s = sequences[j]
			while (i <= count and (func1 keys[i] s.end)) do
			(
				if (func2 keys[i] s.start) then
					append result[j] keys[i]
				i += 1
			)
		)
		result
	)
	
	fn addCopyKey controller sourceKey time =
	(
		local k = addNewKey controller time
		k.value = sourceKey.value
	)
	
	fn _getParentsEx obj &tmp=
	(
		if obj.parent != undefined then
		(
			append tmp obj.parent
			_getParentsEx obj.parent tmp
		)
	)

	fn GetParents obj =
	(
		result = #()
		_getParentsEx obj result
		result
	)
	
	fn compareControllers c1 c2 =
	(
		if ((classOf c1) == (classOf c2)) then
		(
			local k1 = c1.keys
			local k2 = c2.keys
			local n = k1.count
			if (n == k2.count) then
			(
				local result = true
				for i = 1 to n where result do
				(
					local v1,v2
					at time k1[i].time v1 = c1.value
					at time k2[i].time v2 = c2.value
					result = result and ((k1[i].time == k2[i].time) and (v1 == v2))
				)
				result
			) else false
		) else false
	)

	