
---------------------------------------------------------------------------------------------------------------------------------
-- 												Definitions											  					---
---------------------------------------------------------------------------------------------------------------------------------
global GlobalSequencesGathered
global GlobalSequenceCount = 0
global Sequences = #()
local IKChains
local IKChainsKeys
local IKChainsBones

struct TCBKeys (
	time, 
	value, 
	tension, 
	continuity, 
	bias
)

struct ParameterValue (
	type, -- can be #Static or #Animated
	data
)

struct KeyData (
	time,
	value
)

struct SmoothKeyData (
	time,
	value,
	intan,
	outtan
)

struct AnimationData (
	type,
	globalSequenceId = undefined,
	keys
)

struct TransformationData (
	translation,
	rotation,
	scale
)

---------------------------------------------------------------------------------------------------------------------------------
-- 												GlobalSequences										  					  ---
---------------------------------------------------------------------------------------------------------------------------------

fn LoadGlobalSequence value =
(
	local index = findItem GlobalSequencesGathered value
	if index == 0 then
	(
		append GlobalSequencesGathered value
		GlobalSequenceCount = GlobalSequenceCount + 1
		index = GlobalSequenceCount
	)
	index
)

---------------------------------------------------------------------------------------------------------------------------------
-- 												Interpolation Functions											  		  ---
---------------------------------------------------------------------------------------------------------------------------------

--This function has been extracted from: http://www.enlight.ru/demo/faq/smth.phtml?query=alg_curves_splines
--This code according to the site comes directly from Autodesk!!! and had been used on 3dsmax.
--I'm using this function to compute TCB Tangents of Quaternions
fn ComputeTangents prevkey currentkey nextkey =
(
	local qprev, qnext, q, p3, qcurrent
	local qp, qm, qa, qb, qae, qbe
	local tm,cm,cp,bm,bp,tmcm,tmcp,ksm,ksp,kdm,kdp,c
	local dt,f_p,f_n,pt,nt
	qcurrent = copy currentkey.value
	if prevkey != undefined then
	(
		if qcurrent.angle > 360.0 then
		(
			p3 = qcurrent.axis
			q = quat p3.x p3.y 0.0 0.0
			qm = LogN q
		)
		else
		(
			qprev = copy prevkey.value
			if (DotQuat qcurrent qprev) < 0 then qprev = - qprev
			qm = LnDif qprev qcurrent
		)
	)
	if nextkey != undefined then
	(
		qnext = copy nextkey.value
		if (qnext.angle) > 360.0 then
		(
			p3 = qnext.axis
			q = quat p3.x p3.y 0.0 0.0
			qp = LogN q
		)
		else
		(
			if (DotQuat qcurrent qnext) < 0 then qnext = - qnext
			qp = LnDif qcurrent qnext
		)
	)
	if prevkey == undefined then qm = copy qp
	if nextkey == undefined then qp = copy qm
	--Irregular Times settings
	if nextkey != undefined and prevkey != undefined then -- nextkey && prevkey
	(
		pt = prevkey.time
		nt = nextkey.time
		dt = 0.5*(F2R (nt - pt))
		f_p = (F2R (currentkey.time - pt))/dt --if intervals are the same, f_P = 1.0
		f_n = (F2R (nt- currentkey.time))/dt
		c = (abs (currentkey.continuity))
		f_p = f_p + c - c*f_p
		f_n = f_n + c - c*f_n
	)
	else
	(
		f_p = 1.0
		f_n = 1.0
	)
	--end Irregular Times Settings
	cm = 1.0 - currentkey.continuity
	tm = 0.5*(1.0 - currentkey.tension)
	cp = 2.0 - cm
	bm = 1.0 - currentkey.bias
	bp = 2.0 - bm
	tmcm = tm*cm
	tmcp = tm*cp
	ksm = 1.0 - tmcm*bp*f_p
	ksp = (-tmcp)*bm*f_p
	kdm = tmcp*bp*f_n
	kdp = tmcm*bm*f_n - 1.0
	--calculation
	qa = (SumQuat (qm*kdm) (qp*kdp))*0.5
	qb = (SumQuat (qm*ksm) (qp*ksp))*0.5
	qae = qcurrent*(exp qa)
	qbe = qcurrent*(exp qb)
	(DoubleV qae qbe)
)

-- Used for TCB controllers which are not rotation
fn InTan p1 p2 p3 t c b = --these functions need fixing for irregular timings
(
	S = ((1-t)*(1-c)*(1+b)*0.5)
	R = ((1-t)*(1+c)*(1-b)*0.5)
	(((p2-p1)*S)+((p3-p2)*R))
)
fn OutTan p1 p2 p3 t c b =
(
	S = ((1-t)*(1+c)*(1+b)*0.5)
	R = ((1-t)*(1-c)*(1-b)*0.5)
	(((p2-p1)*S)+((p3-p2)*R))
)

--- Bezier Functions for computing tangents

fn BezierOutTan pe pd a d = 3*pe  + (2*d - 5*a - 9*pd)/6.0

fn BezierInTan pe pd a d = 3*pd + (2*a - 9*pe - 5*d)/6.0


/*
fn getDelta t1 t2 = (( t2 - t1 ) as float)

fn getBezierOuttan k1 k2=
(
	local delta = (getDelta k1.time k2.time)
	local beta = (3*framerate)/delta
	local rs = k1.outTangent/beta
	rs
)

fn getBezierIntan k1 k2=
(
	local delta = (getDelta k1.time k2.time)
	local beta = (3*framerate)/delta
	local rs = k2.inTangent/beta
	rs
)
*/

fn controllerIsNonInterp c =
(
	local result = true
	for k in c.keys where result do
		result = result and (k.inTangentType == #step) and (k.outTangentType == #step)
	result
)

-- Functions for Smooth Rotation Controller

fn E2P e1 = point3 e1.x e1.y e1.z
fn P2E p1 = eulerangles p1.x p1.y p1.z

fn GetTangentSmooth p1 p2 p3 =
(
	local s1 = E2P (p1 as eulerangles)
	local s2 = E2P (p2 as eulerangles)
	local s3 = E2P (p3 as eulerangles)
	local e1 =  (s2 + (s2-s1)*0.25 - (s3-s2)*0.25)
	((P2E e1) as quat)
)

---- The next functions get the relative transformation to a parent
fn GetRelativePosition t o parent p =
(
	at time t
	(
		local i = o.pos
		local r = relativeFactor parent
		local c = parent.center
	)
	((i-c)/r - p)
)

fn GetRot t o r pr hp =
(
	at time t local v = o.rotation*r -- 
	if hp then at time t  v = (v/(o.parent.rotation*pr))
	v
)

-- for Bipeds only
fn GetRotBiped t o r pr =
(
	at time t local v = (inverse (biped.getTransform o #rotation))*r
	if o.parent != undefined then at time t  v = (v/((inverse (biped.getTransform o.parent #rotation))*pr))
	v
)

--- BezierProcessing
fn ProcessBezier keys =
(
	local k = #()
	local x
	local outtan
	local intan
	if keys.count != 0 then k[keys.count] = undefined
	for x = 1 to keys.count do
	(
		if x != 1 then intan = (BezierInTan keys[x-1].outtan keys[x].intan keys[x-1].value keys[x].value) else intan = keys[x].value
		if x != keys.count then  outtan = (BezierOutTan keys[x].outtan keys[x+1].intan keys[x].value keys[x+1].value) else outtan = keys[x].value
		k[x] = (SmoothKeyData keys[x].time keys[x].value intan outtan)
	)
	k
)

---------------------------------------------------------------------------------------------------------------------------------
-- 												IK & Constraint Functions											  	  ---
---------------------------------------------------------------------------------------------------------------------------------

fn getSubControl c i =
(
	if (isProperty c #getSubCtrl) then c.getSubCtrl i
	else (
		getPropertyController c (c.GetName i)
	)
)


fn isIKAffected o = classOf o.transform.controller == IKControl
-- only supports Orientation and Look at Constraints
fn isObjectConstrained o =
(
	local c = o.rotation.controller
	if classof c == rotation_list then
	(
		local n = c.count
		local result = false
		for i = 1 to n do
		(
			local tmp = classOf (getSubControl c i) 
			result = result or (tmp == Orientation_Constraint) or (tmp == LookAt_Constraint)
		)
		result
	)
	else false
)

fn getKeysFromObject o pos:true rot:true sca:true =
(
	local result = #()
	local keys
	local count = 0
	if pos then
	(
		keys = o.pos.controller.keys
		count = keys.count
		if count > 0 then result[count] = undefined
		for i = 1 to count do result[i] = keys[i].time
	)
	if rot then
	(
		keys = o.rotation.controller.keys
		local count2 = keys.count
		if count2 > 0 then result[count + count2] = undefined
		for i = 1 to count2 do result[count + i] = keys[i].time
		count += count2
	)
	if sca then
	(
		keys = o.scale.controller.keys
		local count2 = keys.count
		if count2 > 0 then result[count + count2] = undefined
		for i = 1 to count2 do result[count + i] = keys[i].time
		count += count2
	)
	ArrayMakeUnique result
)

fn getIKId o =
(
	result = #()
	for i = 1 to IKChains.count do
		if (findItem IKChainsBones[i] o) != 0 then
			append result i
	result
)

fn getGoalDistance o t =
(
	local p1,p2
	at time t p1 = o.pos
	at time t p2 = o.controller.startjoint.pos
	distance p1 p2
)

fn getAngleofGoalChange o t1 t2 =
(
	local v1, v2
	at time t1 v1 = o.controller.startjoint.pos - o.pos
	at time t2 v2 = o.controller.startjoint.pos - o.pos
	acos ((dot v1 v2)/((sqrt (dot v1 v1)) * (sqrt (dot v2 v2))))
)

fn getObjectConstraintKeys obj =
(
	local c = obj.rotation.controller
	local n = c.count
	local objChain = #(obj)
	join objChain (GetParents obj)
	local orChain = #()
	local laChain = #()
	for i = 1 to n do
	(
		local c2 = getSubControl c i
		if (classof c2 == Orientation_Constraint) then
		(
			local n2 = c2.getNumTargets()
			for j = 1 to n2 do
			(
				local o = c2.getNode j
				local tmp = #(o)
				join tmp (GetParents o)
				join orChain tmp
			)
		)
		else 	if (classof c2 == LookAt_Constraint) then
		(
			local n2 = c2.getNumTargets()
			for j = 1 to n2 do
			(
				local o = c2.getNode j
				local tmp = #(o)
				join tmp (GetParents o)
				join laChain tmp
			)
		)
	)
	orChain = ArrayMakeUnique orChain
	laChain = ArrayMakeUnique laChain
	local needPos = laChain.count != 0
	local keys = #()
	for s in orChain where s != undefined do
		join keys (getKeysFromObject s pos:false sca:false)
	for s in laChain where s != undefined do
		join keys (getKeysFromObject s sca:false)
	for s in objChain where s != undefined do
		join keys (getKeysFromObject s sca:false)
	keys = ArrayMakeUnique keys
	sort keys
	keys
)

-----------------------------------------------------------------------------------------------------------------------------
-- 												Translation											  					  ---
---------------------------------------------------------------------------------------------------------------------------------

fn PositionKeys o =
(
	local keys = #()
	local keys2
	local c = o.pos.controller
	local p = o.pos
	p = at time 0f in coordsys parent o.pos
	case (classOf c) of
	(
		bezier_position: (
			keys2 =#()
			k = c.keys
			keys2[k.count] = undefined
			for i = 1 to k.count do
			(
				at time k[i].time local v = in coordsys parent o.pos
				v = v - p
				local in1 
				if i != 1 then
					at time  (k[i-1].time + 2*(k[i].time - k[i-1].time)/3) in coordsys parent in1 = o.pos - p
				else
					in1 = v
				local out1
				if i != k.count then
					at time  (k[i].time + (k[i+1].time - k[i].time)/3) in coordsys parent out1 = o.pos - p
				else
					out1 = v
				keys2[i] = (SmoothKeyData k[i].time v in1 out1)
			)
			keys = ProcessBezier keys2
		)
		tcb_position: (
			append keys (SmoothKeyData 0f [0,0,0] undefined undefined)
			for k in c.keys do 
			(
				local v = at time k.time in coordsys parent o.pos
				v = v - p
				append keys (SmoothKeyData k.time v undefined undefined)
			)
			append keys (SmoothKeyData 999999999f [0,0,0] undefined undefined)
			for k = 1 to (c.keys.count) do
			(
				local t = (c.keys[k].tension - 25) / 25
				local c_ = (c.keys[k].continuity - 25) / 25				-- "c" is taken by the position controller
				local b = (c.keys[k].bias - 25) / 25
				keys[k+1].intan = (InTan keys[k].value keys[k+1].value keys[k+2].value t c_ b)
				keys[k+1].outtan = (OutTan keys[k].value keys[k+1].value keys[k+2].value t c_ b)
			)
			deleteItem keys (keys.count)
			deleteItem keys 1
		)
		default: (
			for k in c.keys do 
			(
				local v = at time k.time in coordsys parent o.pos
				v = v - p
				append keys (KeyData k.time v)
			)
		)
	)
	keys
)

fn PositionType controller =
(
	local type
	case (classOf controller) of
	(
		bezier_position: 
		(
			if (controllerIsNonInterp controller) then
				type = #NonInterp
			else
				type = #Bezier
		)
		tcb_position: type = #Hermite
		linear_position: type =#Linear
		default: type = #NonInterp
	)
	type
)

fn PositionKeysBiped o =
(
	local keys = #()
	
	--Igni - Added support for biped
	-- Same as rotation, but now for position see RotationKeys for details
	if (not (hasProperty o.transform.controller "vertical" or hasProperty o.transform.controller "horizontal")) then
	(
		return keys
	)
	
	append keys (SmoothKeyData 0f [0,0,0] undefined undefined)

	local p = o.transform.pos
	
	if o.parent != undefined then
	(
		p -= o.parent.transform.pos
		pr = o.parent.transform.rotation
	)
	
	-- Indices into the horizontal and vertical key arrays.
	local i_h = 1
	local i_v = 1
	
	-- Arrays containing the keys into both vertical and horizontal controllers
	keys_v = o.transform.controller.vertical.controller.keys
	keys_h = o.transform.controller.horizontal.controller.keys
	
	local keys_t = #()
	local keys_c = #()
	local keys_b = #()
	
	-- Since the full animation is actually contained within two different tracks, we need to combine those
	-- we do this by going through both arrays each time taking the smallest time and obtaining the position.
	while (i_h <= keys_h.count) or (i_v <= keys_v.count) do
	(
		local translated_pos
		if i_h > keys_h.count then
		(
			at time keys_v[i_v].time translated_pos = o.transform.pos - p
			if o.parent != undefined then at time keys_v[i_v].time translated_pos = ((o.transform.pos - o.parent.transform.pos) * inverse(o.parent.transform.rotation * inverse(pr))) - p
			append keys (SmoothKeyData keys_v[i_v].time translated_pos undefined undefined)
			
			append keys_t (biped.getKey o.transform.controller.vertical.controller i_v).tension
			append keys_c (biped.getKey o.transform.controller.vertical.controller i_v).continuity
			append keys_b (biped.getKey o.transform.controller.vertical.controller i_v).bias
			
			i_v += 1
		)
		else if i_v > keys_v.count then
		(
			at time keys_h[i_h].time translated_pos = o.transform.pos - p
			if o.parent != undefined then at time keys_h[i_h].time translated_pos = ((o.transform.pos - o.parent.transform.pos) * inverse(o.parent.transform.rotation * inverse(pr))) - p
			append keys (SmoothKeyData keys_h[i_h].time translated_pos undefined undefined)
			
			append keys_t (biped.getKey o.transform.controller.horizontal.controller i_h).tension
			append keys_c (biped.getKey o.transform.controller.horizontal.controller i_h).continuity
			append keys_b (biped.getKey o.transform.controller.horizontal.controller i_h).bias
			
			i_h += 1
		)
		else if keys_v[i_v].time < keys_h[i_h].time then
		(
			at time keys_v[i_v].time translated_pos = o.transform.pos - p
			if o.parent != undefined then at time keys_v[i_v].time translated_pos = ((o.transform.pos - o.parent.transform.pos) * inverse(o.parent.transform.rotation * inverse(pr))) - p
			append keys (SmoothKeyData keys_v[i_v].time translated_pos undefined undefined)
			
			append keys_t (biped.getKey o.transform.controller.vertical.controller i_v).tension
			append keys_c (biped.getKey o.transform.controller.vertical.controller i_v).continuity
			append keys_b (biped.getKey o.transform.controller.vertical.controller i_v).bias
			
			i_v += 1
		)
		else if keys_v[i_v].time > keys_h[i_h].time then
		(
			at time keys_h[i_h].time translated_pos = o.transform.pos - p
			if o.parent != undefined then at time keys_h[i_h].time translated_pos = ((o.transform.pos - o.parent.transform.pos) * inverse(o.parent.transform.rotation * inverse(pr))) - p
			append keys (SmoothKeyData keys_h[i_h].time translated_pos undefined undefined)
			
			append keys_t (biped.getKey o.transform.controller.horizontal.controller i_h).tension
			append keys_c (biped.getKey o.transform.controller.horizontal.controller i_h).continuity
			append keys_b (biped.getKey o.transform.controller.horizontal.controller i_h).bias
			
			i_h += 1
		)
		else -- keys_v[i_v].time = keys_h[i_h].time (merge two keys at the same time)
		(
			at time keys_h[i_h].time translated_pos = o.transform.pos - p
			if o.parent != undefined then at time keys_h[i_h].time translated_pos = ((o.transform.pos - o.parent.transform.pos) * inverse(o.parent.transform.rotation * inverse(pr))) - p
			append keys (SmoothKeyData keys_h[i_h].time translated_pos undefined undefined)
			
			append keys_t (biped.getKey o.transform.controller.vertical.controller i_v).tension
			append keys_c (biped.getKey o.transform.controller.vertical.controller i_v).continuity
			append keys_b (biped.getKey o.transform.controller.vertical.controller i_v).bias
			
			i_h += 1
			i_v += 1
		)
	)
	append keys (SmoothKeyData 999999999f [0,0,0] undefined undefined)
	for k = 1 to (keys.count - 2) do
	(
		local t = (keys_t[k] - 25) / 25
		local c_ = (keys_c[k] - 25) / 25				-- "c" is taken by the position controller
		local b = (keys_b[k] - 25) / 25
		keys[k+1].intan = (InTan keys[k].value keys[k+1].value keys[k+2].value t c_ b)
		keys[k+1].outtan = (OutTan keys[k].value keys[k+1].value keys[k+2].value t c_ b)
	)
	deleteItem keys (keys.count)
	deleteItem keys 1
	keys	

)

---------------------------------------------------------------------------------------------------------------------------------
-- 												Rotation											  						  ---
---------------------------------------------------------------------------------------------------------------------------------

fn RotationKeys o =
(
	local keys = #()
	local keys2 = #()
	local tangents
	local c = o.rotation.controller
	local hasParent = o.parent != undefined and (classOf o.parent != Uvwmap) -- This last part is for UVW Gizmo
	local r = (inverse o.rotation)
	local k = c.keys
	local pr = (quat 0 0 0 1)
	if hasParent and (classOf o.parent != Uvwmap) then pr = (inverse o.parent.rotation)
	case (classOf c) of
	(
		tcb_rotation: (
			local keys2 = #()
			local tangents
			if k.count != 0 then keys2[k.count] = undefined
			for i = 1 to k.count do
			(
				local v = GetRot k[i].time o r pr hasParent
				keys2[i] = (TCBKeys k[i].time v ((k[i].tension - 25.0)/25.0) ((k[i].continuity - 25.0)/25.0) ((k[i].bias - 25.0)/25.0) )
			)
			if keys2.count > 1 then
			(
				keys[keys2.count] = undefined
				for i = 1 to keys2.count do
				(
					if i != 1 then
						if (CheckLock keys2[i-1].value keys2[i].value) then
							reportError "TCB" o.name keys2[i].time
					-----
					if i == 1 then tangents = ComputeTangents undefined keys2[i] keys2[i+1] else \
					if i == keys2.count then tangents = ComputeTangents keys2[i-1] keys2[i] undefined else \
					tangents = ComputeTangents keys2[i-1] keys2[i] keys2[i+1]
					-----
					keys[i] = (SmoothKeyData keys2[i].time keys2[i].value tangents.val1 tangents.val2)
				)
			)
			else if keys2.count == 1 then keys[1] (TanKey keys2[1].time keys2[1].value keys2[1].value keys2[1].value)
		)
		bezier_rotation: ( --smooth rotation, Now exported :D.
			local tangent = quat 0 0 0 1
			local k = c.keys
			if k.count != 0 then keys[k.count] = undefined
			for i = 1 to k.count do
			(
				local v = GetRot k[i].time o r pr hasParent
				keys[i] = (SmoothKeyData k[i].time v undefined undefined)
			)
			if keys.count > 1 then
			(
				for k = 1 to (keys.count) do
				(
					if k != 1 then
					(
						if (CheckLock keys[k-1].value keys[k].value) == true then
						(
							reportError "TCB" o.name keys[k].time
						)
					)
					if k == 1 then tangent  = (GetTangentSmooth keys[k].value keys[k].value keys[k+1].value) else
					(
						if k == keys.count then tangent  = (GetTangentSmooth keys[k-1].value keys[k].value keys[k-1].value) else \
						tangent  = (GetTangentSmooth keys[k-1].value keys[k].value keys[k+1].value)
					)
					keys[k].intan = tangent
					keys[k].outtan = tangent
				)
			)
			else
			(
				keys[1].intan = keys[1].value
				keys[1].outtan = keys[1].value
			)
		)
		default: (
			local k = c.keys
			if k.count != 0 then keys[k.count] = undefined
			for i = 1 to k.count do 
			(
				local v = GetRot k[i].time o r pr hasParent
				keys[i] = (KeyData k[i].time v)
			)
		)
	)
	keys
)

fn RotationType controller =
(
	local type
	case (classOf controller) of
	(
		bezier_rotation: type = #Hermite
		tcb_rotation: type = #Hermite
		linear_rotation: type =#Linear
		default: type = #NonInterp
	)
	type
)

fn RotationKeysBiped o =
(
	local keys = #()
	local keys2 = #()
	local tangents
	
	--Igni - Added support for biped
	-- If we are dealing with a biped object we need to do the following
	-- use the Tension, Continuity and Bias from the key but use the rotation at the time of the key as read directly from the bone's orientation.

	-- get the keys from the biped controller and match the orignal code
	if (hasProperty o.transform.controller "turning") then
	(
		k = o.transform.controller.turning.controller.keys
	)
	else
	(
		k = o.transform.controller.keys
	)
	
	local r
	
	r = (biped.getTransform o #rotation)
	
	local pr = (quat 0 0 0 1)
	if o.parent != undefined then
	(
		if classOf o.parent == Biped_Object then
		(
			pr = (biped.getTransform o.parent #rotation)
		)
		else
		(
			pr = (inverse o.parent.rotation)
		)
	)
	
	for i = 1 to k.count do
	(
		local v = GetRotBiped k[i].time o r pr
		
		local biped_k
		if (hasProperty o.transform.controller "turning") then
		(
			biped_k = biped.getKey o.transform.controller.turning.controller i
		)
		else
		(
			biped_k = biped.getKey o.transform.controller i
		)
		
		append keys2 (TCBKeys k[i].time v ((biped_k.tension - 25.0)/25.0) ((biped_k.continuity - 25.0)/25.0) ((biped_k.bias - 25.0)/25.0) )
	)
	if keys2.count > 1 then
	(
		for i = 1 to keys2.count do
		(
			if i != 1 then
				if (CheckLock keys2[i-1].value keys2[i].value) then
					reportError "TCB" o.name keys2[i].time
			-----
			if i == 1 then tangents = ComputeTangents undefined keys2[i] keys2[i+1] else \
			if i == keys2.count then tangents = ComputeTangents keys2[i-1] keys2[i] undefined else \
			tangents = ComputeTangents keys2[i-1] keys2[i] keys2[i+1]
			-----
			append keys (SmoothKeyData keys2[i].time keys2[i].value tangents.val1 tangents.val2)
		)
	)
	else if keys2.count == 1 then append keys (SmoothKeyData keys2[1].time keys2[1].value keys2[1].value keys2[1].value)
	keys
)

fn GetRotIK t o r pr hp =
(
	at time t local v = (inverse o.transform.rotationPart)*r -- 
	if hp then at time t  v = (v/((inverse o.parent.transform.rotationPart)*pr))
	v
)

fn RotationIKKeys o =
(
	local tmp = #()
	local hasParent = o.parent != undefined 
	local r = at time 0f o.transform.rotationPart
	local pr = (quat 0 0 0 1)
	if hasParent then pr = at time 0f o.parent.transform.rotationPart
	if (isIKAffected o) then
	(
		local chains = getIKId o
		for i in chains do
			join tmp IKChainsKeys[i]
	)
	if (isObjectConstrained o) then join tmp (getObjectConstraintKeys o)
	tmp = ArrayMakeUnique tmp
	sort tmp
	local keys = #()
	tmp = getKeysOnSequences2 tmp Sequences inclusive:false
	local j = 1
	for i = 1 to tmp.count do
	(
		if tmp[i].count != 0 then
		(
			local s = Sequences[i]
			local v =  GetRotIK s.start o r pr hasParent
			keys[j] = (KeyData s.start v)
			j += 1
			for k in tmp[i] do
			(
				v = GetRotIK k o r pr hasParent
				keys[j] = (KeyData k v)
				j += 1
			)
			v = GetRotIK s.end o r pr hasParent
			keys[j] = (KeyData s.end v)
			j += 1
		)
	)
	keys
)

---------------------------------------------------------------------------------------------------------------------------------
-- 												Scaling											  						  ---
---------------------------------------------------------------------------------------------------------------------------------

fn ScaleKeys o =
(
	local keys = #()
	local c = o.scale.controller
	local p = o.scale
	p = at time 0f in coordsys parent o.scale
	case (classOf c) of
	(
		bezier_scale: (
			keys2 =#()
			k = c.keys
			keys2[k.count] = undefined
			for i = 1 to k.count do
			(
				at time k[i].time local v = in coordsys parent o.scale
				v = v / p
				local in1 
				if i != 1 then
					at time  (k[i-1].time + 2*(k[i].time - k[i-1].time)/3) in coordsys parent in1 = o.scale / p
				else
					in1 = v
				local out1
				if i != k.count then
					at time  (k[i].time + (k[i+1].time - k[i].time)/3) in coordsys parent out1 = o.scale / p
				else
					out1 = v
				keys2[i] = (SmoothKeyData k[i].time v in1 out1)
			)
			keys = ProcessBezier keys2
		)
		tcb_scale: (
			append keys (SmoothKeyData 0f [0,0,0] undefined undefined)
			for k in c.keys do 
			(
				local v = at time k.time in coordsys parent o.scale
				v = v/p
				append keys (SmoothKeyData k.time v undefined undefined)
			)
			append keys (SmoothKeyData 999999999f [0,0,0] undefined undefined)
			for k = 1 to (c.keys.count) do
			(
				local t = (c.keys[k].tension - 25) / 25
				local c_ = (c.keys[k].continuity - 25) / 25				-- "c" is taken by the position controller
				local b = (c.keys[k].bias - 25) / 25
				keys[k+1].intan = (InTan keys[k].value keys[k+1].value keys[k+2].value t c_ b)
				keys[k+1].outtan = (OutTan keys[k].value keys[k+1].value keys[k+2].value t c_ b)
			)
			deleteItem keys (keys.count)
			deleteItem keys 1
		)
		default: (
			for k in c.keys do 
			(
				local v = at time k.time in coordsys parent o.scale
				v = v/p
				append keys (KeyData k.time v)
			)
		)
	)
	keys
)

fn ScaleType controller =
(
	local type
	case (classOf controller) of
	(
		bezier_scale: 
		(
			if (controllerIsNonInterp controller) then
				type = #NonInterp
			else
				type = #Bezier
		)
		tcb_scale: type = #Hermite
		linear_scale: type =#Linear
		default: type = #NonInterp
	)
	type
)

---------------------------------------------------------------------------------------------------------------------------------
-- 												Float Controllers											                      ---
---------------------------------------------------------------------------------------------------------------------------------

fn FloatKeys c =
(
	local keys = #()
	case (classOf c) of
	(
		bezier_float: (
			keys2 =#()
			k = c.keys
			keys2[k.count] = undefined
			for i = 1 to k.count do
			(
				at time k[i].time local v = c.value
				local in1
				if i != 1 then
					at time  (k[i-1].time + 2*(k[i].time - k[i-1].time)/3) in1 = c.value
				else
					in1 = v
				local out1
				if i != k.count then
					at time  (k[i].time + (k[i+1].time - k[i].time)/3) out1 = c.value
				else
					out1 = v
				keys2[i] = (SmoothKeyData k[i].time v in1 out1)
			)
			keys = ProcessBezier keys2
		)
		tcb_float: (
			append keys (SmoothKeyData 0f 0.0 undefined undefined)
			for k in c.keys do 
			(
				append keys (SmoothKeyData k.time k.value undefined undefined)
			)
			append keys (SmoothKeyData 999999999f 0.0 undefined undefined)
			for k = 1 to (c.keys.count) do
			(
				local t = (c.keys[k].tension - 25) / 25
				local c_ = (c.keys[k].continuity - 25) / 25
				local b = (c.keys[k].bias - 25) / 25
				keys[k+1].intan = (InTan keys[k].value keys[k+1].value keys[k+2].value t c_ b)
				keys[k+1].outtan = (OutTan keys[k].value keys[k+1].value keys[k+2].value t c_ b)
			)
			deleteItem keys (keys.count)
			deleteItem keys 1
		)
		default: (
			for k in c.keys do append keys (KeyData k.time k.value)
		)
	)
	keys
)

fn FloatType controller =
(
	local type
	case (classOf controller) of
	(
		bezier_float: 
		(
			if (controllerIsNonInterp controller) then
				type = #NonInterp
			else
				type = #Bezier
		)
		tcb_float: type = #Hermite
		linear_float: type =#Linear
		default: type = #NonInterp
	)
	type
)

---------------------------------------------------------------------------------------------------------------------------------
-- 												Visibility											  						  ---
---------------------------------------------------------------------------------------------------------------------------------

fn VisibilityKeys obj =
(
	local keys = #()
	local c = obj.visibility.controller
	case (classOf c) of
	(
		On_Off:
		(
			if Sequences[1] != undefined then
			(
				local contkeys = c.keys
				local i = ArrayFindItemEx Sequences contkeys[1].time (fn tmp a b = b < a.end)
				if (i != 0) then
				(
					local top = Sequences.count
					local j = 1
					local lastKey = contkeys.count
					while (i <= top) do
					(
						local state = false
						local end = Sequences[i].end, start = Sequences[i].start
						local vis
						at time start vis = obj.visibility
						append keys (KeyData start vis)
						j = ArrayFindItemEx contkeys end (fn tmp a b = a.time < b)
						while (j != 0 and j <= lastKey and contkeys[j].time < end) do
						(
							if (contkeys[j].time > start) then
							(
								state = true
								if vis then vis = false else vis = true
								append keys (KeyData contkeys[j].time vis)
							)
							j = j +1
						)
						if j <= lastKey and contkeys[j].time == end then state = true
						if state then
						(
							at time end vis = obj.visibility
							append keys (KeyData end vis)
						)
						i = i + 1
					)
					keys = ArrayMap keys (fn tmp a = (KeyData a.time (B2I a.value)))
				)
			)
		)
		default: (
			keys = FloatKeys c
		)
	)
	keys
)

fn VisibilityType controller = FloatType controller

---------------------------------------------------------------------------------------------------------------------------------
-- 												Color & Texture										  			           ---
---------------------------------------------------------------------------------------------------------------------------------

--- Color
fn ColorKeys c =
(
	local keys = #()
	case (classOf c) of
	(
		bezier_position: (
			local keys2 =#()
			k = c.keys
			keys2[k.count] = undefined
			for i = 1 to k.count do
			(
				at time k[i].time local v = k[i].value
				local in1 
				if i != 1 then
					at time  (k[i-1].time + 2*(k[i].time - k[i-1].time)/3) in1 = c.value
				else
					in1 = v
				local out1
				if i != k.count then
					at time  (k[i].time + (k[i+1].time - k[i].time)/3) out1 = c.value
				else
					out1 = v
				keys2[i] = (SmoothKeyData k[i].time v in1 out1)
			)
			keys = ProcessBezier keys2
		)
		tcb_point3: (
			append keys (SmoothKeyData 0f [0,0,0] undefined undefined)
			for k in c.keys do 
			(
				append keys (SmoothKeyData k.time k.value undefined undefined)
			)
			append keys (SmoothKeyData 999999999f [0,0,0] undefined undefined)
			for k = 1 to (c.keys.count) do
			(
				local t = (c.keys[k].tension - 25) / 25
				local c_ = (c.keys[k].continuity - 25) / 25
				local b = (c.keys[k].bias - 25) / 25
				keys[k+1].intan = (InTan keys[k].value keys[k+1].value keys[k+2].value t c_ b)
				keys[k+1].outtan = (OutTan keys[k].value keys[k+1].value keys[k+2].value t c_ b)
			)
			deleteItem keys (keys.count)
			deleteItem keys 1
		)
		default: (
			for k in c.keys do 
			(
				append keys (KeyData k.time k.value)
			)
		)
	)
	keys
)

fn ColorType controller =
(
	local type
	case (classOf controller) of
	(
		bezier_color: 
		(
			if (controllerIsNonInterp controller) then
				type = #NonInterp
			else
				type = #Bezier
		)
		tcb_point3: type = #Hermite
		default: type = #Linear
	)
	type
)

fn GetTextureAnimation idsList playback =
(
	local n = round (1/playback)
	local m = 0
	local result = #()
	if (n == 0) then n = 1
	for i in idsList do
	(
		append result (KeyData (m as time) i)
		m += n
	)
	local animData = AnimationData ()
	animData.keys = result
	animData.type = #NonInterp
	animData.globalSequenceId = LoadGlobalSequence animData.keys[animData.keys.count].time
	ParameterValue #Animated animData
)

---------------------------------------------------------------------------------------------------------------------------------
-- 												MainFunctions										  					  ---
---------------------------------------------------------------------------------------------------------------------------------
global SequencesOriginal

fn checkStartEndKeys keys objName controllerName =
(
	local errName = objName + "->" + controllerName
	local resampledKeys = getKeysOnSequences keys Sequences inclusive:true
	local count = SequencesOriginal.count
	for i = 1 to count do
	(
		local tmp = resampledKeys[i]
		local err = (tmp.count == 0 or tmp[1].time != Sequences[i].start)
		if tmp.count > 1 then
			err = err or (tmp[tmp.count].time != Sequences[i].end)
		if err then
			reportError "StartEndKeys" errName SequencesOriginal[i].name
	)
)

fn GetAnimationData obj controller &processFunc &typeFunc checkSE:true =
(
	local animData = AnimationData ()
	animData.keys = processFunc obj
	animData.type = typeFunc controller
	local usesGlobalSequences = getAfterORT controller == #cycle
	if usesGlobalSequences then animData.globalSequenceId = LoadGlobalSequence animData.keys[animData.keys.count].time
	if (animData.keys.count != 0) then
	(
		local s
		fn getName t = if isProperty t #name then t.name else t as string
		s = getName obj
		if checkSE and (not usesGlobalSequences) then checkStartEndKeys animData.keys s (controller as string)
		animData
	)
	else undefined
)

fn GetAnimationDataBiped obj &processFunc checkSE:true =
(
	local animData = AnimationData ()
	animData.keys = processFunc obj
	animData.type = #Hermite
	if (animData.keys.count != 0) then
	(
		if checkSE then checkStartEndKeys animData.keys obj.name "Biped Controller"
		animData
	)
	else undefined
)

fn GetTransformations obj =
(
	local trans = TransformationData ()
	if (classOf obj != Biped_Object) then
	(
		if obj.pos.isAnimated then trans.translation = GetAnimationData obj obj.pos.controller PositionKeys PositionType
		if  ((isIKAffected obj) or (isObjectConstrained obj)) then
			trans.rotation = GetAnimationData obj obj.rotation.controller RotationIKKeys (fn tmp a = #linear)
		else
			if obj.rotation.isAnimated then trans.rotation = GetAnimationData obj obj.rotation.controller RotationKeys RotationType
		if obj.scale.isAnimated then trans.scale = GetAnimationData obj obj.scale.controller ScaleKeys ScaleType
	)
	else
	(
		trans.translation = GetAnimationDataBiped obj PositionKeysBiped
		trans.rotation = GetAnimationDataBiped obj RotationKeysBiped
	)
	trans
)

fn _ProcessSingleValue controller val &ProcessFunc &TypeFunc =
(
	local ParamValue = ParameterValue ()
	if (controller != undefined) then
	(
		ParamValue.type = #Animated
		ParamValue.data = GetAnimationData controller controller ProcessFunc TypeFunc checkSE:false
		if (ParamValue.data == undefined) then
		(
			ParamValue.type = #Static
			ParamValue.data = val
		)
	)
	else
	(
		ParamValue.type = #Static
		ParamValue.data = val
	)
	ParamValue
)

fn FloatParameterProcess controller val =
(
	_ProcessSingleValue controller val FloatKeys FloatType
)

fn ColorParameterProcess controller val =
(
	_ProcessSingleValue controller val ColorKeys ColorType
)

fn VisibilityParameterProcess obj =
(
	local controller = obj.visibility.controller
	local ParamValue = ParameterValue ()
	if (controller != undefined) then
	(
		local ParamValue = ParameterValue ()
		ParamValue.type = #Animated
		ParamValue.data = GetAnimationData obj controller VisibilityKeys VisibilityType checkSE:(classof controller != ON_OFF)
		if (ParamValue.data == undefined) then ParamValue = undefined
	) else ParamValue = undefined
	ParamValue
)

fn EventProcess obj =
(
	local keys
	if hasNoteTracks obj then
	(
		local track = getNoteTrack obj 1
		keys = #()
		for k in track.keys do append keys k.time
		if (keys.count == 0) then keys = undefined
	)
	keys
)

fn GetGlobalSequences = GlobalSequencesGathered

fn IKRefineKeys o &arr t1 t2 =
(
	local d1 = (getGoalDistance o t1)
	local d2 = (getGoalDistance o t2)
	local maxDist = if d1 > d2 then d1 else d2
	local d = (abs (d2 - d1))/maxDist
	if (d > 0.25) or (getAngleofGoalChange o t1 t2) > 30.0 then
	(
		local th = t1 + (t2-t1)*0.5
		append arr th
		IKRefineKeys o arr t1 th
		IKRefineKeys o arr th t2
	)
)

fn GatherIKChains =
(
	fn getBoneChains start end =
	(
		result = #(end)
		end = end.parent
		while (end != undefined and end != start) do
		(
			append result end
			end = end.parent
		)
		append result start
		result
	)
	IKChains = #()
	IKChainsKeys = #()
	IKChainsBones = #()
	fn checkHILimb o = (classof o == IK_Chain_Object) and (classof o.controller == IKChainControl)
	fn checkSpline o = (classof o == IK_Chain_Object) and (classof o.controller == SplineIKChain)
	for o in Helpers where checkHILimb o do
	(
		append IKChains o
		local target = o.controller.VHTarget
		local tmp = #(o)
		join tmp (GetParents o)
		local topBone = o.controller.startjoint.parent
		if topBone != undefined then
		(
			local cycle = #(topBone)
			join cycle (getParents topBone)
			local out = true
			local i = 1
			while (i <= cycle.count and out) do
			(
				local index = findItem tmp cycle[i]
				if index == 0 then
					append tmp cycle[i]
				else
				(
					deleteItem tmp index
					out = false
				)
				i += 1
			)
		)
		if target != undefined and o.controller.useVHTarget then
		(
			append tmp target
			join tmp (GetParents target)
		)
		tmp = ArrayMakeUnique tmp
		local keys = #()
		for s in tmp where s != undefined do
			join keys (getKeysFromObject s)
		keys = ArrayMakeUnique keys
		sort keys
		local extraKeys = #()
		for i = 2 to keys.count do
			IKRefineKeys o &extraKeys  keys[i-1] keys[i]
		join keys extraKeys
		sort keys
		IKChainsKeys[IKChains.count] = keys
		IKChainsBones[IKChains.count] = getBoneChains o.controller.startjoint o.controller.endjoint
	)
	for o in helpers where checkSpline o  do
	(
		append IKChains o
		local tmp = #(o)
		local cpoints = o.controller.pickShape.modifiers[#Spline_IK_Control].helper_list
		join tmp cpoints
		local topBone = if cpoints.count > 0 then cpoints[1].parent else undefined
		if topBone != undefined then
		(
			local cycle = #(topBone)
			join cycle (getParents topBone)
			local out = true
			local i = 1
			while (i <= cycle.count and out) do
			(
				local index = findItem tmp cycle[i]
				if index == 0 then
					append tmp cycle[i]
				else
				(
					deleteItem tmp index
					out = false
				)
				i += 1
			)
		)
		tmp = ArrayMakeUnique tmp
		local keys = #()
		for s in tmp where s != undefined do
			join keys (getKeysFromObject s)
		keys = ArrayMakeUnique keys
		sort keys
		local extraKeys = #()
		for i = 2 to keys.count do
			IKRefineKeys o &extraKeys  keys[i-1] keys[i]
		join keys extraKeys
		sort keys
		IKChainsKeys[IKChains.count] = keys
		IKChainsBones[IKChains.count] = getBoneChains o.controller.startjoint o.controller.endjoint
	)
)

fn isSmoothType t = ((t == #Bezier) or (t == #TCB))

fn combineTVAnimation c1 c2 c3 =
(
	fn getTimes c =
	(
		result = #()
		local n = c.keys.count
		if (n > 0) then result[n] = undefined
		for i = 1 to n do
			result[i] = c.keys[i].time
		result
	)
	fn produceKeys times isSmooth =
	(
		result = #()
		result[times.count]
		p = [0,0,0]
		for i = 1 to times.count do
			if isSmooth then
				result[i] = SmoothKeyData times[i] (copy p) (copy p) (copy p)
			else
				result[i] = KeyData times[i] (copy p)
		result
	)
	fn saveKeyData keys ct isSmooth index =
	(
		fn getControllerVal c t = at time t (c.value)
		if isSmooth then
		(
			k = keys
			for i = 1 to k.count do
			(
				local t = k[i].time
				local in1 
				local out1
				local v =  getControllerVal ct t
				if i != 1 then in1 = getControllerVal ct (k[i-1].time + 2*(t - k[i-1].time)/3) else in1 = v
				if i != k.count then out1 = getControllerVal ct (k[i].time + (k[i+1].time - t)/3) else out1 = v
				k[i].value[index] = v
				k[i].intan[index] = in1
				k[i].outtan[index] = out1
			)
		) else (
			k = keys
			for i = 1 to k.count do
			(
				local v =  getControllerVal ct k[i].time
				k[i].value[index] = v
			)
		)
	)
	fn isSmoothGet t1 t2 t3 = ((isSmoothType t1) or (isSmoothType t2) or (isSmoothType t3))
	local t1 = FloatType c1
	local t2 = FloatType c2
	local t3 = FloatType c3
	if c1 == undefined then c1 = linear_float ()
	if c2 == undefined then c2 = linear_float ()
	if c3 == undefined then c3 = linear_float ()
	local t = getTimes c1
	join t (getTimes c2)
	join t (getTimes c3)
	sort t
	OptimizeArray t
	local ism = (isSmoothGet c1 c2 c3)
	local keys = produceKeys t ism
	saveKeyData keys c1 ism 1
	saveKeyData keys c2 ism 2
	saveKeyData keys c3 ism 3
	if ism then keys = ProcessBezier keys
	local animData = AnimationData ()
	animData.keys = keys
	if (t1 == #NonInterp and t2 == #NonInterp and t3 == #NonInterp) then 
		animData.type = #NonInterp
	else
		animData.type = (if ism then #Bezier else #Linear)
	local usesGlobalSequences = (getAfterORT c1 == #cycle or getAfterORT c2 == #cycle or getAfterORT c3 == #cycle)
	if usesGlobalSequences then animData.globalSequenceId = LoadGlobalSequence animData.keys[animData.keys.count].time
	animData
)

fn InitiateAnimationData s s2 =
(
	Sequences = s
	SequencesOriginal = s2
	GlobalSequencesGathered = #()
	GlobalSequenceCount = 0
	GatherIKChains ()
)