
macroscript NeoDex_Animation_Tools
    buttonText:"Animation Tools"
	category:"NeoDex Toolkit"
	internalCategory:"Tools" 
(
    local neoDexAHVersion = "1.0"
	-- Includes
	include "NeoDexModules\\UtilityFunctions.ms"
	-- end Includes
	
	mapped fn setContinuityToZero o time &controllerFunc =
	(
		local c = controllerFunc o
		if (c != undefined) then
		(
			local index = ArrayFindItemEx c.keys time (fn tmp a b = a.time == b)
			if (index != 0) then c.keys[index].continuity = 0
		)
	)
	
	fn GetRot t o r pr hp =
	(
		at time t local v = o.rotation*r -- 
		if hp then at time t  v = (v/(o.parent.rotation*pr))
		v
	)
	mapped fn fixSquadRotation s =
	(
		local c = s.rotation.controller
		local b = (classOf c == tcb_rotation or classOf c == bezier_rotation)
		if (b and c.keys.count > 1) then
		(
			local faults = #()
			local keys = c.keys
			local hasParent = s.parent != undefined and (classOf s.parent != Uvwmap) -- This last part is for UVW Gizmo
			local r = (inverse s.rotation)
			local pr = (quat 0 0 0 1)
			if hasParent then pr = (inverse s.parent.rotation)
			for i = 2 to keys.count do
			(
				local v = GetRot keys[i-1].time s r pr hasParent
				local w = GetRot keys[i].time s r pr hasParent
				if (CheckLock v w) then
					append faults (interval keys[i-1].time keys[i].time )
			)
			for f in faults do addNewKey c  ((f.start+f.end)*0.5)
		)
	)
	global sequences
	fn createStartEndkeys sel =
	(
		fn tmp c =
		(
			local superKeys = getKeysOnSequences c.keys sequences
			local count = sequences.count
			for i = 1 to count do
			(
				addNewKey c sequences[i].start
				if (superKeys[i].count > 0) then
					addNewKey c sequences[i].end
			)
		)
		for s in sel do
		(
			if s.pos.IsAnimated then tmp s.pos.controller
			if s.rotation.IsAnimated then tmp s.rotation.controller
			if s.scale.IsAnimated then tmp s.scale.controller
		)
	)

	rollout generalRoll "General Tools" width:162 height:351
	(
		GroupBox grp5 "TCB Tools" pos:[5,109] width:149 height:135
		checkbox PosChk "Position" pos:[20,137] width:110 height:18
		checkbox RotChk "Rotation" pos:[20,158] width:100 height:20
		checkbox ScaChk "Scaling" pos:[20,181] width:65 height:19
		button continuityBtn "Set Continuity to Zero" pos:[19,206] width:113 height:23
		GroupBox grp2 "Animation" pos:[3,6] width:150 height:96
		button addSEkeysbtn "Do it!" pos:[58,69] width:70 height:23
		label lbl1 "Add start and ending keys to selected objects:" pos:[17,23] width:116 height:46
		label lbl2 "Fix non-linear rotation problems on selection:" pos:[22,268] width:105 height:33
		button fixSquadbtn "Do it!" pos:[22,305] width:63 height:24
		groupBox grp3 "Smooth Rotations" pos:[9,250] width:131 height:89
		button SquadAboutBtn "?" pos:[100,307] width:21 height:20
		on continuityBtn pressed do
		(
			fn PosController o =
			(
				local c = o.pos.controller
				if classOf c != tcb_position then c = undefined
				c
			)
			fn RotController o =
			(
				local c = o.rotation.controller
				if classOf c != tcb_rotation then c = undefined
				c
			)
			fn ScaController o =
			(
				local c = o.scale.controller
				if classOf c != tcb_scale then c = undefined
				c
			)
			if PosChk.checked then
				setContinuityToZero selection currentTime PosController
			if RotChk.checked then
				setContinuityToZero selection currentTime RotController
			if ScaChk.checked then
				setContinuityToZero selection currentTime ScaController
		)
		on addSEkeysbtn pressed do
		(
			sequences = getAnimationIntervals ()
			createStartEndkeys selection
		)
		on fixSquadbtn pressed do
			fixSquadRotation selection
		on SquadAboutBtn pressed do
		(
			messageBox (
				"Non-linear rotation keys may animate improperly in wc3 if they go beyond the 90.\n" +
				"This little tool will apply fixes to those keys by creating a new key between them.\n" +
				"Note that you may have to run the tool more than once in some weird occasions where you created rotations of over 180"
			)
		)
	)

	mapped fn saveParents s =
	(
		local p = s.parent
		if p != undefined then setUserProp s "OriginalParent" p.name
		s.parent = undefined
	)

	mapped fn restoreParents s =
	(
		local tmp = getUserProp s "OriginalParent"
		if tmp != undefined then
		(
			local p = getNodeByName tmp
			s.parent = p
		)
	)
	
	rollout skinRoll "Skinning" width:162 height:163
	(
		groupBox grp1 "AT Skinning" pos:[8,9] width:145 height:136
		label lbl1 "Saving and restoring parents." pos:[20,35] width:114 height:33
		button saveBtn "Save and Unlink" pos:[18,77] width:89 height:26
		button restoreBtn "Restore Parents" pos:[20,107] width:86 height:22
		button aboutBtn "?" pos:[117,111] width:25 height:17
		on saveBtn pressed do
			saveParents selection
		on restoreBtn pressed do
			restoreParents selection
		on aboutBtn pressed do
		(
			messageBox (
				"Sometimes you already got a bunch of meshes linked to some bones and suddenly you want to edit the bones." +
				" If you don't want to mess up the meshes, you must unlink them do whatever you need, then relink.\n" +
				"This tool saves the parents of the selected meshes, unlinks them and restores them when you want.\n" +
				"Use this tool before converting bones from IK to FK, for instance."
			)
		)
	)

	mapped fn applyFunc node &func2 =
	(
		func2 node
		applyFunc node.children func2
	)
	
	fn applyFuncToSelection &func =
	(
		fn isselected sel nod =
		(
			for s in sel do
				if nod == s then return true
			false
		)
		rootnodes = #()
		for s in selection do
			if s.parent == undefined or isselected selection s.parent then
			append rootnodes s
		applyFunc rootnodes func
	)

	rotationFuncs = #()
	rotationFuncs[1] = (fn func a = a.rotation.controller = linear_rotation())
	rotationFuncs[2] = (fn func a = a.rotation.controller = bezier_rotation())
	rotationFuncs[3] = (fn func a = a.rotation.controller = tcb_rotation())
	rotationFuncs[4] = (fn func a = OK)
	positionFuncs = #()
	positionFuncs[1] = (fn func a = a.pos.controller = linear_position())
	positionFuncs[2] = (fn func a = a.pos.controller = bezier_position())
	positionFuncs[3] = (fn func a = a.pos.controller = tcb_position())
	positionFuncs[4] = (fn func a = OK)
	scaleFuncs = #()
	scaleFuncs[1] = (fn func a = a.scale.controller = linear_scale())
	scaleFuncs[2] = (fn func a = a.scale.controller = bezier_scale())
	scaleFuncs[3] = (fn func a = a.scale.controller = tcb_scale())
	scaleFuncs[4] = (fn func a = OK)
	
	rollout controllersRoll "Controllers" width:162 height:197
	(
		GroupBox grp3 "Controllers" pos:[3,9] width:151 height:325
		radiobuttons PosRdo "Position:" pos:[24,23] width:53 height:78 labels:#("Linear", "Bezier", "TCB", "Ignore") default:4
		radiobuttons RotRdo "Rotation:" pos:[24,110] width:59 height:78 labels:#("Linear", "Smooth", "TCB", "Ignore") default:4
		radiobuttons ScaRdo "Scaling" pos:[24,205] width:53 height:78 labels:#("Linear", "Bezier", "TCB", "Ignore") default:4
		button controllerBtn "Apply to selected" pos:[14,298] width:121 height:25
		on controllerBtn pressed do
		(
			applyFuncToSelection rotationFuncs[RotRdo.state]
			applyFuncToSelection positionFuncs[PosRdo.state]
			applyFuncToSelection scaleFuncs[ScaRdo.state]
		)
	)

	Main = newrolloutfloater ("NeoDex Animation Tools" + neoDexAHVersion) 200 400
	addrollout generalRoll Main
	addrollout skinRoll Main
	addrollout controllersRoll Main
	
)
