
macroscript NeoDex_Skin_Tools
    buttonText:"Skinning Tools"
    category:"NeoDex Toolkit"
	internalCategory:"Tools" 
(
	include "NeoDexModules\\UtilityFunctions.ms"
	include "NeoDexModules\\GeometryUtilities.ms"
	include "NeoDexModules\\SkinUtilities.ms"
	
	local neoDexSTVersion = ".50"
	
	fn IsSkinned o = (o.modifiers[#Skin] != undefined)

	struct MeshUvwmap
	(
		tvfaces,
		tvertices,
		channel,
		ex
	)

	fn produceMap m faces map =
	(
		local result = MeshUvwmap ()
		result.channel = map
		result.tvfaces = getTVFaces m map
		local tmp = #()
		local remapper = #()
		local num = meshop.getNumMapVerts m map
		local func = meshop.getMapVert
		local tmp = #()
		tmp[getnumVerts m] = undefined
		local j = 0
		for i = 1 to num do
		(
			j += 1
			tmp[j] = func m map i
			remapper[i] = j
		)
		local tvertices = #()
		tvertices[j] = undefined
		for i = 1 to j do tvertices[i] = tmp[i]
		result.tvertices = tvertices
		for f in result.tvfaces do
			for i = 1 to 3 do f[i] = remapper[f[i]]
		result
	)

	struct SkinObject
	(
		bone,
		vertices,
		normals,
		smoothgroups,
		maps,
		faces,
		matrixIndices,
		material,
		matrices,
		mesh
	)

	fn LoadSkinObject m =
	(
		obj = SkinObject ()
		if m.parent != undefined then obj.bone = m.parent else throw "AT Skinned without a bone"
		obj.matrices = #(#(1))
		obj.vertices = #()
		obj.normals = #()
		local numVerts = getNumVerts m
		obj.vertices[numVerts] = undefined
		obj.normals[numVerts] = undefined
		for i = 1 to numVerts do
		(
			obj.vertices[i] = getVert m i
			obj.normals[i] = getNormal m i
		)
		obj.faces = getFaces m
		obj.maps = #()
		for i = 1 to (meshop.getNumMaps m) do
		(
			if (i > 99) then exit
			if (meshop.getMapSupport m i) then
				append obj.maps (produceMap m obj.faces i)
		)
		local numFaces = obj.faces.count
		obj.smoothgroups = #()
		obj.smoothgroups[numFaces] = undefined
		for i = 1 to NumFaces do obj.smoothgroups[i] = getFaceSmoothGroup m i
		obj.matrixIndices = ArrayRepeat numVerts 0
		obj.material = m.material
		obj.mesh = m
		obj
	)

	fn mergeMaps map1 map2 mapNum =
	(
		local result = MeshUvwmap ()
		local tvertices = #()
		join tvertices map1.tvertices
		join tvertices map2.tvertices
		local tvfaces = #()
		join tvfaces map1.tvfaces
		local maxVert = map1.tvertices.count
		for f in map2.tvfaces do
			append tvfaces (f + maxVert)
		result.tvertices = tvertices
		result.tvfaces = tvfaces
		result.channel = mapNum
		result
	)

	fn mergeGeosets geo1 geo2 =
	(
		--Returns new index or current index
		fn expandMatrix &matrix newValue =
		(
			local found = ArrayFindItem matrix newValue
			if found == 0 then
			(
				append matrix newValue
				found = matrix.count - 1
			) else found = found - 1
			found
		)
		local matrices = deepCopy geo1.matrices
		local newGeo = SkinObject ()
		local RefittedIndices = #()
		RefittedIndices[geo2.matrixindices.count] = undefined
		for j = 1 to  geo2.matrices.count do
		(
			local oldId = j - 1
			local newId = expandMatrix matrices geo2.matrices[j]
			for i = 1 to RefittedIndices.count do
				if geo2.matrixindices[i] == oldId then
					RefittedIndices[i] = newId
		)
		local faces = deepCopy geo1.faces
		local maxVert = geo1.vertices.count
		for f in geo2.faces do 
			append faces (f + maxVert)
		newGeo.vertices = #()
		join newGeo.vertices geo1.vertices
		join newGeo.vertices geo2.vertices
		newGeo.normals = #()
		join newGeo.normals geo1.normals
		join newGeo.normals geo2.normals
		local maxMap = geo1.maps.count
		if maxMap < geo2.maps.count then maxVert = geo2.maps.count
		local maxer1 = geo1.maps.count +1, maxer2 = geo2.maps.count + 1
		local maps = #()
		maps[maxMap] = undefined
		for i = 1 to maxMap do 
			maps[i] = (mergeMaps geo1.maps[mod i maxer1] geo2.maps[mod i maxer2] i)
		local smoothgroups = #()
		join smoothgroups geo1.smoothgroups
		join smoothgroups geo2.smoothgroups
		newGeo.smoothgroups = smoothgroups
		newGeo.maps = maps
		newGeo.matrixindices = #()
		join newGeo.matrixindices geo1.matrixindices
		join newGeo.matrixindices RefittedIndices
		newGeo.faces = faces
		newGeo.matrices = matrices
		newGeo.material = geo1.material
		newGeo
	)

	fn mergeGeoGroups geoGroups =
	(
		local result = #()
		for gp in geoGroups do
		(
			local basicMesh = gp[1]
			for i = 2 to gp.count do
				basicMesh = mergeGeosets basicMesh gp[i]
			append result basicMesh
		)
		result
	)
	
	fn getResettedBones mshs thisBones =
	(
		local result = #()
		for m in mshs do
		(
			local tmpBones = #()
			for n in m.matrices do
				for i = 1 to n.count do
				(
					local b = thisBones[n[i]]
					local index = findItem tmpBones b
					if index == 0 then
					(
						append tmpBones b
						index = tmpBones.count
					)
					n[i] = index
				)
			append result tmpBones
		)
		result
	)

	fn OptimizeVertices geo Thresshold:0.05 =
	(
		fn checkVertices v1 v2 thresshold =
		(
			local v = v1 - v2
			local same =  (v.x <= thresshold and v.x >= -thresshold)
			same = same and (v.y <= thresshold and v.y >= -thresshold)
			same = same and (v.z <= thresshold and v.z >= -thresshold)
			same
		)
		local tmp = #()
		tmp[geo.matrices.count] = undefined
		for i = 1 to geo.vertices.count do
		(
			local matindex = geo.matrixindices[i] + 1
			if tmp[matindex] == undefined then tmp[matindex] = #()
			append tmp[matindex] i
		)
		-- Now we split the subgroups by vertex relation
		for i = 1 to tmp.count do
		(
			local gp = tmp[i]
			local result = #()
			for g in gp do
			(
				local noMatch = true
				for j = 1 to result.count where noMatch do
					if (checkVertices geo.vertices[result[j][1]] geo.vertices[g] Thresshold) then
					(
						noMatch = false
						append result[j] g
					)
				if noMatch then append result #(g)
			)
			tmp[i] = result
		)
		-- now we get mappers and remappers
		local mapper = #()
		local remapper = #()
		remapper[geo.vertices.count] = undefined
		local i = 0
		for t in tmp do
		(
			join mapper t
			for j in t do
			(
				i += 1
				for k in j do
					remapper[k] = i
			)
		)
		#(mapper,remapper)
	)

	fn correlateTvertexVertex faces tvfaces numVerts:undefined =
	(
		local mapper = #()
		if numVerts != undefined then mapper[numverts] = undefined
		for i = 1 to faces.count do
		(
			local tf = tvfaces[i]
			local f = faces[i]
			for j = 1 to 3 do
			(
				if mapper[f[j]] == undefined then 
					mapper[f[j]] = #(tf[j])
				else if (findItem mapper[f[j]] tf[j]) == 0 then
					append mapper[f[j]] tf[j]
			)
		)
		mapper
	)

	fn OptimizeMap map obj Thresshold:0.01 =
	(
		fn checkTVertices tv1 tv2 thresshold =
		(
			local v = tv1 - tv2
			local same =  (v.x <= thresshold and v.x >= -thresshold)
			same = same and (v.y <= thresshold and v.y >= -thresshold)
			same
		)
		local tmp = correlateTvertexVertex obj.faces map.tvfaces numVerts:obj.vertices.count
		local remapper = #(), newTvertices = #()
		remapper[map.tvertices.count] = undefined
		local count = 0
		for i = 1 to tmp.count do
		(
			local tvs = tmp[i]
			local result = #()
			for g in tvs do
			(
				local noMatch = true
				for j = 1 to result.count where noMatch do
					if (checkTVertices map.tvertices[result[j][1]] map.tvertices[g] Thresshold) then
					(
						noMatch = false
						append result[j] g
					)
				if noMatch then append result #(g)
			)
			for r in result do
			(
				count += 1
				local tvertex = [0,0,0]
				for s in r do
				(
					tvertex += map.tvertices[s]
					remapper[s] = count
				)
				append newTvertices (tvertex / (r.count as float))
			)
		)
		map.tvertices = newTvertices
		for f in map.tvfaces do
			for i = 1 to 3 do
				f[i] = remapper[f[i]]
	)

	fn optimizeMesh msh Thresshold:0.05 =
	(
		local tmp = OptimizeVertices msh Thresshold:Thresshold
		local mapper = tmp[1]
		local remapper = tmp[2]
		local newVertices = #()
		local newNormals = #()
		local newMatrixIndices = #()
		newVertices[mapper.count] = undefined
		newNormals[mapper.count] = undefined
		newMatrixIndices[mapper.count] = undefined
		for i = 1 to mapper.count do
		(
			local vertex = [0,0,0]
			local normal = [0,0,0]
			for m in mapper[i] do
			(
				vertex += msh.vertices[m]
				normal += msh.normals[m]
			)
			local diver = mapper[i].count as float
			newVertices[i] = (vertex / diver)
			newNormals[i] = (normal / diver)
			newMatrixIndices[i] = msh.matrixIndices[mapper[i][1]]
		)
		msh.vertices = newVertices
		msh.normals = newNormals
		msh.matrixIndices = newMatrixIndices
		for f in msh.faces do
			for i = 1 to 3 do
				f[i] = remapper[f[i]]
		for m in msh.maps do
			OptimizeMap m msh
		OK
	)

	fn createMesh msh bns ignoreFaceGroups:false =
	(
		fn createVertexMap obj map =
		(
			local chan = map.channel
			local tvertices = map.tvertices
			meshop.setMapSupport obj chan true
			meshop.setNumMapVerts obj chan tvertices.count
			local func = meshop.setMapVert
			for i = 1 to tvertices.count do
				func obj chan i tvertices[i]
			func = meshop.setMapFace
			for i = 1 to map.tvfaces.count do
				func obj chan i map.tvfaces[i]
		)
		local obj = mesh vertices:msh.vertices faces:msh.faces
		-- Normals
		for i = 1 to msh.normals.count do
			setNormal obj i msh.normals[i]
		-- Set Facegroups
		if (not ignoreFaceGroups) then
			for i = 1 to msh.faces.count do
				setFaceSmoothGroup obj i msh.smoothgroups[i]
		-- Load Maps
		for m in msh.maps do
			if m.channel < 2 then createVertexMap obj m
		--  materials
		obj.material = msh.material
		-- Skinning
		select obj
		max modify mode
		local SkinMod = Skin ()
		addModifier obj SkinMod
		modPanel.setCurrentObject SkinMod
		local func = skinOps.addBone
		for b in bns do func SkinMod b 0
		update obj
		max views redraw
		func = skinOps.ReplaceVertexWeights
		for i = 1 to msh.vertices.count do
		(
			local matrix = msh.matrices[msh.matrixIndices[i]+1]
			local weight = (1.0/(matrix.count as float))
			local weights = ArrayRepeat matrix.count weight
			func SkinMod i matrix weights
		)
		max create mode
	)

	fn createFinalMeshes meshes assocBones ignoreFaceGroups:false =
	(
		for i = 1 to meshes.count do
			createMesh meshes[i] assocBones[i] ignoreFaceGroups:ignoreFaceGroups
	)

	fn TransformATSkinToNormalSkin objs:objects Thresshold:0.1 deleteOriginals:false ignoreFaceGroups:false ignoreHidden:true mergeMeshes:true =
	(
		local targets = #()
		local origMeshes = #()
		ignoreHidden = not ignoreHidden
		for o in objs do
			if ((not o.isHidden) or ignoreHidden) and (isMesh o) and ((getWc3Skin o) == #ATSkin) then
			(
				append targets (LoadSkinObject o)
				append origMeshes o
			)
		local geoGroups = ArrayGroup targets (fn tmp a b = (a.material == b.material))
		local numGeos = geoGroups.count
		local j = 0
		local thisBones = #()
		for i = 1 to numGeos do
		(
			for g in geoGroups[i] do
			(
				local index = findItem thisbones g.bone
				if index == 0 then
				(
					j = j + 1
					g.matrices = #(#(j))
					append thisbones g.bone
				)
				else g.matrices = #(#(index))
			)
		)
		local colMatrix = collideMatrix (ArrayMap targets (fn tmp a = a.mesh)) 1
		for i = 1 to targets.count do
			for j = i + 1 to targets.count do
				if colMatrix[i][j] then
					applySyntheticSkin targets[i] targets[j] Thresshold false
		local finalMeshes = mergeGeoGroups geoGroups
		--Reset Bones
		local geoBones = getResettedBones finalMeshes thisBones
		for m in finalMeshes do
			optimizeMesh m Thresshold:Thresshold
		--optimizeMeshes finalMeshes
		createFinalMeshes finalMeshes geoBones ignoreFaceGroups:ignoreFaceGroups useInnerBone:mergeMeshes
		if deleteOriginals then
			for o in origMeshes do delete o
	)
	
	rollout mainRollout "Skin Tools" width:257 height:295
	(
		GroupBox grp1 "At Skinning -> Normal Skinning" pos:[14,14] width:225 height:267
		checkbox chk1 "Ignore Hidden" pos:[38,38] width:95 height:18 checked:true
		checkbox chk2 "Ignore Visibility" pos:[38,63] width:91 height:16 checked:true
		checkbox chk3 "Ignore Smoothgroups" pos:[37,88] width:126 height:20
		GroupBox grp2 "Optimize Meshes" pos:[37,114] width:160 height:67
		checkbox chk5 "Optimize" pos:[50,133] width:129 height:16 enabled:true checked:true
		spinner spn1 "Collide Thresshold" pos:[71,157] width:115 height:16 range:[0,5,0.1] scale:0.01
		checkbox chk6 "Delete old AT Skinned Objects" pos:[34,190] width:166 height:15
		radiobuttons rdo1 "Apply on:" pos:[29,215] width:173 height:30 labels:#("Selection", "All objects") default:2 columns:2
		button btn1 "Convert" pos:[74,253] width:96 height:18
		button btn2 "?" pos:[206,32] width:21 height:16
		on rdo1 changed newState do
			neodexSkinToolsApply = newState
		on btn1 pressed do
		(
			local len = animationRange
			local sliderTmp = sliderTime
			animationRange = interval 0f 100f
			sliderTime = 0f
			--try (
				local objs = if rdo1.state == 2 then objects else selection
				TransformATSkinToNormalSkin objs:objs Thresshold:spn1.value \
					deleteOriginals:chk6.checked ignoreFaceGroups:chk3.checked ignoreHidden:chk1.checked
				messageBox "Done!"
			--)
			--catch (
				--messageBox "Conversion failed, probably you have some unsupported structures :("
				--if (not gmax) then format "*** % ***\n" (getCurrentException()) to:listener
			--)
			animationRange = len
			sliderTime = sliderTmp
		)
		on btn2 pressed do
		(
			messageBox (
				"This tool allows you to convert AT Skinned meshes into unified meshes with skin modifiers.\n" + \
				"This tool will also optimize the linked meshes. It will only recognize meshes which are marked as AT Skinned."
			)
		)
	)
	
	createDialog mainRollout
	mainRollout.rdo1.state = neodexSkinToolsApply
)