
fn RotatePoint p q =
(
	local q2 = quat p.x p.y p.z 0
	q2 = q*q2*(inverse q)
	[q2.x,q2.y,q2.z]
)

fn round a = (ceil ((floor (a*2))/2) as integer)

fn RibbonGenerateMesh r =
(
	fn GenerateVertices t r tf=
	(
		at time t
		(
			local tmp = (tf.frame/framerate)
			local pos = r.pos - [0,0,r.gravity*tmp*tmp]
			local q = (r.rotation as quat)
			local p2 = [r.above,0,0]
			local p1 = [-r.below,0,0]
			#(((RotatePoint p1 q) + pos),((RotatePoint p2 q) + pos))
		)
	)
	local numPlanes = round (r.life*r.emission)
	local step = (framerate/(r.emission as float)) as time
	local start, end
	local vertices = #()
	local tvertices = #()
	local faces = #()
	local face1 = [2,1,4]
	local face2 = [3,4,1]
	local pos = undefined
	local i = currentTime
	while (numPlanes != 0) do
	(
		numPlanes -= 1
		at time i if r.visibility then 
			if start == undefined then start = i else end = i
		i -= step
	)
	if start != undefined then
	(
		while (start >= end) do
		(
			at time start local pos2 = r.pos
			if (pos != pos2) then
			(
				if vertices.count != 0 then
				(
					append faces face1
					append faces face2
					face1 += 2
					face2 += 2
				)
				join vertices (GenerateVertices start r (currentTime-start))
			)
			start -= step
		)
		
		local geo = mesh vertices:vertices faces:faces
		setNumTVerts geo vertices.count
		setTVert geo 1 [0,0,0]
		setTVert geo 2 [0,1,0]
		local j = 3
		local totalPlanes = faces.count/2
		for i = 1 to totalPlanes do
		(
			local tmp = (i/(totalPlanes as float))
			setTVert geo j [tmp,0,0]
			setTVert geo (j+1) [tmp,1,0]
			j += 2
		)
		buildTVFaces geo
		
		for i = 1 to faces.count do setTVFace geo i faces[i]
		if r.rmaterial != undefined then
		(
			local rm = r.rmaterial
			local m = Standard ()
			m.name = "Tmp: " + r.name
			m.opacity = rm.alpha
			m.twoSided = rm.twosides
			m.useSelfIllumColor = true
			m.selfIllumColor = r.VertexColor
			m.diffuseMap = rm.texture
			if rm.filtermode > 1 then m.opacityMap = rm.texture
			geo.material = m
		)
		geo.visibility = linear_float()
		geo.visibility.controller.value = r.alpha
		geo
	)
	else undefined
)

fn PickRibbonEmitters =
(
	result = #()
	for h in helpers do
		if (not h.isHidden) and (classof h) as string == "BlizRibbon" then
			append result h
	result
)

fn PickParticleEmitters2 =
(
	result = #()
	for h in helpers do
		if (not h.isHidden) and (classof h)== BlizzParticle2 then
			append result h
	result
)

global Wc3TmpObjects = #()
global Wc3LightState = 2
global Wc3RenderUseSkylight = true
global Wc3AmbLight = #(color 0 0 0, color 255 255 255, color 211 204 127, color 75 92 117)
global Wc3BeforeAmbient = color 0 0 0
global Wc3GlobalLight = #(color 255 255 255, color 0 0 0, color 128 51 44, color 138 163 180)
global Wc3BeforeGlobalTint = color 255 255 255
fn preRenderWc3 =
(
	local isRenderAnimated = (rendTimeType > 1)
	Wc3TmpObjects = #()
	local ribbons = PickRibbonEmitters ()
	local emitters = PickParticleEmitters2 ()
	for r in ribbons do append Wc3TmpObjects (RibbonGenerateMesh r)
	--for p in emitters do append Wc3TmpObjects (createEmitter p currentTime)
	Wc3BeforeAmbient = ambientColor
	Wc3BeforeGlobalTint = lightTintColor
	ambientColor = Wc3AmbLight[Wc3LightState]
	lightTintColor =  Wc3GlobalLight[Wc3LightState]
)

fn postRenderWc3 =
(
	for w in Wc3TmpObjects where w != undefined do delete w
	Wc3TmpObjects = #()
	ambientColor = Wc3BeforeAmbient
	lightTintColor =  Wc3BeforeGlobalTint
)

global Wc3RenderAuto = false
